@@ -1,11 +1,11 @@
 /*
- * $Id: cond.c,v 2.6 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: cond.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
  *
  * cond.c - evaluate conditional expressions
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "cond.pro"
 
 /**/
 int
@@ -97,7 +98,7 @@ evalcond(Cond c)
     case 'w':
 	return (doaccess(c->left, W_OK));
     case 'x':
-	if (!geteuid()) {
+	if (privasserted()) {
 	    unsigned short mode = dostat(c->left);
 	    return (mode & 0111) || S_ISDIR(mode);
 	}
@@ -160,7 +161,7 @@ evalcond(Cond c)
 
 
 /**/
-int
+static int
 doaccess(char *s, int c)
 {
     return !access(unmeta(s), c);
@@ -170,7 +171,7 @@ doaccess(char *s, int c)
 static struct stat st;
 
 /**/
-struct stat *
+static struct stat *
 getstat(char *s)
 {
 /* /dev/fd/n refers to the open file descriptor n.  We always use fstat *
@@ -188,7 +189,7 @@ getstat(char *s)
 
 
 /**/
-unsigned short
+static unsigned short
 dostat(char *s)
 {
     struct stat *statp;
@@ -202,7 +203,7 @@ dostat(char *s)
 /* pem@aaii.oz; needed since dostat now uses "stat" */
 
 /**/
-unsigned short
+static unsigned short
 dolstat(char *s)
 {
     if (lstat(unmeta(s), &st) < 0)
@@ -212,7 +213,7 @@ dolstat(char *s)
 
 
 /**/
-int
+static int
 optison(char *s)
 {
     int i;
