@@ -1,11 +1,11 @@
 /*
- * $Id: exec.c,v 2.93 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: exec.c,v 3.1.2.11 1997/06/01 06:13:15 hzoli Exp $
  *
  * exec.c - command execution
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "exec.pro"
 
 #define execerr() if (!forked) { lastval = 1; return; } else _exit(1)
 	
@@ -90,7 +91,7 @@ setlimits(char *nam)
 /* fork and set limits */
 
 /**/
-pid_t
+static pid_t
 zfork(void)
 {
     pid_t pid;
@@ -123,7 +124,7 @@ static char list_pipe_text[JOBTEXTSIZE];
 /* execute a current shell command */
 
 /**/
-int
+static int
 execcursh(Cmd cmd)
 {
     if (!list_pipe)
@@ -138,7 +139,7 @@ execcursh(Cmd cmd)
 #define POUNDBANGLIMIT 64
 
 /**/
-int
+static int
 zexecve(char *pth, char **argv)
 {
     int eno;
@@ -228,7 +229,7 @@ zexecve(char *pth, char **argv)
 /* test whether we really want to believe the error number */
 
 /**/
-int
+static int
 isgooderr(int e, char *dir)
 {
     /*
@@ -252,6 +253,12 @@ execute(Cmdnam not_used_yet, int dash)
     char **argv, **pp;
     int eno = 0, ee;
 
+    arg0 = (char *) peekfirst(args);
+    if (isset(RESTRICTED) && strchr(arg0, '/')) {
+	zerr("%s: restricted", arg0, 0);
+	_exit(1);
+    }
+
     /* If the parameter STTY is set in the command's environment, *
      * we first run the stty command with the value of this       *
      * parameter as it arguments.                                 */
@@ -267,12 +274,11 @@ execute(Cmdnam not_used_yet, int dash)
 	exargs = NULL;
     }
 
-    arg0 = (char *) peekfirst(args);
     cn = (Cmdnam) cmdnamtab->getnode(cmdnamtab, arg0);
 
     /* If ARGV0 is in the commands environment, we use *
      * that as argv[0] for this external command       */
-    if ((z = zgetenv("ARGV0"))) {
+    if (unset(RESTRICTED) && (z = zgetenv("ARGV0"))) {
 	setdata(firstnode(args), (void *) ztrdup(z));
 	delenv(z - 6);
     } else if (dash) {
@@ -605,7 +611,7 @@ sublist_done:
  */
 
 /**/
-int
+static int
 execpline(Sublist l, int how, int last1)
 {
     int ipipe[2], opipe[2];
@@ -648,7 +654,6 @@ execpline(Sublist l, int how, int last1)
     pline_level--;
     if (how & Z_ASYNC) {
 	lastwj = newjob;
-	jobtab[thisjob].stat |= STAT_NOSTTY;
 	if (l->flags & PFLAG_COPROC)
 	    zclose(ipipe[1]);
 	if (how & Z_DISOWN) {
@@ -783,7 +788,7 @@ int subsh_close = -1;
 /* execute pipeline.  This function assumes the `pline' is not NULL. */
 
 /**/
-void
+static void
 execpline2(Pline pline, int how, int input, int output, int last1)
 {
     pid_t pid;
@@ -857,7 +862,7 @@ execpline2(Pline pline, int how, int input, int output, int last1)
 /* make the argv array */
 
 /**/
-char **
+static char **
 makecline(LinkList list)
 {
     LinkNode node;
@@ -901,7 +906,7 @@ untokenize(char *s)
 /* Open a file for writing redicection */
 
 /**/
-int
+static int
 clobber_open(struct redir *f)
 {
     struct stat buf;
@@ -909,13 +914,12 @@ clobber_open(struct redir *f)
 
     if (isset(CLOBBER) || IS_CLOBBER_REDIR(f->type))
 	return open(unmeta(f->name), O_WRONLY | O_CREAT | O_TRUNC, 0666);
-    if (stat(unmeta(f->name), &buf) || S_ISREG(buf.st_mode))
-	return open(unmeta(f->name), O_WRONLY | O_CREAT | O_EXCL, 0666);
+    if ((fd = open(unmeta(f->name), O_WRONLY | O_CREAT | O_EXCL, 0666)) >= 0)
+	return fd;
     if ((fd = open(unmeta(f->name), O_WRONLY)) < 0)
 	return fd;
     if (!fstat(fd, &buf) && !S_ISREG(buf.st_mode))
 	return fd;
-    /* someone replaced the file between stat and open */
     close(fd);
     errno = EEXIST;
     return -1;
@@ -927,7 +931,7 @@ clobber_open(struct redir *f)
 /* close an multio (success) */
 
 /**/
-void
+static void
 closemn(struct multio **mfds, int fd)
 {
     struct multio *mn = mfds[fd];
@@ -963,7 +967,7 @@ closemn(struct multio **mfds, int fd)
 /* close all the mnodes (failure) */
 
 /**/
-void
+static void
 closemnodes(struct multio **mfds)
 {
     int i, j;
@@ -977,7 +981,7 @@ closemnodes(struct multio **mfds)
 }
 
 /**/
-void
+static void
 closeallelse(struct multio *mn)
 {
     int i, j;
@@ -1006,7 +1010,7 @@ closeallelse(struct multio *mn)
  * multiple redirection.                                              */
 
 /**/
-void
+static void
 addfd(int forked, int *save, struct multio **mfds, int fd1, int fd2, int rflag)
 {
     int pipes[2];
@@ -1047,7 +1051,7 @@ addfd(int forked, int *save, struct multio **mfds, int fd1, int fd2, int rflag)
 }
 
 /**/
-void
+static void
 addvars(LinkList l, int export)
 {
     Varasg v;
@@ -1095,18 +1099,23 @@ addvars(LinkList l, int export)
 	    if (xtr)
 		fprintf(stderr, "%s ", val);
 	    if (export) {
-		if (export < 0)
+		if (export < 0) {
 		    /* We are going to fork so do not bother freeing this */
-		    paramtab->removenode(paramtab, v->name);
+		    pm = (Param) paramtab->removenode(paramtab, v->name);
+		    if (isset(RESTRICTED) && (pm->flags & PM_RESTRICTED)) {
+			zerr("%s: restricted", pm->nam, 0);
+			zsfree(val);
+			return;
+		    }
+		}
 		allexp = opts[ALLEXPORT];
 		opts[ALLEXPORT] = 1;
-		pm = setsparam(v->name, ztrdup(val));
+		pm = setsparam(v->name, val);
 		opts[ALLEXPORT] = allexp;
 	    } else
-		pm = setsparam(v->name, ztrdup(val));
+		pm = setsparam(v->name, val);
 	    if (errflag)
 		return;
-	    zsfree(val);
 	    continue;
 	}
 	ptr = arr = (char **) zalloc(sizeof(char **) * (countlinknodes(vl) + 1));
@@ -1128,7 +1137,7 @@ addvars(LinkList l, int export)
 }
 
 /**/
-void
+static void
 execcmd(Cmd cmd, int input, int output, int how, int last1)
 {
     HashNode hn = NULL;
@@ -1196,6 +1205,11 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 	    }
 	    if (!(hn->flags & BINF_PREFIX)) {
 		is_builtin = 1;
+#ifdef DYNAMIC
+		/* autoload the builtin if necessary */
+		if (!((Builtin) hn)->handlerfunc)
+		    load_module(((Builtin) hn)->optstr);
+#endif
 		assign = (hn->flags & BINF_MAGICEQUALS);
 		break;
 	    }
@@ -1267,6 +1281,11 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		    }
 		    return;
 		}
+	    } else if (isset(RESTRICTED) && (cflags & BINF_EXEC) &&
+		       (cmd->flags & CFLAG_EXEC)) {
+		zerrnam("exec", "%s: restricted", (char *) getdata(firstnode(args)), 0);
+		lastval = 1;
+		return;
 	    }
 
 	    if (errflag || checked ||
@@ -1289,6 +1308,11 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 	    }
 	    if (!(hn->flags & BINF_PREFIX)) {
 		is_builtin = 1;
+#ifdef DYNAMIC
+		/* autoload the builtin if necessary */
+		if (!((Builtin) hn)->handlerfunc)
+		    load_module(((Builtin) hn)->optstr);
+#endif
 		break;
 	    }
 	    cflags &= ~BINF_BUILTIN & ~BINF_COMMAND;
@@ -1480,7 +1504,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
     while (nonempty(cmd->redir)) {
 	fn = (Redir) ugetnode(cmd->redir);
 	DPUTS(fn->type == HEREDOC || fn->type == HEREDOCDASH,
-	      "BUG: unexpanded here document\n");
+	      "BUG: unexpanded here document");
 	if (fn->type == INPIPE) {
 	    if (fn->fd2 == -1) {
 		closemnodes(mfds);
@@ -1503,6 +1527,10 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		fixfds(save);
 		execerr();
 	    }
+	    if (isset(RESTRICTED) && IS_WRITE_FILE(fn->type)) {
+		zerr("writing redirection not allowed in restricted mode", NULL, 0);
+		execerr();
+	    }
 	    if (unset(EXECOPT))
 		continue;
 	    switch(fn->type) {
@@ -1614,7 +1642,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 	    static int (*func[]) _((Cmd)) = {
 		execcursh, exectime, execfuncdef, execfor, execwhile,
 		execrepeat, execif, execcase, execselect, execcond,
-		execarith
+		execarith, execautofn
 	    };
 
 	    if (last1 == 1)
@@ -1723,13 +1751,13 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 		execute((Cmdnam) hn, cflags & BINF_DASH);
 	    } else {		/* ( ... ) */
 		DPUTS(cmd->vars && nonempty(cmd->vars),
-		      "BUG: assigment before complex command.");
+		      "BUG: assigment before complex command");
 		list_pipe = 0;
 		if (subsh_close >= 0)
 		    zclose(subsh_close);
                 subsh_close = -1;
 		/* If we're forked (and we should be), no need to return */
-		DPUTS(last1 != 1 && !forked, "BUG? not exiting?");
+		DPUTS(last1 != 1 && !forked, "BUG: not exiting?");
 		execlist(cmd->u.list, 0, 1);
 	    }
 	}
@@ -1744,7 +1772,7 @@ execcmd(Cmd cmd, int input, int output, int how, int last1)
 /* Arrange to have variables restored. */
 
 /**/
-void
+static void
 save_params(Cmd cmd, LinkList *restore_p, LinkList *remove_p)
 {
     Param pm;
@@ -1761,7 +1789,8 @@ save_params(Cmd cmd, LinkList *restore_p, LinkList *remove_p)
 	if ((pm = (Param) paramtab->getnode(paramtab, s))) {
 	    if (!(pm->flags & PM_SPECIAL)) {
 		paramtab->removenode(paramtab, s);
-	    } else if (!(pm->flags & PM_READONLY)) {
+	    } else if (!(pm->flags & PM_READONLY) &&
+		       (unset(RESTRICTED) || !(pm->flags & PM_RESTRICTED))) {
 		Param tpm = (Param) alloc(sizeof *tpm);
 
 		tpm->nam = s;
@@ -1792,7 +1821,7 @@ save_params(Cmd cmd, LinkList *restore_p, LinkList *remove_p)
 /* Restore saved parameters after executing a shfunc or builtin */
 
 /**/
-void
+static void
 restore_params(LinkList restorelist, LinkList removelist)
 {
     Param pm;
@@ -1804,7 +1833,7 @@ restore_params(LinkList restorelist, LinkList removelist)
 	    if ((pm = (Param) paramtab->getnode(paramtab, s)) &&
 		!(pm->flags & PM_SPECIAL)) {
 		pm->flags &= ~PM_READONLY;
-		unsetparam_pm(pm, 0);
+		unsetparam_pm(pm, 0, 0);
 	    }
 	}
     }
@@ -1817,7 +1846,7 @@ restore_params(LinkList restorelist, LinkList removelist)
 
 		DPUTS(!tpm || PM_TYPE(pm->flags) != PM_TYPE(tpm->flags) ||
 		      !(pm->flags & PM_SPECIAL),
-		      "BUG in restoring special parameters.");
+		      "BUG: in restoring special parameters");
 		tpm->flags = pm->flags;
 		switch (PM_TYPE(pm->flags)) {
 		case PM_SCALAR:
@@ -1841,7 +1870,7 @@ restore_params(LinkList restorelist, LinkList removelist)
 /* restore fds after redirecting a builtin */
 
 /**/
-void
+static void
 fixfds(int *save)
 {
     int old_errno = errno;
@@ -1854,7 +1883,7 @@ fixfds(int *save)
 }
 
 /**/
-void
+static void
 entersubsh(int how, int cl, int fake)
 {
     int sig;
@@ -1997,7 +2026,7 @@ gethere(char *str, int typ)
 /* open here string fd */
 
 /**/
-int
+static int
 getherestr(struct redir *fn)
 {
     char *s, *t;
@@ -2093,7 +2122,7 @@ getoutput(char *cmd, int qt)
 /* read output of command substitution */
 
 /**/
-LinkList
+static LinkList
 readoutput(int in, int qt)
 {
     LinkList ret;
@@ -2141,7 +2170,7 @@ readoutput(int in, int qt)
 }
 
 /**/
-List
+static List
 parsecmd(char *cmd)
 {
     char *str;
@@ -2217,8 +2246,7 @@ getoutputfile(char *cmd)
 #if !defined(PATH_DEV_FD) && defined(HAVE_FIFOS)
 /* get a temporary named pipe */
 
-/**/
-char *
+static char *
 namedpipe(void)
 {
     char *tnam = gettempname();
@@ -2302,7 +2330,7 @@ getproc(char *cmd)
 /* > >(...) or < <(...) (does not use named pipes) */
 
 /**/
-int
+static int
 getpipe(char *cmd)
 {
     List list;
@@ -2326,7 +2354,7 @@ getpipe(char *cmd)
 /* open pipes with fds >= 10 */
 
 /**/
-void
+static void
 mpipe(int *pp)
 {
     pipe(pp);
@@ -2337,7 +2365,7 @@ mpipe(int *pp)
 /* Do process substitution with redirection */
 
 /**/
-void
+static void
 spawnpipes(LinkList l)
 {
     LinkNode n;
@@ -2357,7 +2385,7 @@ spawnpipes(LinkList l)
 /* evaluate a [[ ... ]] */
 
 /**/
-int
+static int
 execcond(Cmd cmd)
 {
     return !evalcond(cmd->u.cond);
@@ -2366,7 +2394,7 @@ execcond(Cmd cmd)
 /* evaluate a ((...)) arithmetic command */
 
 /**/
-int
+static int
 execarith(Cmd cmd)
 {
     char *e;
@@ -2381,7 +2409,7 @@ execarith(Cmd cmd)
 /* perform time ... command */
 
 /**/
-int
+static int
 exectime(Cmd cmd)
 {
     int jb;
@@ -2399,7 +2427,7 @@ exectime(Cmd cmd)
 /* Define a shell function */
 
 /**/
-int
+static int
 execfuncdef(Cmd cmd)
 {
     Shfunc shf;
@@ -2427,15 +2455,12 @@ execfuncdef(Cmd cmd)
     return 0;
 }
 
-/* Main entry point to execute a shell function.  It will retrieve *
- * an autoloaded shell function if it is currently undefined.      */
+/* Main entry point to execute a shell function. */
 
 /**/
-void
+static void
 execshfunc(Cmd cmd, Shfunc shf)
 {
-    List funcdef;
-    char *nam;
     LinkList last_file_list = NULL;
 
     if (errflag)
@@ -2449,34 +2474,50 @@ execshfunc(Cmd cmd, Shfunc shf)
 	deletejob(jobtab + thisjob);
     }
 
-    /* Are we dealing with an autoloaded shell function? */
-    if (shf->flags & PM_UNDEFINED) {
-	nam = (char *) peekfirst(cmd->args);
-	if (!(funcdef = getfpfunc(nam))) {
-	    zerr("function not found: %s", nam, 0);
-	    lastval = 1;
-	} else {
-	    PERMALLOC {
-		shf->flags &= ~PM_UNDEFINED;
-		funcdef = shf->funcdef = (List) dupstruct(funcdef);
-	    } LASTALLOC;
+    doshfunc(shf->funcdef, cmd->args, shf->flags, 0);
 
-	    /* Execute the function definition, we just retrived */
-	    doshfunc(shf->funcdef, cmd->args, shf->flags, 0);
-
-	    /* See if this file defined the autoloaded function *
-	     * by name.  If so, we execute it again.            */
-	    if ((shf = (Shfunc) shfunctab->getnode(shfunctab, nam))
-		&& shf->funcdef && shf->funcdef != funcdef)
-		doshfunc(shf->funcdef, cmd->args, shf->flags, 0);
-	}
-    } else
-	/* Normal shell function execution */
-	doshfunc(shf->funcdef, cmd->args, shf->flags, 0);
     if (!list_pipe)
 	deletefilelist(last_file_list);
 }
 
+/* Function to execute the special type of command that represents an *
+ * autoloaded shell function.  The command structure tells us which   *
+ * function it is.  This function is actually called as part of the   *
+ * execution of the autoloaded function itself, so when the function  *
+ * has been autoloaded, its list is just run with no frills.          */
+
+/**/
+static int
+execautofn(Cmd cmd)
+{
+    Shfunc shf = cmd->u.autofn->shf;
+    List l = getfpfunc(shf->nam);
+    if(l == &dummy_list) {
+	zerr("%s: function definition file not found", shf->nam, 0);
+	return 1;
+    }
+    if(isset(KSHAUTOLOAD)) {
+	VARARR(char, n, strlen(shf->nam) + 1);
+	strcpy(n, shf->nam);
+	execlist(l, 1, 0);
+	shf = (Shfunc) shfunctab->getnode(shfunctab, n);
+	if(!shf || (shf->flags & PM_UNDEFINED)) {
+	    zerr("%s: function not defined by file", n, 0);
+	    return 1;
+	}
+    } else {
+	freestruct(shf->funcdef);
+	PERMALLOC {
+	    shf->funcdef = dupstruct(stripkshdef(l, shf->nam));
+	} LASTALLOC;
+	shf->flags &= ~PM_UNDEFINED;
+    }
+    HEAPALLOC {
+	execlist(dupstruct(shf->funcdef), 1, 0);
+    } LASTALLOC;
+    return lastval;
+}
+
 /* execute a shell function */
 
 /**/
@@ -2486,11 +2527,9 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
  * value (lastval) to its value before the shell function   *
  * was executed.                                            */
 {
-    Param pm;
     char **tab, **x, *oargv0 = NULL;
     int xexittr, newexittr, oldzoptind, oldlastval;
-    LinkList olist;
-    char *s, *ou;
+    char *ou;
     void *xexitfn, *newexitfn;
     char saveopts[OPT_SIZE];
     int obreaks = breaks;
@@ -2538,27 +2577,14 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
 		argzero = ztrdup(argzero);
 	    }
 	}
-	PERMALLOC {
-	    olist = locallist;		/* save the old locallist since shell functions may be nested */
-	    locallist = newlinklist();	/* make a new list of local variables that we have to destroy */
-	} LASTALLOC;
-	locallevel++;
+	startparamscope();
 	ou = underscore;
 	underscore = ztrdup(underscore);
 	execlist(dupstruct(list), 1, 0);
 	zsfree(underscore);
 	underscore = ou;
-	locallevel--;
+	endparamscope();
 
-	/* destroy the local variables we have created in the shell function */
-	while ((s = (char *) getlinknode(locallist))) {
-	    if((pm = (Param) paramtab->getnode(paramtab, s)) && (pm->level > locallevel))
-		unsetparam_pm(pm, 0);
-	    zsfree(s);
-	}
-	zfree(locallist, sizeof(struct linklist));
-
-	locallist = olist;	/* restore the old list of local variables */
 	if (retflag) {
 	    retflag = 0;
 	    breaks = obreaks;
@@ -2572,8 +2598,9 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
 	pparams = tab;
 
 	if (isset(LOCALOPTIONS)) {
-	    /* restore all shell options except PRIVILEGED */
+	    /* restore all shell options except PRIVILEGED and RESTRICTED */
 	    saveopts[PRIVILEGED] = opts[PRIVILEGED];
+	    saveopts[RESTRICTED] = opts[RESTRICTED];
 	    memcpy(opts, saveopts, sizeof(opts));
 	} else {
 	    /* just restore a couple. */
@@ -2612,10 +2639,11 @@ doshfunc(List list, LinkList doshargs, int flags, int noreturnval)
     } LASTALLOC;
 }
 
-/* Search fpath for an undefined function. */
+/* Search fpath for an undefined function.  Finds the file, and returns the *
+ * list of its contents.                                                    */
 
 /**/
-List
+static List
 getfpfunc(char *s)
 {
     char **pp, buf[PATH_MAX];
@@ -2654,7 +2682,39 @@ getfpfunc(char *s)
 	    }
 	}
     }
-    return NULL;
+    return &dummy_list;
+}
+
+/* Handle the most common type of ksh-style autoloading, when doing a      *
+ * zsh-style autoload.  Given the list read from an autoload file, and the *
+ * name of the function being defined, check to see if the file consists   *
+ * entirely of a single definition for that function.  If so, use the      *
+ * contents of that definition.  Otherwise, use the entire file.           */
+
+/**/
+static List
+stripkshdef(List l, char *name)
+{
+    Sublist s;
+    Pline p;
+    Cmd c;
+    if(!l)
+	return NULL;
+    if(l->type != Z_SYNC || l->right)
+	return l;
+    s = l->left;
+    if(s->flags || s->right)
+	return l;
+    p = s->left;
+    if(p->right)
+	return l;
+    c = p->left;
+    if(c->type != FUNCDEF || c->flags ||
+	nonempty(c->redir) || nonempty(c->vars) ||
+	empty(c->args) || lastnode(c->args) != firstnode(c->args) ||
+	strcmp(name, peekfirst(c->args)))
+	return l;
+    return c->u.list;
 }
 
 /* check to see if AUTOCD applies here */
@@ -2662,7 +2722,7 @@ getfpfunc(char *s)
 extern int doprintdir;
 
 /**/
-char *
+static char *
 cancd(char *s)
 {
     int nocdpath = s[0] == '.' &&
@@ -2701,7 +2761,7 @@ cancd(char *s)
 }
 
 /**/
-int
+static int
 cancd2(char *s)
 {
     struct stat buf;
