@@ -1,9 +1,11 @@
 /*
+ * $Id: globals.h,v 3.1.2.10 1997/05/11 05:01:43 hzoli Exp $
+ *
  * globals.h - global variables for zsh
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -71,14 +73,20 @@ extern int  nulstrlen;
 
 /* NULL-terminated arrays containing path, cdpath, etc. */
  
-EXTERN char **path;		/* $path     */
-EXTERN char **cdpath;		/* $cdpath   */
-EXTERN char **fpath;		/* $fpath    */
-EXTERN char **watch;		/* $watch    */
-EXTERN char **mailpath;		/* $mailpath */
-EXTERN char **manpath;		/* $manpath  */
-EXTERN char **fignore;		/* $fignore  */
-EXTERN char **psvar;		/* $psvar    */
+EXTERN char **path;		/* $path        */
+EXTERN char **cdpath;		/* $cdpath      */
+EXTERN char **fpath;		/* $fpath       */
+EXTERN char **watch;		/* $watch       */
+EXTERN char **mailpath;		/* $mailpath    */
+EXTERN char **manpath;		/* $manpath     */
+EXTERN char **fignore;		/* $fignore     */
+EXTERN char **psvar;		/* $psvar       */
+#ifdef DYNAMIC
+EXTERN char **module_path;	/* $module_path */
+
+/* List of modules */
+EXTERN LinkList modules;
+#endif
  
 EXTERN char *yytext;
 
@@ -182,6 +190,10 @@ EXTERN int infunc;
  
 EXTERN int isnewlin;
 
+/* != 0 if we are after a for keyword */
+
+EXTERN int infor;
+
 /* the lists of history events */
  
 EXTERN LinkList histlist;
@@ -194,6 +206,10 @@ EXTERN LinkList dirstack;
  
 EXTERN LinkList bufstack;
 
+/* extra functions to call before displaying the prompt */
+
+EXTERN LinkList prepromptfns;
+
 /* total # of characters waiting to be read. */
 
 EXTERN int inbufct;
@@ -266,29 +282,13 @@ EXTERN HashTable shfunctab;
 
 EXTERN HashTable builtintab;
  
-/* hash table for completion info for commands */
- 
-EXTERN HashTable compctltab;
-
-/* hash table for multi-character bindings */
-
-EXTERN HashTable keybindtab;
-
-/* hash table for emacs multi-character bindings */
-
-EXTERN HashTable emkeybindtab;
-
-/* hash table for vi multi-character bindings */
-
-EXTERN HashTable vikeybindtab;
-
 /* hash table for named directories */
 
 EXTERN HashTable nameddirtab;
  
-/* default completion infos */
- 
-EXTERN struct compctl cc_compos, cc_default, cc_first, cc_dummy;
+/* hash table for option names */
+
+extern HashTable optiontab;
  
 /* the job table */
  
@@ -298,10 +298,6 @@ EXTERN struct job jobtab[MAXJOB];
  
 EXTERN struct tms shtms;
  
-/* the list of sched jobs pending */
- 
-EXTERN struct schedcmd *schedcmds;
- 
 /* the last l for s/l/r/ history substitution */
  
 EXTERN char *hsubl;
@@ -318,19 +314,19 @@ EXTERN char *cached_username;   /* $USERNAME   */
 EXTERN char *zsh_name;		/* ZSH_NAME    */
 
 EXTERN char *underscore;	/* $_          */
-EXTERN zlong lastval;		/* $?          */
-EXTERN zlong mypid;		/* $$          */
-EXTERN zlong lastpid;		/* $!          */
-EXTERN zlong ppid;		/* $PPID       */
+EXTERN long lastval;            /* $?          */
+EXTERN long mypid;		/* $$          */
+EXTERN long lastpid;		/* $!          */
+EXTERN long ppid;		/* $PPID       */
 EXTERN char *ifs;		/* $IFS        */
 EXTERN char *pwd;		/* $PWD        */
 EXTERN char *oldpwd;		/* $OLDPWD     */
 
-EXTERN zlong columns;		/* $COLUMNS    */
-EXTERN zlong lines;		/* $LINES      */
+EXTERN long columns;            /* $COLUMNS    */
+EXTERN long lines;              /* $LINES      */
 
 EXTERN char *zoptarg;		/* $OPTARG     */
-EXTERN zlong zoptind;		/* $OPTIND     */
+EXTERN long zoptind;		/* $OPTIND     */
 EXTERN char *prompt;		/* $PROMPT     */
 EXTERN char *prompt2;		/* etc.        */
 EXTERN char *prompt3;
@@ -350,11 +346,10 @@ EXTERN pid_t mypgrp;		/* the process group of the shell */
  
 EXTERN char *argzero;           /* $0 */
  
-EXTERN char *hackzero;
 EXTERN char *scriptname;        /* name of script being sourced */
 
-EXTERN zlong lineno;		/* $LINENO       */
-EXTERN zlong shlvl;		/* $SHLVL        */
+EXTERN long lineno;             /* $LINENO       */
+EXTERN long shlvl;              /* $SHLVL        */
  
 EXTERN long lastval2;
 
@@ -379,10 +374,6 @@ EXTERN struct timeval shtimer;
 EXTERN char *nullcmd;
 EXTERN char *readnullcmd;
  
-/* the List of local variables we have to destroy */
- 
-EXTERN LinkList locallist;
-
 /* what level of localness we are at */
  
 EXTERN int locallevel;
@@ -433,10 +424,6 @@ EXTERN FILE *shout;
 
 EXTERN FILE *bshin;
 
-/* the FILE for xtrace output */
-
-EXTERN FILE *xtrerr;
-
 /* != 0 means we are reading input from a string */
  
 EXTERN int strin;
@@ -530,19 +517,63 @@ EXTERN int ttyfrozen;
 /* != 0 if we are allocating in the heaplist */
  
 EXTERN int useheap;
- 
-/* Words on the command line, for use in completion */
- 
-EXTERN int clwsize, clwnum, clwpos;
-EXTERN char **clwords;
 
-/* Non-zero if a completion list was displayed. */
+/* Current allocation pointers.  ncalloc() is either zalloc() or halloc(); *
+ * alloc() is either zcalloc() or hcalloc().                               */
 
-EXTERN int listshown;
+EXTERN void *(*ncalloc) _((size_t));
+EXTERN void *(*alloc) _((size_t));
 
-/* Non-zero if refresh() should clear the list below the prompt. */
+/* != 0 if the line editor is active */
+EXTERN int zleactive;
 
-EXTERN int clearlist;
+/* cursor position */
+EXTERN int cs;
+
+/* the line buffer */
+EXTERN unsigned char *line;
+
+/* line length */
+EXTERN int ll;
+
+/* inwhat says what exactly we are in     *
+ * (its value is one of the IN_* things). */
+EXTERN int inwhat;
+
+/* 1 if x added to complete in a blank between words */
+EXTERN int addedx;
+
+/* left prompt and right prompt */
+EXTERN char *lpmpt, *rpmpt;
+
+/* height of left prompt */
+EXTERN int lppth;
+
+/* != 0 if we need to call resetvideo() */
+EXTERN int resetneeded;
+
+#ifdef TIOCGWINSZ
+EXTERN int winchanged;		/* window size changed */
+#endif
+
+/* flag for whether terminal has automargin (wraparound) capability */
+EXTERN int hasam;
+
+/* compctl entry point pointers */
+
+typedef int (*CompctlReadFn) _((char *, char **, char *, char *));
+
+extern CompctlReadFn compctlreadptr;
+
+/* ZLE entry point pointers */
+
+typedef void (*ZleVoidFn) _((void));
+typedef void (*ZleVoidIntFn) _((int));
+typedef unsigned char * (*ZleReadFn) _((char *, char *, int));
+
+extern ZleVoidFn trashzleptr, gotwordptr, refreshptr;
+extern ZleVoidIntFn spaceinlineptr;
+extern ZleReadFn zlereadptr;
 
 /* pid of process undergoing 'process substitution' */
  
@@ -577,11 +608,6 @@ EXTERN int signal_queue[MAX_QUEUE_SIZE];
 EXTERN int queue_front;
 EXTERN int queue_rear;
 
-/* Previous values of errflag and breaks if the signal handler had to
- * change them. And a flag saying if it did that. */
-
-EXTERN int prev_errflag, prev_breaks, errbrk_saved;
-
 /* 1 if aliases should not be expanded */
  
 EXTERN int noaliases;
@@ -629,7 +655,7 @@ char *tccapnams[TC_COUNT] =
 {
     "cl", "le", "LE", "nd", "RI", "up", "UP", "do",
     "DO", "dc", "DC", "ic", "IC", "cd", "ce", "al", "dl", "ta",
-    "md", "so", "us", "me", "se", "ue", "ch"
+    "md", "so", "us", "me", "se", "ue"
 };
 #else
 extern char *tccapnams[TC_COUNT];
@@ -652,7 +678,7 @@ char *tokstrings[WHILE + 1] = {
     ")",	/* OUTPAR	     */
     "||",	/* DBAR		     */
     "&&",	/* DAMPER	     */
-    ">",	/* OUTANG	  10 */
+    ")",	/* OUTANG	  10 */
     ">|",	/* OUTANGBANG	     */
     ">>",	/* DOUTANG	     */
     ">>|",	/* DOUTANGBANG	     */
@@ -673,6 +699,7 @@ char *tokstrings[WHILE + 1] = {
     "((",	/* DINPAR	     */
     "))",	/* DOUTPAR	     */
     "&|",	/* AMPERBANG	  30 */
+    ";&",	/* SEMIAMP	     */
 };
 #else
 extern char *tokstrings[];
@@ -694,119 +721,4 @@ char *cmdnames[] =
 extern char *cmdnames[];
 #endif
  
-#ifndef GLOBALS
-extern struct option optns[OPT_SIZE];
-#else
-struct option optns[OPT_SIZE] = {
-# define x OPT_REV|
-    {NULL, 0, 0, 0},
-    {"allexport", 		'a',  'a',  OPT_EMULATE},
-    {"alwayslastprompt", 	0,    0,    0},
-    {"alwaystoend", 		0,    0,    0},
-    {"appendhistory", 		0,    0,    0},
-    {"autocd", 			'J',  0,    OPT_EMULATE},
-    {"autolist", 		'9',  0,    0},
-    {"automenu", 		0,    0,    0},
-    {"autonamedirs", 		0,    0,    0},
-    {"autoparamkeys", 		0,    0,    0},
-    {"autoparamslash", 		0,    0,    OPT_CSH},
-    {"autopushd", 		'N',  0,    0},
-    {"autoremoveslash", 	0,    0,    0},
-    {"autoresume", 		'W',  0,    0},
-    {"badpattern", 		x'2', 0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"banghist", 		x'K', 0,    OPT_NONBOURNE},
-    {"beep", 			x'B', 0,    OPT_ALL},
-    {"bgnice", 			'6',  0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"braceccl", 		0,    0,    OPT_EMULATE},
-    {"bsdecho", 		0,    0,    OPT_EMULATE|OPT_SH},
-    {"cdablevars", 		'T',  0,    OPT_EMULATE},
-    {"chaselinks", 		'w',  0,    OPT_EMULATE},
-    {"clobber", 		x'C', x'C', OPT_EMULATE|OPT_ALL},
-    {"completealiases", 	0,    0,    0},
-    {"completeinword", 		0,    0,    0},
-    {"correct", 		'0',  0,    0},
-    {"correctall", 		'O',  0,    0},
-    {"cshjunkiehistory", 	0,    0,    OPT_EMULATE|OPT_CSH},
-    {"cshjunkieloops", 		0,    0,    OPT_EMULATE|OPT_CSH},
-    {"cshjunkiequotes", 	0,    0,    OPT_EMULATE|OPT_CSH},
-    {"cshnullglob", 		0,    0,    OPT_EMULATE|OPT_CSH},
-    {"equals", 			0,    0,    OPT_EMULATE|OPT_ZSH},
-    {"errexit", 		'e',  'e',  OPT_EMULATE},
-    {"exec", 			x'n', x'n', OPT_ALL},
-    {"extendedglob", 		0,    0,    OPT_EMULATE},
-    {"extendedhistory", 	0,    0,    OPT_CSH},
-    {"flowcontrol", 		0,    0,    OPT_ALL},
-    {"functionargzero",		0,    0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"glob", 			x'F', x'f', OPT_EMULATE|OPT_ALL},
-    {"globassign", 		0,    0,    OPT_EMULATE|OPT_CSH},
-    {"globcomplete", 		0,    0,    0},
-    {"globdots", 		'4',  0,    OPT_EMULATE},
-    {"globsubst", 		0,    0,    OPT_EMULATE|OPT_NONZSH},
-    {"hashcmds", 		0,    0,    OPT_ALL},
-    {"hashdirs", 		0,    0,    OPT_ALL},
-    {"hashlistall", 		0,    0,    OPT_ALL},
-    {"histallowclobber", 	0,    0,    0},
-    {"histbeep", 		0,    0,    OPT_ALL},
-    {"histignoredups", 		'h',  0,    0},
-    {"histignorespace", 	'g',  0,    0},
-    {"histnostore", 		0,    0,    0},
-    {"histreduceblanks",	0,    0,    0},
-    {"histverify", 		0,    0,    0},
-    {"hup", 			0,    0,    OPT_EMULATE|OPT_ZSH},
-    {"ignorebraces", 		'I',  0,    OPT_EMULATE|OPT_SH},
-    {"ignoreeof", 		'7',  0,    0},
-    {"interactive", 		'i',  'i',  OPT_SPECIAL},
-    {"interactivecomments", 	'k',  0,    OPT_BOURNE},
-    {"ksharrays", 		0,    0,    OPT_EMULATE|OPT_BOURNE},
-    {"kshoptionprint", 		0,    0,    OPT_EMULATE|OPT_KSH},
-    {"listambiguous", 		0,    0,    0},
-    {"listbeep", 		0,    0,    OPT_ALL},
-    {"listtypes", 		'X',  0,    OPT_CSH},
-    {"localoptions", 		0,    0,    OPT_EMULATE|OPT_KSH},
-    {"login", 			'l',  'l',  OPT_SPECIAL},
-    {"longlistjobs", 		'R',  0,    0},
-    {"magicequalsubst", 	0,    0,    OPT_EMULATE},
-    {"mailwarning", 		'U',  0,    0},
-    {"markdirs", 		'8',  'X',  0},
-    {"menucomplete", 		'Y',  0,    0},
-    {"monitor", 		'm',  'm',  OPT_SPECIAL},
-    {"multios", 		0,    0,    OPT_EMULATE|OPT_ZSH},
-    {"nomatch", 		x'3', 0,    OPT_EMULATE|OPT_NONBOURNE},
-    {"notify", 			'5',  'b',  OPT_ZSH},
-    {"nullglob", 		'G',  0,    OPT_EMULATE},
-    {"numericglobsort", 	0,    0,    OPT_EMULATE},
-    {"overstrike", 		0,    0,    0},
-    {"pathdirs", 		'Q',  0,    OPT_EMULATE},
-    {"posixbuiltins",		0,    0,    OPT_EMULATE|OPT_BOURNE},
-    {"printeightbit", 		0,    0,    0},
-    {"printexitvalue", 		'1',  0,    0},
-    {"privileged", 		'p',  'p',  OPT_SPECIAL},
-    {"promptcr", 		x'V', 0,    OPT_ALL},
-    {"promptsubst", 		0,    0,    OPT_KSH},
-    {"pushdignoredups", 	0,    0,    OPT_EMULATE},
-    {"pushdminus", 		0,    0,    OPT_EMULATE},
-    {"pushdsilent", 		'E',  0,    0},
-    {"pushdtohome", 		'D',  0,    OPT_EMULATE},
-    {"rcexpandparam", 		'P',  0,    OPT_EMULATE},
-    {"rcquotes", 		0,    0,    OPT_EMULATE},
-    {"rcs", 			x'f', 0,    OPT_ALL},
-    {"recexact", 		'S',  0,    0},
-    {"rmstarsilent", 		'H',  0,    OPT_BOURNE},
-    {"shfileexpansion",		0,    0,    OPT_EMULATE|OPT_BOURNE},
-    {"shglob", 			0,    0,    OPT_EMULATE|OPT_BOURNE},
-    {"shinstdin", 		's',  's',  OPT_SPECIAL},
-    {"shoptionletters",		0,    0,    OPT_EMULATE|OPT_BOURNE},
-    {"shortloops", 		0,    0,    OPT_NONBOURNE},
-    {"shwordsplit", 		'y',  0,    OPT_EMULATE|OPT_BOURNE},
-    {"singlecommand",		't',  't',  OPT_SPECIAL},
-    {"singlelinezle", 		'M',  0,    OPT_KSH},
-    {"sunkeyboardhack", 	'L',  0,    0},
-    {"unset", 			x'u', x'u', OPT_EMULATE|OPT_BSHELL},
-    {"verbose", 		'v',  'v',  0},
-    {"xtrace", 			'x',  'x',  0},
-    {"zle", 			'Z',  0,    OPT_SPECIAL},
-};
-# undef x
-#endif
-
 EXTERN short int typtab[256];
