@@ -1,11 +1,11 @@
 /*
- * $Id: hashtable.c,v 2.11 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: hashtable.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
  *
  * hashtable.c - hash tables
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -29,12 +29,53 @@
  *
  */
 
+#include <config.h>
+
+#ifdef ZSH_HASH_DEBUG
+# define HASHTABLE_DEBUG_MEMBERS \
+    /* Members of struct hashtable used for debugging hash tables */ \
+    HashTable next, last;	/* linked list of all hash tables           */ \
+    char *tablename;		/* string containing name of the hash table */ \
+    PrintTableStats printinfo;	/* pointer to function to print table stats */
+#else /* !ZSH_HASH_DEBUG */
+# define HASHTABLE_DEBUG_MEMBERS
+#endif /* !ZSH_HASH_DEBUG */
+
+#define HASHTABLE_INTERNAL_MEMBERS \
+    ScanStatus scan;		/* status of a scan over this hashtable     */ \
+    HASHTABLE_DEBUG_MEMBERS
+
+typedef struct scanstatus *ScanStatus;
+
 #include "zsh.h"
+#include "hashtable.pro"
+
+/* Structure for recording status of a hashtable scan in progress.  When a *
+ * scan starts, the .scan member of the hashtable structure points to one  *
+ * of these.  That member being non-NULL disables resizing of the          *
+ * hashtable (when adding elements).  When elements are deleted, the       *
+ * contents of this structure is used to make sure the scan won't stumble  *
+ * into the deleted element.                                               */
+
+struct scanstatus {
+    int sorted;
+    union {
+	struct {
+	    HashNode *tab;
+	    int ct;
+	} s;
+	HashNode u;
+    } u;
+};
 
 /********************************/
 /* Generic Hash Table functions */
 /********************************/
 
+#ifdef ZSH_HASH_DEBUG
+static HashTable firstht, lastht;
+#endif /* ZSH_HASH_DEBUG */
+
 /* Generic hash function */
 
 /**/
@@ -53,17 +94,51 @@ hasher(char *str)
 
 /**/
 HashTable
-newhashtable(int size)
+newhashtable(int size, char const *name, PrintTableStats printinfo)
 {
     HashTable ht;
 
     ht = (HashTable) zcalloc(sizeof *ht);
+#ifdef ZSH_HASH_DEBUG
+    ht->next = NULL;
+    if(!firstht)
+	firstht = ht;
+    ht->last = lastht;
+    if(lastht)
+	lastht->next = ht;
+    lastht = ht;
+    ht->printinfo = printinfo ? printinfo : printhashtabinfo;
+    ht->tablename = ztrdup(name);
+#endif /* ZSH_HASH_DEBUG */
     ht->nodes = (HashNode *) zcalloc(size * sizeof(HashNode));
     ht->hsize = size;
     ht->ct = 0;
+    ht->scan = NULL;
     return ht;
 }
 
+/* Delete a hash table.  After this function has been used, any *
+ * existing pointers to the hash table are invalid.             */
+
+/**/
+void
+deletehashtable(HashTable ht)
+{
+    ht->emptytable(ht);
+#ifdef ZSH_HASH_DEBUG
+    if(ht->next)
+	ht->next->last = ht->last;
+    else
+	lastht = ht->last;
+    if(ht->last)
+	ht->last->next = ht->next;
+    else
+	firstht = ht->next;
+#endif /* ZSH_HASH_DEBUG */
+    zfree(ht->nodes, ht->hsize * sizeof(HashNode));
+    zfree(ht, sizeof(*ht));
+}
+
 /* Add a node to a hash table.                          *
  * nam is the key to use in hashing.  dat is a pointer  *
  * to the node to add.  If there is already a node in   *
@@ -89,15 +164,25 @@ addhashnode(HashTable ht, char *nam, void *nodeptr)
     if (!hp) {
 	hn->next = NULL;
 	ht->nodes[hashval] = hn;
-	if (++ht->ct == ht->hsize * 2)
+	if (++ht->ct >= ht->hsize * 2 && !ht->scan)
 	    expandhashtable(ht);
 	return;
     }
 
     /* else check if the first node contains the same key */
     if (!strcmp(hp->nam, hn->nam)) {
-	hn->next = hp->next;
 	ht->nodes[hashval] = hn;
+	replacing:
+	hn->next = hp->next;
+	if(ht->scan)
+	    if(ht->scan->sorted) {
+		HashNode *tab = ht->scan->u.s.tab;
+		int i;
+		for(i = ht->scan->u.s.ct; i--; )
+		    if(tab[i] == hp)
+			tab[i] = hn;
+	    } else if(ht->scan->u.u == hp)
+		ht->scan->u.u = hn;
 	ht->freenode(hp);
 	return;
     }
@@ -107,17 +192,15 @@ addhashnode(HashTable ht, char *nam, void *nodeptr)
     hp = hp->next;
     for (; hp; hq = hp, hp = hp->next) {
 	if (!strcmp(hp->nam, hn->nam)) {
-	    hn->next = hp->next;
 	    hq->next = hn;
-	    ht->freenode(hp);
-	    return;
+	    goto replacing;
 	}
     }
 
     /* else just add it at the front of the list */
     hn->next = ht->nodes[hashval];
     ht->nodes[hashval] = hn;
-    if (++ht->ct == ht->hsize * 2)
+    if (++ht->ct >= ht->hsize * 2 && !ht->scan)
         expandhashtable(ht);
 }
 
@@ -187,7 +270,17 @@ removehashnode(HashTable ht, char *nam)
     /* else check if the key in the first one matches */
     if (!strcmp(hp->nam, nam)) {
 	ht->nodes[hashval] = hp->next;
+	gotit:
 	ht->ct--;
+	if(ht->scan)
+	    if(ht->scan->sorted) {
+		HashNode *tab = ht->scan->u.s.tab;
+		int i;
+		for(i = ht->scan->u.s.ct; i--; )
+		    if(tab[i] == hp)
+			tab[i] = NULL;
+	    } else if(ht->scan->u.u == hp)
+		ht->scan->u.u = hp->next;
 	return hp;
     }
 
@@ -197,8 +290,7 @@ removehashnode(HashTable ht, char *nam)
     for (; hp; hq = hp, hp = hp->next) {
 	if (!strcmp(hp->nam, nam)) {
 	    hq->next = hp->next;
-	    ht->ct--;
-	    return hp;
+	    goto gotit;
 	}
     }
 
@@ -227,172 +319,118 @@ enablehashnode(HashNode hn, int flags)
 /* Compare two hash table entries */
 
 /**/
-int
-hnamcmp(struct hashnode **a, struct hashnode **b)
+static int
+hnamcmp(const void *ap, const void *bp)
 {
-    return ztrcmp((unsigned char *) (*a)->nam, (unsigned char *) (*b)->nam);
+    HashNode a = *(HashNode *)ap;
+    HashNode b = *(HashNode *)bp;
+    return ztrcmp((unsigned char *) a->nam, (unsigned char *) b->nam);
 }
 
-/* Scan the nodes in a hash table and execute scanfunc on nodes based on the flags *
- * that are set/unset.  scanflags is passed unchanged to scanfunc (if executed).   *
- *                                                                                 *
- * If sorted = 1, then sort entries of hash table before scanning.                 *
- * If sorted = 0, don't sort entries before scanning.                              *
- * If (flags1 > 0), then execute func on a node only if these flags are set.       *
- * If (flags2 > 0), then execute func on a node only if these flags are NOT set.   *
- * The conditions above for flags1/flags2 must both be true.                       */
+/* Scan the nodes in a hash table and execute scanfunc on nodes based on
+ * the flags that are set/unset.  scanflags is passed unchanged to
+ * scanfunc (if executed).
+ *
+ * If sorted != 0, then sort entries of hash table before scanning.
+ * If flags1 > 0, then execute scanfunc on a node only if at least one of
+ *                these flags is set.
+ * If flags2 > 0, then execute scanfunc on a node only if all of
+ *                these flags are NOT set.
+ * The conditions above for flags1/flags2 must both be true.
+ *
+ * It is safe to add, remove or replace hash table elements from within
+ * the scanfunc.  Replaced elements will appear in the scan exactly once,
+ * the new version if it was not scanned before the replacement was made.
+ * Added elements might or might not appear in the scan.
+ */
 
 /**/
 void
 scanhashtable(HashTable ht, int sorted, int flags1, int flags2, ScanFunc scanfunc, int scanflags)
 {
-    HashNode hn, *hnsorttab, *htp;
-    int i;
+    struct scanstatus st;
 
     if (sorted) {
-	hnsorttab = (HashNode *) zalloc(ht->ct * sizeof(HashNode));
+	int i, ct = ht->ct;
+	VARARR(HashNode, hnsorttab, ct);
+	HashNode *htp, hn;
 
 	for (htp = hnsorttab, i = 0; i < ht->hsize; i++)
 	    for (hn = ht->nodes[i]; hn; hn = hn->next)
 		*htp++ = hn;
+	qsort((void *)hnsorttab, ct, sizeof(HashNode), hnamcmp);
 
-	qsort((void *) & hnsorttab[0], ht->ct, sizeof(HashNode),
-	           (int (*) _((const void *, const void *))) hnamcmp);
+	st.sorted = 1;
+	st.u.s.tab = hnsorttab;
+	st.u.s.ct = ct;
+	ht->scan = &st;
 
-	/* Ignore the flags */
-	if (!flags1 && !flags2) {
-	    for (htp = hnsorttab, i = 0; i < ht->ct; i++, htp++)
+	for (htp = hnsorttab, i = 0; i < ct; i++, htp++)
+	    if (*htp && ((*htp)->flags & flags1) + !flags1 &&
+		!((*htp)->flags & flags2))
 		scanfunc(*htp, scanflags);
-	} else if (flags1 && !flags2) {
-	/* Only exec scanfunc if flags1 are set */
-	    for (htp = hnsorttab, i = 0; i < ht->ct; i++, htp++)
-		if ((*htp)->flags & flags1)
-		    scanfunc(*htp, scanflags);
-	} else if (!flags1 && flags2) {
-	/* Only exec scanfunc if flags2 are NOT set */
-	    for (htp = hnsorttab, i = 0; i < ht->ct; i++, htp++)
-		if (!((*htp)->flags & flags2))
-		    scanfunc(*htp, scanflags);
-	} else {
-	/* Only exec scanfun if flags1 are set, and flags2 are NOT set */
-	    for (htp = hnsorttab, i = 0; i < ht->ct; i++, htp++)
-		if (((*htp)->flags & flags1) && !((*htp)->flags & flags2))
-		    scanfunc(*htp, scanflags);
-	}
-    free(hnsorttab);
-    return;
-    }
 
-    /* Don't sort, just use hash order. */
+	ht->scan = NULL;
+    } else {
+	int i, hsize = ht->hsize;
+	HashNode *nodes = ht->nodes;
 
-    /* Ignore the flags */
-    if (!flags1 && !flags2) {
-	for (i = 0; i < ht->hsize; i++)
-	    for (hn = ht->nodes[i]; hn; hn = hn->next)
-		scanfunc(hn, scanflags);
-	return;
-    }
+	st.sorted = 0;
+	ht->scan = &st;
 
-    /* Only exec scanfunc if flags1 are set */
-    if (flags1 && !flags2) {
-	for (i = 0; i < ht->hsize; i++)
-	    for (hn = ht->nodes[i]; hn; hn = hn->next)
-		if (hn->flags & flags1)
+	for (i = 0; i < hsize; i++)
+	    for (st.u.u = nodes[i]; st.u.u; ) {
+		HashNode hn = st.u.u;
+		st.u.u = st.u.u->next;
+		if ((hn->flags & flags1) + !flags1 && !(hn->flags & flags2))
 		    scanfunc(hn, scanflags);
-	return;
-    }
+	    }
 
-    /* Only exec scanfunc if flags2 are NOT set */
-    if (!flags1 && flags2) {
-	for (i = 0; i < ht->hsize; i++)
-	    for (hn = ht->nodes[i]; hn; hn = hn->next)
-		if (!(hn->flags & flags2))
-		    scanfunc(hn, scanflags);
-	return;
+	ht->scan = NULL;
     }
-
-    /* Only exec scanfun if flags1 are set, and flags2 are NOT set */
-    for (i = 0; i < ht->hsize; i++)
-	for (hn = ht->nodes[i]; hn; hn = hn->next)
-	    if ((hn->flags & flags1) && !(hn->flags & flags2))
-		scanfunc(hn, scanflags);
 }
 
-
 /* Scan all nodes in a hash table and executes scanfunc on the *
  * nodes which meet all the following criteria:                *
  * The hash key must match the glob pattern given by `com'.    *
- * If (flags1 > 0), then all flags in flags1 must be set.      *
+ * If (flags1 > 0), then any flag in flags1 must be set.       *
  * If (flags2 > 0), then all flags in flags2 must NOT be set.  *
  *                                                             *
  * scanflags is passed unchanged to scanfunc (if executed).    *
- * The return value if the number of matches.                  */
+ * The return value is the number of matches.                  */
 
 /**/
 int
 scanmatchtable(HashTable ht, Comp com, int flags1, int flags2, ScanFunc scanfunc, int scanflags)
 {
-    HashNode hn;
-    int i, match = 0;
+    int i, hsize = ht->hsize;
+    HashNode *nodes = ht->nodes;
+    int match = 0;
+    struct scanstatus st;
 
-    /* ignore the flags */
-    if (!flags1 && !flags2) {
-	for (i = 0; i < ht->hsize; i++) {
-	    for (hn = ht->nodes[i]; hn; hn = hn->next) {
-		if (domatch(hn->nam, com, 0)) {
-		    scanfunc(hn, scanflags);
-		    match++;
-		}
-	    }
-	}
-    return match;
-    }
+    st.sorted = 0;
+    ht->scan = &st;
 
-    /* flags in flags1 must be set */
-    if (flags1 && !flags2) {
-	for (i = 0; i < ht->hsize; i++) {
-	    for (hn = ht->nodes[i]; hn; hn = hn->next) {
-		if (domatch(hn->nam, com, 0) && (hn->flags & flags1)) {
-		    scanfunc(hn, scanflags);
-		    match++;
-		}
-	    }
-	}
-    return match;
-    }
-
-    /* flags in flags2 must NOT be set */
-    if (!flags1 && flags2) {
-	for (i = 0; i < ht->hsize; i++) {
-	    for (hn = ht->nodes[i]; hn; hn = hn->next) {
-		if (domatch(hn->nam, com, 0) && !(hn->flags & flags2)) {
-		    scanfunc(hn, scanflags);
-		    match++;
-		}
-	    }
-	}
-    return match;
-    }
-
-    /* flags in flags1 must be set,    *
-     * flags in flags2 must NOT be set */
-    for (i = 0; i < ht->hsize; i++) {
-	for (hn = ht->nodes[i]; hn; hn = hn->next) {
-	    if (domatch(hn->nam, com, 0) && (hn->flags & flags1) && !(hn->flags & flags2)) {
+    for (i = 0; i < hsize; i++)
+	for (st.u.u = nodes[i]; st.u.u; ) {
+	    HashNode hn = st.u.u;
+	    st.u.u = st.u.u->next;
+	    if ((hn->flags & flags1) + !flags1 && !(hn->flags & flags2) &&
+		domatch(hn->nam, com, 0))
 		scanfunc(hn, scanflags);
 		match++;
-	    }
 	}
-    }
+
+    ht->scan = NULL;
+
     return match;
 }
 
-
 /* Expand hash tables when they get too many entries. *
  * The new size is 4 times the previous size.         */
 
 /**/
-void
+static void
 expandhashtable(HashTable ht)
 {
     struct hashnode **onodes, **ha, *hn, *hp;
@@ -420,8 +458,8 @@ expandhashtable(HashTable ht)
 /* Empty the hash table and resize it if necessary */
 
 /**/
-void
-emptyhashtable(HashTable ht, int newsize)
+static void
+resizehashtable(HashTable ht, int newsize)
 {
     struct hashnode **ha, *hn, *hp;
     int i;
@@ -450,14 +488,23 @@ emptyhashtable(HashTable ht, int newsize)
     ht->ct = 0;
 }
 
-/* Print info about hash table */
+/* Generic method to empty a hash table */
+
+/**/
+void
+emptyhashtable(HashTable ht)
+{
+    resizehashtable(ht, ht->hsize);
+}
 
 #ifdef ZSH_HASH_DEBUG
 
+/* Print info about hash table */
+
 #define MAXDEPTH 7
 
 /**/
-void
+static void
 printhashtabinfo(HashTable ht)
 {
     HashNode hn;
@@ -488,22 +535,33 @@ printhashtabinfo(HashTable ht)
     printf("number of hash values with chain of length %d+ : %4d\n", MAXDEPTH, chainlen[MAXDEPTH]);
     printf("total number of nodes                         : %4d\n", total);
 }
-#endif
+
+/**/
+int
+bin_hashinfo(char *nam, char **args, char *ops, int func)
+{
+    HashTable ht;
+    printf("----------------------------------------------------\n");
+    for(ht = firstht; ht; ht = ht->next) {
+	ht->printinfo(ht);
+	printf("----------------------------------------------------\n");
+    }
+    return 0;
+}
+
+#endif /* ZSH_HASH_DEBUG */
 
 /********************************/
 /* Command Hash Table Functions */
 /********************************/
 
-/* size of the initial cmdnamtab hash table */
-#define INITIAL_CMDNAMTAB 201
-
 /* Create a new command hash table */
  
 /**/
 void
 createcmdnamtable(void)
 {
-    cmdnamtab = newhashtable(INITIAL_CMDNAMTAB);
+    cmdnamtab = newhashtable(201, "cmdnamtab", NULL);
 
     cmdnamtab->hash        = hasher;
     cmdnamtab->emptytable  = emptycmdnamtable;
@@ -516,19 +574,15 @@ createcmdnamtable(void)
     cmdnamtab->enablenode  = NULL;
     cmdnamtab->freenode    = freecmdnamnode;
     cmdnamtab->printnode   = printcmdnamnode;
-#ifdef ZSH_HASH_DEBUG
-    cmdnamtab->printinfo   = printhashtabinfo;
-    cmdnamtab->tablename   = ztrdup("cmdnamtab");
-#endif
 
     pathchecked = path;
 }
 
 /**/
-void
+static void
 emptycmdnamtable(HashTable ht)
 {
-    emptyhashtable(ht, INITIAL_CMDNAMTAB);
+    emptyhashtable(ht);
     pathchecked = path;
 }
 
@@ -546,12 +600,7 @@ hashdir(char **dirp)
     if (isrelative(*dirp) || !(dir = opendir(unmeta(*dirp))))
 	return;
 
-    while ((fn = zreaddir(dir))) {
-	/* Ignore `.' and `..'. */
-	if (fn[0] == '.' &&
-	    (fn[1] == '\0' ||
-	     (fn[1] == '.' && fn[2] == '\0')))
-	    continue;
+    while ((fn = zreaddir(dir, 1))) {
 	if (!cmdnamtab->getnode(cmdnamtab, fn)) {
 	    cn = (Cmdnam) zcalloc(sizeof *cn);
 	    cn->flags = 0;
@@ -566,7 +615,7 @@ hashdir(char **dirp)
  * the command hashtable.                       */
 
 /**/
-void
+static void
 fillcmdnamtable(HashTable ht)
 {
     char **pq;
@@ -578,7 +627,7 @@ fillcmdnamtable(HashTable ht)
 }
 
 /**/
-void
+static void
 freecmdnamnode(HashNode hn)
 {
     Cmdnam cn = (Cmdnam) hn;
@@ -593,7 +642,7 @@ freecmdnamnode(HashNode hn)
 /* Print an element of the cmdnamtab hash table (external command) */
  
 /**/
-void
+static void
 printcmdnamnode(HashNode hn, int printflags)
 {
     Cmdnam cn = (Cmdnam) hn;
@@ -651,7 +700,7 @@ printcmdnamnode(HashNode hn, int printflags)
 void
 createshfunctable(void)
 {
-    shfunctab = newhashtable(7);
+    shfunctab = newhashtable(7, "shfunctab", NULL);
 
     shfunctab->hash        = hasher;
     shfunctab->emptytable  = NULL;
@@ -664,10 +713,6 @@ createshfunctable(void)
     shfunctab->enablenode  = enableshfuncnode;
     shfunctab->freenode    = freeshfuncnode;
     shfunctab->printnode   = printshfuncnode;
-#ifdef ZSH_HASH_DEBUG
-    shfunctab->printinfo   = printhashtabinfo;
-    shfunctab->tablename   = ztrdup("shfunctab");
-#endif
 }
 
 /* Remove an entry from the shell function hash table.   *
@@ -675,7 +720,7 @@ createshfunctable(void)
  * it will disable the trapping of that signal.          */
 
 /**/
-HashNode
+static HashNode
 removeshfuncnode(HashTable ht, char *nam)
 {
     HashNode hn;
@@ -693,7 +738,7 @@ removeshfuncnode(HashTable ht, char *nam)
  * it will disable the trapping of that signal.          */
 
 /**/
-void
+static void
 disableshfuncnode(HashNode hn, int flags)
 {
     hn->flags |= DISABLED;
@@ -710,7 +755,7 @@ disableshfuncnode(HashNode hn, int flags)
  * it will re-enable the trapping of that signal.        */
 
 /**/
-void
+static void
 enableshfuncnode(HashNode hn, int flags)
 {
     Shfunc shf = (Shfunc) hn;
@@ -727,7 +772,7 @@ enableshfuncnode(HashNode hn, int flags)
 }
 
 /**/
-void
+static void
 freeshfuncnode(HashNode hn)
 {
     Shfunc shf = (Shfunc) hn;
@@ -741,7 +786,7 @@ freeshfuncnode(HashNode hn)
 /* Print a shell function */
  
 /**/
-void
+static void
 printshfuncnode(HashNode hn, int printflags)
 {
     Shfunc f = (Shfunc) hn;
@@ -766,7 +811,7 @@ printshfuncnode(HashNode hn, int printflags)
 	printf("undefined ");
     if (f->flags & PM_TAGGED)
 	printf("traced ");
-    if (!f->funcdef) {
+    if ((f->flags & PM_UNDEFINED) || !f->funcdef) {
 	nicezputs(f->nam, stdout);
 	printf(" () { }\n");
 	return;
@@ -790,7 +835,7 @@ createbuiltintable(void)
 {
     Builtin bn;
 
-    builtintab = newhashtable(85);
+    builtintab = newhashtable(85, "builtintab", NULL);
 
     builtintab->hash        = hasher;
     builtintab->emptytable  = NULL;
@@ -798,15 +843,11 @@ createbuiltintable(void)
     builtintab->addnode     = addhashnode;
     builtintab->getnode     = gethashnode;
     builtintab->getnode2    = gethashnode2;
-    builtintab->removenode  = NULL;
+    builtintab->removenode  = removehashnode;
     builtintab->disablenode = disablehashnode;
     builtintab->enablenode  = enablehashnode;
-    builtintab->freenode    = NULL;
+    builtintab->freenode    = freebuiltinnode;
     builtintab->printnode   = printbuiltinnode;
-#ifdef ZSH_HASH_DEBUG
-    builtintab->printinfo   = printhashtabinfo;
-    builtintab->tablename   = ztrdup("builtintab");
-#endif
 
     for (bn = builtins; bn->nam; bn++)
 	builtintab->addnode(builtintab, bn->nam, bn);
@@ -815,7 +856,7 @@ createbuiltintable(void)
 /* Print a builtin */
 
 /**/
-void
+static void
 printbuiltinnode(HashNode hn, int printflags)
 {
     Builtin bn = (Builtin) hn;
@@ -834,6 +875,20 @@ printbuiltinnode(HashNode hn, int printflags)
     printf("%s\n", bn->nam);
 }
 
+/**/
+static void
+freebuiltinnode(HashNode hn)
+{
+    Builtin bn = (Builtin) hn;
+ 
+    zsfree(bn->nam);
+    zsfree(bn->optstr);
+    zsfree(bn->defopts);
+    zfree(bn, sizeof(struct builtin));
+}
+
+
+
 /**************************************/
 /* Reserved Word Hash Table Functions */
 /**************************************/
@@ -846,7 +901,7 @@ createreswdtable(void)
 {
     Reswd rw;
 
-    reswdtab = newhashtable(23);
+    reswdtab = newhashtable(23, "reswdtab", NULL);
 
     reswdtab->hash        = hasher;
     reswdtab->emptytable  = NULL;
@@ -859,10 +914,6 @@ createreswdtable(void)
     reswdtab->enablenode  = enablehashnode;
     reswdtab->freenode    = NULL;
     reswdtab->printnode   = printreswdnode;
-#ifdef ZSH_HASH_DEBUG
-    reswdtab->printinfo   = printhashtabinfo;
-    reswdtab->tablename   = ztrdup("reswdtab");
-#endif
 
     for (rw = reswds; rw->nam; rw++)
 	reswdtab->addnode(reswdtab, rw->nam, rw);
@@ -871,7 +922,7 @@ createreswdtable(void)
 /* Print a reserved word */
 
 /**/
-void
+static void
 printreswdnode(HashNode hn, int printflags)
 {
     Reswd rw = (Reswd) hn;
@@ -900,7 +951,7 @@ printreswdnode(HashNode hn, int printflags)
 void
 createaliastable(void)
 {
-    aliastab = newhashtable(23);
+    aliastab = newhashtable(23, "aliastab", NULL);
 
     aliastab->hash        = hasher;
     aliastab->emptytable  = NULL;
@@ -913,10 +964,6 @@ createaliastable(void)
     aliastab->enablenode  = enablehashnode;
     aliastab->freenode    = freealiasnode;
     aliastab->printnode   = printaliasnode;
-#ifdef ZSH_HASH_DEBUG
-    aliastab->printinfo   = printhashtabinfo;
-    aliastab->tablename   = ztrdup("aliastab");
-#endif
 
     /* add the default aliases */
     aliastab->addnode(aliastab, ztrdup("run-help"), createaliasnode(ztrdup("man"), 0));
@@ -939,7 +986,7 @@ createaliasnode(char *txt, int flags)
 }
 
 /**/
-void
+static void
 freealiasnode(HashNode hn)
 {
     Alias al = (Alias) hn;
@@ -952,7 +999,7 @@ freealiasnode(HashNode hn)
 /* Print an alias */
 
 /**/
-void
+static void
 printaliasnode(HashNode hn, int printflags)
 {
     Alias a = (Alias) hn;
@@ -1099,12 +1146,9 @@ printparamnode(HashNode hn, int printflags)
 /* Named Directory Hash Table Functions */
 /****************************************/
 
-/* size of the initial name directory hash table */
-#define INITIAL_NAMEDDIR 201
-
 /* != 0 if all the usernames have already been *
  * added to the named directory hash table.    */
-int allusersadded;
+static int allusersadded;
 
 /* Create new hash table for named directories */
 
@@ -1112,7 +1156,7 @@ int allusersadded;
 void
 createnameddirtable(void)
 {
-    nameddirtab = newhashtable(INITIAL_NAMEDDIR);
+    nameddirtab = newhashtable(201, "nameddirtab", NULL);
 
     nameddirtab->hash        = hasher;
     nameddirtab->emptytable  = emptynameddirtable;
@@ -1125,10 +1169,6 @@ createnameddirtable(void)
     nameddirtab->enablenode  = NULL;
     nameddirtab->freenode    = freenameddirnode;
     nameddirtab->printnode   = printnameddirnode;
-#ifdef ZSH_HASH_DEBUG
-    nameddirtab->printinfo   = printhashtabinfo;
-    nameddirtab->tablename   = ztrdup("nameddirtab");
-#endif
 
     allusersadded = 0;
     finddir(NULL);		/* clear the finddir cache */
@@ -1137,10 +1177,10 @@ createnameddirtable(void)
 /* Empty the named directories table */
 
 /**/
-void
+static void
 emptynameddirtable(HashTable ht)
 {
-    emptyhashtable(ht, INITIAL_NAMEDDIR);
+    emptyhashtable(ht);
     allusersadded = 0;
     finddir(NULL);		/* clear the finddir cache */
 }
@@ -1149,7 +1189,7 @@ emptynameddirtable(HashTable ht)
  * to the named directories table.                     */
 
 /**/
-void
+static void
 fillnameddirtable(HashTable ht)
 {
     if (!allusersadded) {
@@ -1173,7 +1213,7 @@ fillnameddirtable(HashTable ht)
  * initialising the `diff' member.          */
 
 /**/
-void
+static void
 addnameddirnode(HashTable ht, char *nam, void *nodeptr)
 {
     Nameddir nd = (Nameddir) nodeptr;
@@ -1187,7 +1227,7 @@ addnameddirnode(HashTable ht, char *nam, void *nodeptr)
  * hash table, clearing the finddir() cache. */
 
 /**/
-HashNode
+static HashNode
 removenameddirnode(HashTable ht, char *nam)
 {
     HashNode hn = removehashnode(ht, nam);
@@ -1200,7 +1240,7 @@ removenameddirnode(HashTable ht, char *nam)
 /* Free up the memory used by a named directory hash node. */
 
 /**/
-void
+static void
 freenameddirnode(HashNode hn)
 {
     Nameddir nd = (Nameddir) hn;
@@ -1213,7 +1253,7 @@ freenameddirnode(HashNode hn)
 /* Print a named directory */
 
 /**/
-void
+static void
 printnameddirnode(HashNode hn, int printflags)
 {
     Nameddir nd = (Nameddir) hn;
@@ -1229,158 +1269,3 @@ printnameddirnode(HashNode hn, int printflags)
     quotedzputs(nd->dir, stdout);
     putchar('\n');
 }
-
-/********************************/
-/* Compctl Hash Table Functions */
-/********************************/
-
-/**/
-void
-createcompctltable(void)
-{
-    compctltab = newhashtable(23);
-
-    compctltab->hash        = hasher;
-    compctltab->emptytable  = NULL;
-    compctltab->filltable   = NULL;
-    compctltab->addnode     = addhashnode;
-    compctltab->getnode     = gethashnode2;
-    compctltab->getnode2    = gethashnode2;
-    compctltab->removenode  = removehashnode;
-    compctltab->disablenode = NULL;
-    compctltab->enablenode  = NULL;
-    compctltab->freenode    = freecompctlp;
-    compctltab->printnode   = printcompctlp;
-#ifdef ZSH_HASH_DEBUG
-    compctltab->printinfo   = printhashtabinfo;
-    compctltab->tablename   = ztrdup("compctltab");
-#endif
-}
-
-/**/
-void
-freecompctl(Compctl cc)
-{
-    if (cc == &cc_default ||
- 	cc == &cc_first ||
-	cc == &cc_compos ||
-	--cc->refc > 0)
-	return;
-
-    zsfree(cc->keyvar);
-    zsfree(cc->glob);
-    zsfree(cc->str);
-    zsfree(cc->func);
-    zsfree(cc->explain);
-    zsfree(cc->prefix);
-    zsfree(cc->suffix);
-    zsfree(cc->hpat);
-    zsfree(cc->subcmd);
-    if (cc->cond)
-	freecompcond(cc->cond);
-    if (cc->ext) {
-	Compctl n, m;
-
-	n = cc->ext;
-	do {
-	    m = (Compctl) (n->next);
-	    freecompctl(n);
-	    n = m;
-	}
-	while (n);
-    }
-    if (cc->xor && cc->xor != &cc_default)
-	freecompctl(cc->xor);
-    zfree(cc, sizeof(struct compctl));
-}
-
-/**/
-void
-freecompctlp(HashNode hn)
-{
-    Compctlp ccp = (Compctlp) hn;
-
-    zsfree(ccp->nam);
-    freecompctl(ccp->cc);
-    zfree(ccp, sizeof(struct compctlp));
-}
-
-/***********************************************************/
-/* Emacs Multi-Character Key Bindings Hash Table Functions */
-/***********************************************************/
-
-/* size of the initial hashtable for multi-character *
- * emacs key bindings.                               */
-#define INITIAL_EMKEYBINDTAB 67
-
-/**/
-void
-createemkeybindtable(void)
-{
-    emkeybindtab = newhashtable(INITIAL_EMKEYBINDTAB);
-
-    emkeybindtab->hash        = hasher;
-    emkeybindtab->emptytable  = emptyemkeybindtable;
-    emkeybindtab->filltable   = NULL;
-    emkeybindtab->addnode     = addhashnode;
-    emkeybindtab->getnode     = gethashnode2;
-    emkeybindtab->getnode2    = gethashnode2;
-    emkeybindtab->removenode  = removehashnode;
-    emkeybindtab->disablenode = NULL;
-    emkeybindtab->enablenode  = NULL;
-    emkeybindtab->freenode    = freekeynode;
-
-    /* need to combine printbinding and printfuncbinding for this */
-    emkeybindtab->printnode   = NULL;
-#ifdef ZSH_HASH_DEBUG
-    emkeybindtab->printinfo   = printhashtabinfo;
-    emkeybindtab->tablename   = ztrdup("emkeybindtab");
-#endif
-}
-
-/**/
-void
-emptyemkeybindtable(HashTable ht)
-{
-    emptyhashtable(ht, INITIAL_EMKEYBINDTAB);
-}
-
-/********************************************************/
-/* Vi Multi-Character Key Bindings Hash Table Functions */
-/********************************************************/
-
-/* size of the initial hash table for *
- * multi-character vi key bindings.   */
-#define INITIAL_VIKEYBINDTAB 20
-
-/**/
-void
-createvikeybindtable(void)
-{
-    vikeybindtab = newhashtable(INITIAL_VIKEYBINDTAB);
-
-    vikeybindtab->hash        = hasher;
-    vikeybindtab->emptytable  = emptyvikeybindtable;
-    vikeybindtab->filltable   = NULL;
-    vikeybindtab->addnode     = addhashnode;
-    vikeybindtab->getnode     = gethashnode2;
-    vikeybindtab->getnode2    = gethashnode2;
-    vikeybindtab->removenode  = removehashnode;
-    vikeybindtab->disablenode = NULL;
-    vikeybindtab->enablenode  = NULL;
-    vikeybindtab->freenode    = freekeynode;
-
-    /* need to combine printbinding and printfuncbinding for this */
-    vikeybindtab->printnode   = NULL;
-#ifdef ZSH_HASH_DEBUG
-    vikeybindtab->printinfo   = printhashtabinfo;
-    vikeybindtab->tablename   = ztrdup("vikeybindtab");
-#endif
-}
-
-/**/
-void
-emptyvikeybindtable(HashTable ht)
-{
-    emptyhashtable(ht, INITIAL_VIKEYBINDTAB);
-}
