@@ -1,9 +1,11 @@
 /*
+ * $Id: hist.c,v 3.1.2.7 1997/06/02 06:52:04 hzoli Exp $
+ *
  * hist.c - history expansion
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -28,14 +30,15 @@
  */
 
 #include "zsh.h"
+#include "hist.pro"
 
 /*
  * Note on curhist: with history inactive, this points to the
  * last line actually added to the history list.  With history active,
  * the line does not get added to the list until hend(), if at all.
  * However, curhist is incremented to reflect the current line anyway.
- * If the resulting line was not added to the list, a flag is set so
- * that curhist will be decremented in hbegin().
+ * Thus if the line is not added to the list, curhist must be
+ * decremented in hend().
  */
 
 /* Bits of histactive variable */
@@ -152,7 +155,7 @@ hgetc(void)
 }
 
 /**/
-void
+static void
 safeinungetc(int c)
 {
     if (lexstop)
@@ -161,18 +164,10 @@ safeinungetc(int c)
 	inungetc(c);
 }
 
-/**/
-void
-herrflush(void)
-{
-    while (!lexstop && inbufct && !strin)
-	hwaddc(ingetc());
-}
-
 /* extract :s/foo/bar/ delimiters and arguments */
 
 /**/
-int
+static int
 getsubsargs(char *subline)
 {
     int del;
@@ -190,8 +185,8 @@ getsubsargs(char *subline)
     zsfree(hsubr);
     hsubr = ptr2;
     if (hsubl && !strstr(subline, hsubl)) {
-	herrflush();
 	zerr("substitution failed", NULL, 0);
+	inerrflush();
 	return 1;
     }
     return 0;
@@ -200,7 +195,7 @@ getsubsargs(char *subline)
 /* Get the maximum no. of words for a history entry. */
 
 /**/
-int
+static int
 getargc(Histent ehist)
 {
     return ehist->nwords ? ehist->nwords-1 : 0;
@@ -209,7 +204,7 @@ getargc(Histent ehist)
 /* Perform history substitution, returning the next character afterwards. */
 
 /**/
-int
+static int
 histsubchar(int c)
 {
     int ev, farg, evset = -1, larg, argc, cflag = 0, bflag = 0;
@@ -276,7 +271,7 @@ histsubchar(int c)
 	    mev = ev = hconsearch(hsubl = ztrdup(buf), &marg);
 	    evset = 0;
 	    if (ev == -1) {
-		herrflush();
+		inerrflush();
 		zerr("no such event: %s", buf, 0);
 		return -1;
 	    }
@@ -302,7 +297,7 @@ histsubchar(int c)
 		c = ingetc();
 	    }
 	    *ptr = 0;
-	    if (!*buf) {
+	    if (!*buf)
 		if (c != '%') {
 		    if (isset(CSHJUNKIEHISTORY))
 			ev = curhist - 1;
@@ -318,7 +313,6 @@ histsubchar(int c)
 		    else
 			ev = defev;
 		    evset = 0;
-		}
 	    } else if ((t0 = atoi(buf))) {
 		ev = (t0 < 0) ? curhist + t0 : t0;
 		evset = 1;
@@ -329,8 +323,9 @@ histsubchar(int c)
 		ev = curhist;
 		evset = 1;
 	    } else if ((ev = hcomsearch(buf)) == -1) {
-		herrflush();
 		zerr("event not found: %s", buf, 0);
+		while (c != '\n' && !lexstop)
+		    c = ingetc();
 		return -1;
 	    } else
 		evset = 1;
@@ -352,8 +347,9 @@ histsubchar(int c)
 		    ehist = gethist(defev = mev);
 		    argc = getargc(ehist);
 		} else {
-		    herrflush();
 		    zerr("Ambiguous history reference", NULL, 0);
+		    while (c != '\n' && !lexstop)
+			c = ingetc();
 		    return -1;
 		}
 
@@ -410,28 +406,28 @@ histsubchar(int c)
 		break;
 	    case 'h':
 		if (!remtpath(&sline)) {
-		    herrflush();
+		    inerrflush();
 		    zerr("modifier failed: h", NULL, 0);
 		    return -1;
 		}
 		break;
 	    case 'e':
 		if (!rembutext(&sline)) {
-		    herrflush();
+		    inerrflush();
 		    zerr("modifier failed: e", NULL, 0);
 		    return -1;
 		}
 		break;
 	    case 'r':
 		if (!remtext(&sline)) {
-		    herrflush();
+		    inerrflush();
 		    zerr("modifier failed: r", NULL, 0);
 		    return -1;
 		}
 		break;
 	    case 't':
 		if (!remlpaths(&sline)) {
-		    herrflush();
+		    inerrflush();
 		    zerr("modifier failed: t", NULL, 0);
 		    return -1;
 		}
@@ -443,7 +439,7 @@ histsubchar(int c)
 		if (hsubl && hsubr)
 		    subst(&sline, hsubl, hsubr, gbal);
 		else {
-		    herrflush();
+		    inerrflush();
 		    zerr("no previous substitution", NULL, 0);
 		    return -1;
 		}
@@ -461,7 +457,7 @@ histsubchar(int c)
 		upcase(&sline);
 		break;
 	    default:
-		herrflush();
+		inerrflush();
 		zerr("illegal modifier: %c", NULL, c);
 		return -1;
 	    }
@@ -569,8 +565,6 @@ hbegin(void)
     chwords = zalloc((chwordlen = 16)*sizeof(short));
     chwordpos = 0;
 
-    if (histactive & HA_JUNKED)
-	curhist--;
     curhistent = gethistent(curhist);
     if (!curhistent->ftim)
 	curhistent->ftim = time(NULL);
@@ -586,36 +580,17 @@ hbegin(void)
 /* compare current line with history entry using only text in words */
 
 /**/
-int
+static int
 histcmp(Histent he)
 {
     int kword, lword;
     int nwords = chwordpos/2;
 
-    /* If the history entry came from a file, the words were not divided by
-     * the lexer so we have to resort to a whitespace-ignoring compare.
+    /* If the history entry came from a file, the words were not
+     * divided by the lexer so we have to resort to strcmp.
      */
-    if (he->flags & HIST_READ) {
-	char *str1 = he->text, *str2 = chline;
-
-	while (inblank(*str1)) str1++;
-	while (inblank(*str2)) str2++;
-	while (*str1 && *str2) {
-	    if (inblank(*str1)) {
-		if (!inblank(*str2))
-		    break;
-		do str1++; while (inblank(*str1));
-		do str2++; while (inblank(*str2));
-	    }
-	    else {
-		if (*str1 != *str2)
-		    break;
-		str1++;
-		str2++;
-	    }
-	}
-	return *str1 - *str2;
-    }
+    if (he->flags & HIST_READ)
+	return strcmp(he->text, chline);
 
     if (nwords != he->nwords)
 	return 1;
@@ -664,8 +639,6 @@ hend(void)
     int flag, save = 1;
 
     DPUTS(!chline, "BUG: chline is NULL in hend()");
-    if (histdone & HISTFLAG_SETTY)
-	settyinfo(&shttyinfo);
     if (histactive & (HA_NOSTORE|HA_NOINC)) {
 	zfree(chline, hlinesz);
 	zfree(chwords, chwordlen*sizeof(short));
@@ -681,12 +654,11 @@ hend(void)
 	save = 0;
     else {
 	*hptr = '\0';
-	if (hptr[-1] == '\n') {
+	if (hptr[-1] == '\n')
 	    if (chline[1]) {
 		*--hptr = '\0';
 	    } else
 		save = 0;
-	}
 	if (!*chline || !strcmp(chline, "\n") ||
 	    (isset(HISTIGNORESPACE) && spaceflag))
 	    save = 0;
@@ -710,6 +682,7 @@ hend(void)
     }
     if (save) {
 	Histent he;
+	int keepflags = 0;
 
 #ifdef DEBUG
 	/* debugging only */
@@ -719,41 +692,35 @@ hend(void)
 	}
 #endif
 	/* get rid of pesky \n which we've already nulled out */
-	if (chwordpos > 1 && !chline[chwords[chwordpos-2]])
+	if (!chline[chwords[chwordpos-2]])
 	    chwordpos -= 2;
 	/* strip superfluous blanks, if desired */
 	if (isset(HISTREDUCEBLANKS))
 	    histreduceblanks();
+
 	if (isset(HISTIGNOREDUPS) && (he = gethistent(curhist - 1))
-	 && he->text && histcmp(he) == 0) {
+	 && he->text && !histcmp(he)) {
 	    /* This history entry compares the same as the previous.
 	     * In case minor changes were made, we overwrite the
-	     * previous one with the current one.  This also gets the
-	     * timestamp right.  Perhaps, preserve the HIST_OLD flag.
+	     * previous one with the current one.  This also gets
+	     * the timestamp right.  However, keep the old flags.
 	     */
-	    zsfree(he->text);
-	    he->text = ztrdup(chline);
-	    if (chwordpos)
-		memcpy(he->words, chwords, chwordpos * sizeof(short));
-	    he->stim = time(NULL);	/* set start time */
-	    he->ftim = 0;
+	    keepflags = he->flags;
 	    curhist--;
 	}
-	else {
-	    Histent curhistent = gethistent(curhist);
-	    zsfree(curhistent->text);
-	    if (curhistent->nwords)
-		zfree(curhistent->words, curhistent->nwords*2*sizeof(short));
 
-	    curhistent->text = ztrdup(chline);
-	    curhistent->stim = time(NULL);
-	    curhistent->ftim = 0L;
-	    curhistent->flags = 0;
+	he =  gethistent(curhist);
+	zsfree(he->text);
+	he->text = ztrdup(chline);
+	if (he->nwords)
+	    zfree(he->words, he->nwords*2*sizeof(short));
+	he->stim = time(NULL);
+	he->ftim = 0L;
+	he->flags = keepflags;
 
-	    if ((curhistent->nwords = chwordpos/2)) {
-		curhistent->words = (short *)zalloc(chwordpos * sizeof(short));
-		memcpy(curhistent->words, chwords, chwordpos * sizeof(short));
-	    }
+	if ((he->nwords = chwordpos/2)) {
+	    he->words = (short *)zalloc(chwordpos * sizeof(short));
+	    memcpy(he->words, chwords, chwordpos * sizeof(short));
 	}
     } else
 	curhist--;
@@ -777,7 +744,7 @@ remhist(void)
 	    zsfree(he->text);
 	    he->text = NULL;
 	    histactive |= HA_JUNKED;
-	    /* curhist-- is delayed until the next hbegin() */
+	    curhist--;
 	}
     } else
 	histactive |= HA_NOSTORE;
@@ -844,7 +811,7 @@ histbackword(void)
 /* Get the start and end point of the current history word */
 
 /**/
-void
+static void
 hwget(char **startptr)
 {
     int pos = hwgetword > -1 ? hwgetword : chwordpos - 2;
@@ -917,7 +884,7 @@ hgetline(void)
 /* get an argument specification */
 
 /**/
-int
+static int
 getargspec(int argc, int marg, int evset)
 {
     int c, ret = -1;
@@ -937,12 +904,12 @@ getargspec(int argc, int marg, int evset)
 	ret = argc;
     else if (c == '%') {
 	if (evset) {
-	    herrflush();
+	    inerrflush();
 	    zerr("Ambiguous history reference", NULL, 0);
 	    return -2;
 	}
 	if (marg == -1) {
-	    herrflush();
+	    inerrflush();
 	    zerr("%% with no previous word matched", NULL, 0);
 	    return -2;
 	}
@@ -955,7 +922,7 @@ getargspec(int argc, int marg, int evset)
 /* do ?foo? search */
 
 /**/
-int
+static int
 hconsearch(char *str, int *marg)
 {
     int t0, t1 = 0;
@@ -1111,7 +1078,7 @@ subst(char **strptr, char *in, char *out, int gbal)
 }
 
 /**/
-char *
+static char *
 convamps(char *out, char *in, int inlen)
 {
     char *ptr, *ret, *pp;
@@ -1168,21 +1135,21 @@ quietgetevent(int ev)
 }
 
 /**/
-Histent
+static Histent
 gethist(int ev)
 {
     Histent ret;
 
     ret = quietgethist(ev);
     if (!ret) {
-	herrflush();
+	inerrflush();
 	zerr("no such event: %d", NULL, ev);
     }
     return ret;
 }
 
 /**/
-char *
+static char *
 getargs(Histent elist, int arg1, int arg2)
 {
     short *words = elist->words;
@@ -1190,7 +1157,7 @@ getargs(Histent elist, int arg1, int arg2)
 
     if (arg2 < arg1 || arg1 >= nwords || arg2 >= nwords) {
 	/* remember, argN is indexed from 0, nwords is total no. of words */
-	herrflush();
+	inerrflush();
 	zerr("no such word in event", NULL, 0);
 	return NULL;
     }
@@ -1262,7 +1229,7 @@ quote(char **tr)
 }
 
 /**/
-int
+static int
 quotebreak(char **tr)
 {
     char *ptr, *rptr, **str = (char **)tr;
@@ -1294,9 +1261,10 @@ quotebreak(char **tr)
     return 0;
 }
 
+#if 0
 /* read an arbitrary amount of data into a buffer until stop is found */
 
-#if 0 /**/
+/**/
 char *
 hdynread(int stop)
 {
@@ -1325,7 +1293,7 @@ hdynread(int stop)
 #endif
 
 /**/
-char *
+static char *
 hdynread2(int stop)
 {
     int bsiz = 256, ct = 0, c;
@@ -1453,9 +1421,6 @@ readhistfile(char *s, int err)
 		ent->ftim = ent->stim = tim;
 	    }
 
-	    if (ent->ftim < ent->stim)	/* 3.1.6 history file compatibility */
-		ent->ftim += ent->stim;
-
 	    zsfree(ent->text);
 	    ent->text = ztrdup(pt);
 	    ent->flags = HIST_OLD|HIST_READ;
@@ -1463,19 +1428,19 @@ readhistfile(char *s, int err)
 		zfree(ent->words, ent->nwords*2*sizeof(short));
 
 	    /* Divide up the words.  We don't know how it lexes,
-	       so just look for white-space.
+	       so just look for spaces.
 	       */
 	    nwordpos = 0;
 	    start = pt;
 	    do {
-		while (inblank(*pt))
+		while (*pt == ' ')
 		    pt++;
 		if (*pt) {
 		    if (nwordpos >= nwordlist)
 			wordlist = (short *) realloc(wordlist,
 					(nwordlist += 16)*sizeof(short));
 		    wordlist[nwordpos++] = pt - start;
-		    while (*pt && !inblank(*pt))
+		    while (*pt && *pt != ' ')
 			pt++;
 		    wordlist[nwordpos++] = pt - start;
 		}
