@@ -1,9 +1,11 @@
 /*
+ * $Id: init.c,v 3.1.2.13 1997/06/01 06:13:15 hzoli Exp $
+ *
  * init.c - main loop and initialization routines
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,73 +32,17 @@
 #define GLOBALS
 #include "zsh.h"
 
-static int noexitct = 0;
+#include "zshpaths.h"
+#include "zshxmods.h"
+#include "init.pro"
 
-/**/
-int
-main(int argc, char **argv)
-{
-    char **t;
-#ifdef USE_LOCALE
-    setlocale(LC_ALL, "");
-#endif
-
-    global_permalloc();
-
-    for (t = argv; *t; *t = metafy(*t, -1, META_ALLOC), t++);
-
-    if (!(zsh_name = strrchr(argv[0], '/')))
-	zsh_name = argv[0];
-    else
-	zsh_name++;
-    if (*zsh_name == '-')
-	zsh_name++;
-
-    fdtable_size = OPEN_MAX;
-    fdtable = zcalloc(fdtable_size);
-
-    emulate(zsh_name, 1);   /* initialises most options */
-    opts[LOGINSHELL] = (**argv == '-');
-    opts[MONITOR] = 1;   /* may be unset in init_io() */
-    opts[PRIVILEGED] = (getuid() != geteuid() || getgid() != getegid());
-    opts[USEZLE] = 1;   /* may be unset in init_io() */
-    parseargs(argv);   /* sets INTERACTIVE, SHINSTDIN and SINGLECOMMAND */
-
-    SHTTY = -1;
-    init_io();
-    setupvals();
-    init_signals();
-    global_heapalloc();
-    run_init_scripts();
-    init_misc();
-
-    for (;;) {
-	do
-	    loop(1,0);
-	while (tok != ENDINPUT);
-	if (!(isset(IGNOREEOF) && interact)) {
-#if 0
-	    if (interact)
-		fputs(islogin ? "logout\n" : "exit\n", shout);
-#endif
-	    zexit(lastval, 0);
-	    continue;
-	}
-	noexitct++;
-	if (noexitct >= 10) {
-	    stopmsg = 1;
-	    zexit(lastval, 0);
-	}
-	zerrnam("zsh", (!islogin) ? "use 'exit' to exit."
-		: "use 'logout' to logout.", NULL, 0);
-    }
-}
+int noexitct = 0;
 
 /* keep executing lists until EOF found */
 
 /**/
 void
-loop(int toplevel, int justonce)
+loop(int toplevel)
 {
     List list;
 #ifdef DEBUG
@@ -107,37 +53,20 @@ loop(int toplevel, int justonce)
     for (;;) {
 	freeheap();
 	errflag = 0;
-	if (isset(SHINSTDIN)) {
-	    setblock_stdin();
-	    if (interact)
-		preprompt();
-	}
+	if (interact && isset(SHINSTDIN))
+	    preprompt();
 	hbegin();		/* init history mech        */
 	intr();			/* interrupts on            */
 	lexinit();              /* initialize lexical state */
 	if (!(list = parse_event())) {	/* if we couldn't parse a list */
 	    hend();
-	    if ((tok == ENDINPUT && !errflag) || justonce)
+	    if (tok == ENDINPUT && !errflag)
 		break;
 	    continue;
 	}
 	if (hend()) {
 	    int toksav = tok;
-	    List prelist;
 
-	    if (toplevel && (prelist = getshfunc("preexec")) != &dummy_list) {
-		Histent he = gethistent(curhist);
-		LinkList args;
-		PERMALLOC {
-		    args = newlinklist();
-		    addlinknode(args, "preexec");
-		    if (he && he->text)
-			addlinknode(args, he->text);
-		} LASTALLOC;
-		doshfunc(prelist, args, 0, 1);
-		freelinklist(args, (FreeFunc) NULL);
-		errflag = 0;
-	    }
 	    if (stopmsg)	/* unset 'you have stopped jobs' flag */
 		stopmsg--;
 	    execlist(list, 0, 0);
@@ -163,39 +92,12 @@ loop(int toplevel, int justonce)
 		dotrap(SIGEXIT);
 	    exit(lastval);
 	}
-	if (justonce)
-	    break;
     }
     popheap();
 }
 
-/**/
-void
-emulate(const char *zsh_name, int fully)
-{
-    int optno;
-
-    /* Work out the new emulation mode */
-    if(!strcmp(zsh_name, "csh"))
-	emulation = EMULATE_CSH;
-    else if(!strcmp(zsh_name, "ksh"))
-	emulation = EMULATE_KSH;
-    else if(!strcmp(zsh_name, "sh"))
-	emulation = EMULATE_SH;
-    else
-	emulation = EMULATE_ZSH;
-
-    /* Set options: each non-special option is set according to the *
-     * current emulation mode if either it is considered relevant   *
-     * to emulation or we are doing a full emulation (as indicated  *
-     * by the `fully' parameter).                                   */
-    for(optno = OPT_SIZE; --optno; )
-	if((fully && !(optns[optno].flags & OPT_SPECIAL)) ||
-	    	(optns[optno].flags & OPT_EMULATE))
-	    opts[optno] = defset(optno);
-}
-
 static char *cmd;
+static int restricted;
 
 /**/
 void
@@ -206,7 +108,7 @@ parseargs(char **argv)
     LinkList paramlist;
     int bourne = (emulation == EMULATE_KSH || emulation == EMULATE_SH);
 
-    hackzero = argzero = *argv++;
+    argzero = *argv++;
     SHIN = 0;
 
     /* There's a bit of trickery with opts[INTERACTIVE] here.  It starts *
@@ -222,7 +124,6 @@ parseargs(char **argv)
 
     /* loop through command line options (begins with "-" or "+") */
     while (*argv && (**argv == '-' || **argv == '+')) {
-	char *args = *argv;
 	action = (**argv == '-');
 	if(!argv[0][1])
 	    *argv = "--";
@@ -236,9 +137,14 @@ parseargs(char **argv)
 	    }
 
 	    if (**argv == 'c') {         /* -c command */
-		cmd = *argv;
+		if (!*++argv) {
+		    zerr("string expected after -c", NULL, 0);
+		    exit(1);
+		}
+		cmd = *argv++;
 		opts[INTERACTIVE] &= 1;
 		opts[SHINSTDIN] = 0;
+		goto doneoptions;
 	    } else if (**argv == 'o') {
 		if (!*++*argv)
 		    argv++;
@@ -248,22 +154,18 @@ parseargs(char **argv)
 		}
 		if(!(optno = optlookup(*argv)))
 		    zerr("no such option: %s", *argv, 0);
+		else if (optno == RESTRICTED)
+		    restricted = action;
 		else
 		    dosetopt(optno, action, 1);
-              break;
-	    } else if (isspace(STOUC(**argv))) {
-		/* zsh's typtab not yet set, have to use ctype */
-		while (*++*argv)
-		    if (!isspace(STOUC(**argv))) {
-			zerr("bad option string: `%s'", args, 0);
-			exit(1);
-		    }
 		break;
 	    } else {
 	    	if (!(optno = optlookupc(**argv))) {
 		    zerr("bad option: -%c", NULL, **argv);
 		    exit(1);
-		} else
+		} else if (optno == RESTRICTED)
+		    restricted = action;
+		else
 		    dosetopt(optno, action, 1);
 	    }
 	}
@@ -271,13 +173,6 @@ parseargs(char **argv)
     }
     doneoptions:
     paramlist = newlinklist();
-    if (cmd) {
-	if (!*argv) {
-	    zerr("string expected after -%s", cmd, 0);
-	    exit(1);
-	}
-	cmd = *argv++;
-    }
     if (*argv) {
 	if (unset(SHINSTDIN)) {
 	    argzero = *argv;
@@ -343,63 +238,19 @@ init_io(void)
 	SHTTY = -1;
     }
 
-    /* Send xtrace output to stderr -- see execcmd() */
-    xtrerr = stderr;
-
     /* Make sure the tty is opened read/write. */
     if (isatty(0)) {
 	zsfree(ttystrname);
-	if ((ttystrname = ztrdup(ttyname(0)))) {
-	    SHTTY = movefd(open(ttystrname, O_RDWR | O_NOCTTY));
-#ifdef TIOCNXCL
-	    /*
-	     * See if the terminal claims to be busy.  If so, and fd 0
-	     * is a terminal, try and set non-exclusive use for that.
-	     * This is something to do with Solaris over-cleverness.
-	     */
-	    if (SHTTY == -1 && errno == EBUSY)
-		ioctl(0, TIOCNXCL, 0);
-#endif
-	}
-	/*
-	 * xterm, rxvt and probably all terminal emulators except
-	 * dtterm on Solaris 2.6 & 7 have a bug. Applications are
-	 * unable to open /dev/tty or /dev/pts/<terminal number here>
-	 * because something in Sun's STREAMS modules doesn't like
-	 * it. The open() call fails with EBUSY which is not even
-	 * listed as a possibility in the open(2) man page.  So we'll
-	 * try to outsmart The Company.  -- <dave@srce.hr>
-	 *
-	 * Presumably there's no harm trying this on any OS, given that
-	 * isatty(0) worked but opening the tty didn't.  Possibly we won't
-	 * get the tty read/write, but it's the best we can do -- pws
-	 *
-	 * Try both stdin and stdout before trying /dev/tty. -- Bart
-	 */
-#if defined(HAVE_FCNTL_H) && defined(F_GETFL)
-#define rdwrtty(fd)	((fcntl(fd, F_GETFL, 0) & O_RDWR) == O_RDWR)
-#else
-#define rdwrtty(fd)	1
-#endif
-	if (SHTTY == -1 && rdwrtty(0)) {
-	    SHTTY = movefd(dup(0));
-	}
+	if ((ttystrname = ztrdup(ttyname(0))))
+	    SHTTY = movefd(open(ttystrname, O_RDWR));
     }
-    if (SHTTY == -1 && isatty(1) && rdwrtty(1) &&
-	(SHTTY = movefd(dup(1))) != -1) {
+    if (SHTTY == -1 && (SHTTY = movefd(open("/dev/tty", O_RDWR))) != -1) {
 	zsfree(ttystrname);
-	ttystrname = ztrdup(ttyname(1));
-    }
-    if (SHTTY == -1 &&
-	(SHTTY = movefd(open("/dev/tty", O_RDWR | O_NOCTTY))) != -1) {
-	zsfree(ttystrname);
-	ttystrname = ztrdup(ttyname(SHTTY));
+	ttystrname = ztrdup("/dev/tty");
     }
     if (SHTTY == -1) {
 	zsfree(ttystrname);
 	ttystrname = ztrdup("");
-    } else if (!ttystrname) {
-	ttystrname = ztrdup("/dev/tty");
     }
 
     /* We will only use zle if shell is interactive, *
@@ -414,12 +265,11 @@ init_io(void)
 #ifdef JOB_CONTROL
     /* If interactive, make the shell the foreground process */
     if (opts[MONITOR] && interact && (SHTTY != -1)) {
+	attachtty(GETPGRP());
 	if ((mypgrp = GETPGRP()) > 0) {
 	    while ((ttpgrp = gettygrp()) != -1 && ttpgrp != mypgrp) {
-		sleep(1);	/* give parent time to change pgrp */
+		sleep(1);
 		mypgrp = GETPGRP();
-		if (mypgrp == mypid)
-		    attachtty(mypgrp);
 		if (mypgrp == gettygrp())
 		    break;
 		killpg(mypgrp, SIGTTIN);
@@ -458,9 +308,6 @@ init_shout(void)
 #endif
 }
 
-/* flag for whether terminal has automargin (wraparound) capability */
-extern hasam;
-
 /* Initialise termcap */
 
 /**/
@@ -562,12 +409,10 @@ setupvals(void)
     int i;
 #endif
 
-    lineno = 1;
     noeval = 0;
     curhist = 0;
     histsiz = DEFAULT_HISTSIZE;
     inithist();
-    clwords = (char **) zcalloc((clwsize = 16) * sizeof(char *));
 
     cmdstack = (unsigned char *) zalloc(256);
     cmdsp = 0;
@@ -598,6 +443,10 @@ setupvals(void)
     mailpath = mkarray(NULL);
     watch    = mkarray(NULL);
     psvar    = mkarray(NULL);
+#ifdef DYNAMIC
+    module_path = mkarray(ztrdup(MODULE_DIR));
+    modules = newlinklist();
+#endif
 
     /* Set default prompts */
     if (opts[INTERACTIVE]) {
@@ -618,10 +467,9 @@ setupvals(void)
 
     zoptarg = ztrdup("");
     zoptind = 1;
-    schedcmds = NULL;
 
-    ppid  = (zlong) getppid();
-    mypid = (zlong) getpid();
+    ppid  = (long) getppid();
+    mypid = (long) getpid();
     term  = ztrdup("");
 
     /* The following variable assignments cause zsh to behave more *
@@ -656,7 +504,7 @@ setupvals(void)
     else if ((ptr = zgetenv("PWD")) && ispwd(ptr))
 	pwd = ztrdup(ptr);
     else
-	pwd = metafy(zgetcwd(), -1, META_REALLOC);
+	pwd = metafy(zgetcwd(), -1, META_DUP);
 
     oldpwd = ztrdup(pwd);  /* initialize `OLDPWD' = `PWD' */
 
@@ -668,29 +516,17 @@ setupvals(void)
     createcmdnamtable();    /* create hash table for external commands */
     createshfunctable();    /* create hash table for shell functions   */
     createbuiltintable();   /* create hash table for builtin commands  */
-    createcompctltable();   /* create hash table for compctls          */
     createnameddirtable();  /* create hash table for named directories */
     createparamtable();     /* create paramater hash table             */
 
 #ifdef TIOCGWINSZ
-    adjustwinsize(0);
+    adjustwinsize();
 #else
-    /* columns and lines are normally zero, unless something different *
-     * was inhereted from the environment.  If either of them are zero *
-     * the setiparam calls below set them to the defaults from termcap */
-    setiparam("COLUMNS", columns);
-    setiparam("LINES", lines);
+    /* Using zero below sets the defaults from termcap */
+    setiparam("COLUMNS", 0);
+    setiparam("LINES", 0);
 #endif
 
-    /* create hash table for multi-character emacs bindings */
-    createemkeybindtable();
-
-    /* create hash table for multi-character vi bindings */
-    createvikeybindtable();
-
-    initkeybindings();	    /* initialize key bindings */
-    compctlsetup();
-
 #ifdef HAVE_GETRLIMIT
     for (i = 0; i != RLIM_NLIMITS; i++) {
 	getrlimit(i, current_limits + i);
@@ -700,13 +536,13 @@ setupvals(void)
 
     breaks = loops = 0;
     lastmailcheck = time(NULL);
-    locallist = NULL;
     locallevel = sourcelevel = 0;
     trapreturn = 0;
     noerrexit = -1;
     nohistsave = 1;
     dirstack = newlinklist();
     bufstack = newlinklist();
+    prepromptfns = newlinklist();
     hsubl = hsubr = NULL;
     lastpid = 0;
     bshin = SHIN ? fdopen(SHIN, "r") : stdin;
@@ -785,7 +621,7 @@ run_init_scripts(void)
 	    char *s = getsparam("ENV");
 	    if (islogin)
 		sourcehome(".profile");
-	    noerrs = 2;
+	    noerrs = 1;
 	    if (s && !parsestr(s)) {
 		singsub(&s);
 		noerrs = 0;
@@ -834,6 +670,8 @@ run_init_scripts(void)
 void
 init_misc(void)
 {
+    if (*zsh_name == 'r' || restricted)
+	dosetopt(RESTRICTED, 1, 0);
     if (cmd) {
 	if (SHIN >= 10)
 	    fclose(bshin);
@@ -875,13 +713,13 @@ source(char *s)
     SHIN = tempfd;
     bshin = fdopen(SHIN, "r");
     subsh  = 0;
-    lineno = 1;
+    lineno = 0;
     loops  = 0;
     dosetopt(SHINSTDIN, 0, 1);
     scriptname = s;
 
     sourcelevel++;
-    loop(0, 0);			/* loop through the file to be sourced        */
+    loop(0);			/* loop through the file to be sourced        */
     sourcelevel--;
     fclose(bshin);
     fdtable[SHIN] = 0;
@@ -923,26 +761,97 @@ sourcehome(char *s)
 
 /**/
 void
-compctlsetup(void)
+init_bltinmods(void)
 {
-    static char
-        *os[] =
-    {"setopt", "unsetopt", NULL}, *vs[] =
-    {"export", "typeset", "vared", "unset", NULL}, *cs[] =
-    {"which", "builtin", NULL}, *bs[] =
-    {"bindkey", NULL};
-
-    compctl_process(os, CC_OPTIONS, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-		    NULL, NULL, 0);
-    compctl_process(vs, CC_VARS, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-		    NULL, NULL, 0);
-    compctl_process(bs, CC_BINDINGS, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-		    NULL, NULL, 0);
-    compctl_process(cs, CC_COMMPATH, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
-		    NULL, NULL, 0);
-    cc_compos.mask = CC_COMMPATH;
-    cc_default.refc = 10000;
-    cc_default.mask = CC_FILES;
-    cc_first.refc = 10000;
-    cc_first.mask = 0;
+    static struct module mod = { NULL, 0, NULL, NULL };
+#include "bltinmods.list"
+    mod.nam = NULL;
 }
+
+/* ZLE entry point pointers.  They are defined here because the initial *
+ * values depend on whether ZLE is linked in or not -- if it is, we     *
+ * avoid wasting space with the fallback functions.  No other source    *
+ * file needs to know which modules are linked in.                      */
+
+#ifdef LINKED_XMOD_zle
+
+ZleVoidFn trashzleptr;
+ZleVoidFn gotwordptr;
+ZleVoidFn refreshptr;
+ZleVoidIntFn spaceinlineptr;
+ZleReadFn zlereadptr;
+
+#else /* !LINKED_XMOD_zle */
+
+ZleVoidFn trashzleptr = noop_function;
+ZleVoidFn gotwordptr = noop_function;
+ZleVoidFn refreshptr = noop_function;
+ZleVoidIntFn spaceinlineptr = noop_function_int;
+# ifdef UNLINKED_XMOD_zle
+ZleReadFn zlereadptr = autoload_zleread;
+# else /* !UNLINKED_XMOD_zle */
+ZleReadFn zlereadptr = fallback_zleread;
+# endif /* !UNLINKED_XMOD_zle */
+
+/**/
+void
+noop_function(void)
+{
+    /* do nothing */
+}
+
+/**/
+void
+noop_function_int(int nothing)
+{
+    /* do nothing */
+}
+
+# ifdef UNLINKED_XMOD_zle
+
+/**/
+static unsigned char *
+autoload_zleread(char *lp, char *rp, int ha)
+{
+    zlereadptr = fallback_zleread;
+    load_module("zle");
+    return zleread(lp, rp, ha);
+}
+
+# endif /* UNLINKED_XMOD_zle */
+
+/**/
+unsigned char *
+fallback_zleread(char *lp, char *rp, int ha)
+{
+    char *pptbuf;
+    int pptlen;
+
+    pptbuf = putprompt(lp, &pptlen, NULL, 1);
+    write(2, (WRITE_ARG_2_T)pptbuf, pptlen);
+    free(pptbuf);
+    return (unsigned char *)shingetline();
+}
+
+#endif /* !LINKED_XMOD_zle */
+
+/* compctl entry point pointers.  Similar to the ZLE ones. */
+
+#ifdef LINKED_XMOD_comp1
+
+CompctlReadFn compctlreadptr;
+
+#else /* !LINKED_XMOD_comp1 */
+
+CompctlReadFn compctlreadptr = fallback_compctlread;
+
+/**/
+int
+fallback_compctlread(char *name, char **args, char *ops, char *reply)
+{
+    zwarnnam(name, "option valid only in functions called from completion",
+	    NULL, 0);
+    return 1;
+}
+
+#endif /* !LINKED_XMOD_comp1 */
