@@ -1,9 +1,11 @@
 /*
+ * $Id: input.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
+ *
  * input.c - read and store lines of input
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -68,6 +70,7 @@
  */
 
 #include "zsh.h"
+#include "input.pro"
 
 /* Input buffer variables:  inbufct and inbufflags are in globals.h. */
 static char *inbuf;		/* Current input buffer */
@@ -76,6 +79,8 @@ static char *inbufpush;		/* Character at which to re-push alias */
 static int inbufleft;		/* Characters left in current input
 				   stack element */
 
+static int lastc;		/* used as flag that end of line was reached */
+
 
  /* Input must be stacked since the input queue is used by
   * various different parts of the shell.
@@ -102,7 +107,8 @@ static int instacksz = INSTACK_INITIAL;
 /* Read a line from bshin.  Convert tokens and   *
  * null characters to Meta c^32 character pairs. */
 
-static char *
+/**/
+char *
 shingetline(void)
 {
     char *line = NULL;
@@ -150,18 +156,14 @@ shingetline(void)
 int
 ingetc(void)
 {
-    int lastc;
-
     if (lexstop)
-	return ' ';
+	return lastc = ' ';
     for (;;) {
 	if (inbufleft) {
 	    inbufleft--;
 	    inbufct--;
 	    if (itok(lastc = STOUC(*inbufptr++)))
 		continue;
-	    if (((inbufflags & INP_LINENO) || !strin) && lastc == '\n')
-		lineno++;
 	    return lastc;
 	}
 
@@ -178,24 +180,24 @@ ingetc(void)
 	 */
 	if (strin || errflag) {
 	    lexstop = 1;
-	    return ' ';
+	    return lastc = ' ';
 	}
 	/* As a last resort, get some more input */
 	if (inputline())
-	    return ' ';
+	    return lastc = ' ';
     }
 }
 
 /* Read a line from the current command stream and store it as input */
 
 /**/
-int
+static int
 inputline(void)
 {
     unsigned char *ingetcline, *ingetcpmptl = NULL, *ingetcpmptr = NULL;
 
     /* If reading code interactively, work out the prompts. */
-    if (interact && isset(SHINSTDIN)) {
+    if (interact && isset(SHINSTDIN))
 	if (!isfirstln)
 	    ingetcpmptl = (unsigned char *)prompt2;
 	else {
@@ -203,7 +205,6 @@ inputline(void)
 	    if (rprompt)
 		ingetcpmptr = (unsigned char *)rprompt;
 	}
-    }
     if (!(interact && isset(SHINSTDIN) && SHTTY != -1 && isset(USEZLE))) {
 	/*
 	 * If not using zle, read the line straight from the input file.
@@ -225,20 +226,8 @@ inputline(void)
 	    free(pptbuf);
 	}
 	ingetcline = (unsigned char *)shingetline();
-    } else {
-	/*
-	 * Since we may have to read multiple lines before getting
-	 * a complete piece of input, we tell zle not to restore the
-	 * original tty settings after reading each chunk.  Instead,
-	 * this is done when the history mechanism for the current input
-	 * terminates, which is not until we have the whole input.
-	 * This is supposed to minimise problems on systems that clobber
-	 * typeahead when the terminal settings are altered.
-	 *                     pws 1998/03/12
-	 */	
-	histdone |= HISTFLAG_SETTY;
-	ingetcline = zleread((char *)ingetcpmptl, (char *)ingetcpmptr);
-    }
+    } else
+	ingetcline = zleread((char *)ingetcpmptl, (char *)ingetcpmptr, 1);
     if (!ingetcline) {
 	return lexstop = 1;
     }
@@ -254,20 +243,23 @@ inputline(void)
 	zputs((char *)ingetcline, stderr);
 	fflush(stderr);
     }
-    if (*ingetcline && ingetcline[strlen((char *)ingetcline) - 1] == '\n' &&
-	interact && isset(SUNKEYBOARDHACK) && isset(SHINSTDIN) &&
-	SHTTY != -1 && *ingetcline && ingetcline[1] &&
-	ingetcline[strlen((char *)ingetcline) - 2] == '`') {
-	/* Junk an unmatched "`" at the end of the line. */
-	int ct;
-	unsigned char *ptr;
+    if (*ingetcline && ingetcline[strlen((char *)ingetcline) - 1] == '\n') {
+	/* We've now read a complete line. */
+	lineno++;
+	if (interact && isset(SUNKEYBOARDHACK) && isset(SHINSTDIN) &&
+	    SHTTY != -1 && *ingetcline && ingetcline[1] &&
+	    ingetcline[strlen((char *)ingetcline) - 2] == '`') {
+	    /* Junk an unmatched "`" at the end of the line. */
+	    int ct;
+	    unsigned char *ptr;
 
-	for (ct = 0, ptr = ingetcline; *ptr; ptr++)
-	    if (*ptr == '`')
-		ct++;
-	if (ct & 1) {
-	    ptr[-2] = '\n';
-	    ptr[-1] = '\0';
+	    for (ct = 0, ptr = ingetcline; *ptr; ptr++)
+		if (*ptr == '`')
+		    ct++;
+	    if (ct & 1) {
+		ptr[-2] = '\n';
+		ptr[-1] = '\0';
+	    }
 	}
     }
     isfirstch = 1;
@@ -283,7 +275,7 @@ inputline(void)
  */
 
 /**/
-void
+static void
 inputsetline(char *str, int flags)
 {
     if ((inbufflags & INP_FREE) && inbuf) {
@@ -331,8 +323,6 @@ inungetc(int c)
 	    inbufptr--;
 	    inbufct++;
 	    inbufleft++;
-	    if (((inbufflags & INP_LINENO) || !strin) && c == '\n')
-		lineno--;
 	}
 #ifdef DEBUG
         else if (!(inbufflags & INP_CONT)) {
@@ -380,7 +370,7 @@ stuff(char *fn)
 {
     FILE *in;
     char *buf;
-    off_t len;
+    int len;
 
     if (!(in = fopen(unmeta(fn), "r"))) {
 	zerr("can't open %s", fn, 0);
@@ -415,7 +405,7 @@ inerrflush(void)
      * it is only used in the history code, where that is the only
      * completely safe way of discarding input.
      */
-    while (!lexstop && inbufct)
+    while ((strin || lastc != '\n') && !lexstop)
 	ingetc();
 }
 
@@ -479,7 +469,7 @@ inpush(char *str, int flags, Alias inalias)
 /* Remove the top element of the stack */
 
 /**/
-void
+static void
 inpoptop(void)
 {
     if (inbuf && (inbufflags & INP_FREE))
