@@ -1,11 +1,11 @@
 /*
- * $Id: jobs.c,v 2.17 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: jobs.c,v 3.1.2.7 1997/06/03 03:52:01 hzoli Exp $
  *
  * jobs.c - job control
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "jobs.pro"
 
 /* empty job structure for quick clearing of jobtab entries */
 
@@ -40,7 +41,7 @@ struct timeval dtimeval, now;
 /* Diff two timevals for elapsed-time computations */
 
 /**/
-struct timeval *
+static struct timeval *
 dtime(struct timeval *dt, struct timeval *t1, struct timeval *t2)
 {
     dt->tv_sec = t2->tv_sec - t1->tv_sec;
@@ -137,9 +138,8 @@ update_job(Job jn)
     job = jn - jobtab;   /* compute job number */
 
     if (somestopped) {
-	if (shout && job == thisjob) {
-	    if (!jn->ty)
-		jn->ty = (struct ttyinfo *) zalloc(sizeof(struct ttyinfo));
+	if (jn->stty_in_env && !jn->ty) {
+	    jn->ty = (struct ttyinfo *) zalloc(sizeof(struct ttyinfo));
 	    gettyinfo(jn->ty);
 	}
 	if (jn->stat & STAT_STOPPED)
@@ -155,8 +155,7 @@ update_job(Job jn)
 	}
     }
 
-    if (shout && !ttyfrozen && !jn->stty_in_env &&
-	job == thisjob && !somestopped && !(jn->stat & STAT_NOSTTY))
+    if (shout && !ttyfrozen && !jn->stty_in_env)
 	gettyinfo(&shttyinfo);
 
     if (isset(MONITOR)) {
@@ -166,7 +165,7 @@ update_job(Job jn)
 	if (mypgrp != pgrp && inforeground &&
 	    (jn->gleader == pgrp || (pgrp > 1 && kill(-pgrp, 0) == -1))) {
 	    attachtty(mypgrp);
-	    adjustwinsize(0);   /* check window size and adjust if necessary */
+	    adjustwinsize();   /* check window size and adjust if necessary */
 	}
     }
 
@@ -213,9 +212,192 @@ update_job(Job jn)
     }
 }
 
-/* lng = 0 means jobs    *
- * lng = 1 means jobs -l *
- * lng = 2 means jobs -p 
+/* set the previous job to something reasonable */
+
+/**/
+static void
+setprevjob(void)
+{
+    int i;
+
+    for (i = MAXJOB - 1; i; i--)
+	if ((jobtab[i].stat & STAT_INUSE) && (jobtab[i].stat & STAT_STOPPED) &&
+	    i != curjob && i != thisjob) {
+	    prevjob = i;
+	    return;
+	}
+
+    for (i = MAXJOB - 1; i; i--)
+	if ((jobtab[i].stat & STAT_INUSE) && i != curjob && i != thisjob) {
+	    prevjob = i;
+	    return;
+	}
+
+    prevjob = -1;
+}
+
+static long clktck = 0;
+
+/**/
+static void
+set_clktck(void)
+{
+#ifdef _SC_CLK_TCK
+    if (!clktck)
+	/* fetch clock ticks per second from *
+	 * sysconf only the first time       */
+	clktck = sysconf(_SC_CLK_TCK);
+#else
+# ifdef __NeXT__
+    /* NeXTStep 3.3 defines CLK_TCK wrongly */
+    clktck = 60;
+# else
+#  ifdef CLK_TCK
+    clktck = CLK_TCK;
+#  else
+#   ifdef HZ
+     clktck = HZ;
+#   else
+     clktck = 60;
+#   endif
+#  endif
+# endif
+#endif
+}
+
+/**/
+static void
+printhhmmss(double secs)
+{
+    int mins = (int) secs / 60;
+    int hours = mins / 60;
+
+    secs -= 60 * mins;
+    mins -= 60 * hours;
+    if (hours)
+	fprintf(stderr, "%d:%02d:%05.2f", hours, mins, secs);
+    else if (mins)
+	fprintf(stderr,      "%d:%05.2f",        mins, secs);
+    else
+	fprintf(stderr,           "%.3f",              secs);
+}
+
+/**/
+static void
+printtime(struct timeval *real, struct timeinfo *ti, char *desc)
+{
+    char *s;
+    double elapsed_time, user_time, system_time;
+    int percent;
+
+    if (!desc)
+	desc = "";
+
+    set_clktck();
+    /* go ahead and compute these, since almost every TIMEFMT will have them */
+    elapsed_time = real->tv_sec + real->tv_usec / 1000000.0;
+    user_time    = ti->ut / (double) clktck;
+    system_time  = ti->st / (double) clktck;
+    percent      =  100.0 * (ti->ut + ti->st)
+	/ (clktck * real->tv_sec + clktck * real->tv_usec / 1000000.0);
+
+    if (!(s = getsparam("TIMEFMT")))
+	s = DEFAULT_TIMEFMT;
+
+    for (; *s; s++)
+	if (*s == '%')
+	    switch (*++s) {
+	    case 'E':
+		fprintf(stderr, "%4.2fs", elapsed_time);
+		break;
+	    case 'U':
+		fprintf(stderr, "%4.2fs", user_time);
+		break;
+	    case 'S':
+		fprintf(stderr, "%4.2fs", system_time);
+		break;
+	    case '*':
+		switch (*++s) {
+		case 'E':
+		    printhhmmss(elapsed_time);
+		    break;
+		case 'U':
+		    printhhmmss(user_time);
+		    break;
+		case 'S':
+		    printhhmmss(system_time);
+		    break;
+		default:
+		    fprintf(stderr, "%%*");
+		    s--;
+		    break;
+		}
+		break;
+	    case 'P':
+		fprintf(stderr, "%d%%", percent);
+		break;
+	    case 'J':
+		fprintf(stderr, "%s", desc);
+		break;
+	    case '%':
+		putc('%', stderr);
+		break;
+	    case '\0':
+		s--;
+		break;
+	    default:
+		fprintf(stderr, "%%%c", *s);
+		break;
+	} else
+	    putc(*s, stderr);
+    putc('\n', stderr);
+    fflush(stderr);
+}
+
+/**/
+static void
+dumptime(Job jn)
+{
+    Process pn;
+
+    if (!jn->procs)
+	return;
+    for (pn = jn->procs; pn; pn = pn->next)
+	printtime(dtime(&dtimeval, &pn->bgtime, &pn->endtime), &pn->ti, pn->text);
+}
+
+/* Check whether shell should report the amount of time consumed   *
+ * by job.  This will be the case if we have preceded the command  *
+ * with the keyword time, or if REPORTTIME is non-negative and the *
+ * amount of time consumed by the job is greater than REPORTTIME   */
+
+/**/
+static int
+should_report_time(Job j)
+{
+    Value v;
+    char *s = "REPORTTIME";
+    int reporttime;
+
+    /* if the time keyword was used */
+    if (j->stat & STAT_TIMED)
+	return 1;
+
+    if (!(v = getvalue(&s, 0)) || (reporttime = getintvalue(v)) < 0)
+	return 0;
+
+    /* can this ever happen? */
+    if (!j->procs)
+	return 0;
+
+    set_clktck();
+    return ((j->procs->ti.ut + j->procs->ti.st) / clktck >= reporttime);
+}
+
+/* !(lng & 3) means jobs    *
+ *  (lng & 1) means jobs -l *
+ *  (lng & 2) means jobs -p
+ *  (lng & 4) means jobs -d
  *
  * synch = 0 means asynchronous
  * synch = 1 means synchronous
@@ -282,7 +464,7 @@ printjob(Job jn, int lng, int synch)
 	    putc('\n', fout);
 	for (pn = jn->procs; pn;) {
 	    len2 = ((job == thisjob) ? 5 : 10) + len;	/* 2 spaces */
-	    if (lng)
+	    if (lng & 3)
 		qn = pn->next;
 	    else
 		for (qn = pn->next; qn; qn = qn->next) {
@@ -302,18 +484,17 @@ printjob(Job jn, int lng, int synch)
 		    fprintf(fout, (job > 9) ? "        " : "       ");
 	    else
 		fprintf(fout, "zsh: ");
-	    if (lng)
-		if (lng == 1)
-		    fprintf(fout, "%ld ", (long) pn->pid);
-		else {
-		    pid_t x = jn->gleader;
+	    if (lng & 1)
+		fprintf(fout, "%ld ", (long) pn->pid);
+	    else if (lng & 2) {
+		pid_t x = jn->gleader;
 
-		    fprintf(fout, "%ld ", (long) x);
-		    do
-			skip++;
-		    while ((x /= 10));
+		fprintf(fout, "%ld ", (long) x);
+		do
 		    skip++;
-		    lng = 0;
+		while ((x /= 10));
+		skip++;
+		lng &= ~3;
 	    } else
 		fprintf(fout, "%*s", skip, "");
 	    if (pn->status == SP_RUNNING)
@@ -346,11 +527,13 @@ printjob(Job jn, int lng, int synch)
 	fflush(fout);
     }
 
-/* print "(pwd now: foo)" messages */
+/* print "(pwd now: foo)" messages: with (lng & 4) we are printing
+ * the directory where the job is running, otherwise the current directory
+ */
 
-    if (interact && job == thisjob && strcmp(jn->pwd, pwd)) {
-	fprintf(shout, "(pwd now: ");
-	fprintdir(pwd, shout);
+    if ((lng & 4) || (interact && job == thisjob && strcmp(jn->pwd, pwd))) {
+	fprintf(shout, "(pwd %s: ", (lng & 4) ? "" : "now");
+	fprintdir((lng & 4) ? jn->pwd : pwd, shout);
 	fprintf(shout, ")\n");
 	fflush(shout);
     }
@@ -404,30 +587,6 @@ deletejob(Job jn)
     *jn = zero;
 }
 
-/* set the previous job to something reasonable */
-
-/**/
-void
-setprevjob(void)
-{
-    int i;
-
-    for (i = MAXJOB - 1; i; i--)
-	if ((jobtab[i].stat & STAT_INUSE) && (jobtab[i].stat & STAT_STOPPED) &&
-	    i != curjob && i != thisjob) {
-	    prevjob = i;
-	    return;
-	}
-
-    for (i = MAXJOB - 1; i; i--)
-	if ((jobtab[i].stat & STAT_INUSE) && i != curjob && i != thisjob) {
-	    prevjob = i;
-	    return;
-	}
-
-    prevjob = -1;
-}
-
 /* add a process to the current job */
 
 /**/
@@ -507,7 +666,7 @@ waitforpid(pid_t pid)
 /* wait for a job to finish */
 
 /**/
-void
+static void
 waitjob(int job, int sig)
 {
     Job jn = jobtab + job;
@@ -642,163 +801,6 @@ spawnjob(void)
     thisjob = -1;
 }
 
-static long clktck = 0;
-
-static void
-set_clktck(void)
-{
-#ifdef _SC_CLK_TCK
-    if (!clktck)
-	/* fetch clock ticks per second from *
-	 * sysconf only the first time       */
-	clktck = sysconf(_SC_CLK_TCK);
-#else
-# ifdef __NeXT__
-    /* NeXTStep 3.3 defines CLK_TCK wrongly */
-    clktck = 60;
-# else
-#  ifdef CLK_TCK
-    clktck = CLK_TCK;
-#  else
-#   ifdef HZ
-     clktck = HZ;
-#   else
-     clktck = 60;
-#   endif
-#  endif
-# endif
-#endif
-}
-
-/* Check whether shell should report the amount of time consumed   *
- * by job.  This will be the case if we have preceded the command  *
- * with the keyword time, or if REPORTTIME is non-negative and the *
- * amount of time consumed by the job is greater than REPORTTIME   */
-
-/**/
-int
-should_report_time(Job j)
-{
-    Value v;
-    char *s = "REPORTTIME";
-    int reporttime;
-
-    /* if the time keyword was used */
-    if (j->stat & STAT_TIMED)
-	return 1;
-
-    if (!(v = getvalue(&s, 0)) || (reporttime = getintvalue(v)) < 0)
-	return 0;
-
-    /* can this ever happen? */
-    if (!j->procs)
-	return 0;
-
-    set_clktck();
-    return ((j->procs->ti.ut + j->procs->ti.st) / clktck >= reporttime);
-}
-
-/**/
-void
-printhhmmss(double secs)
-{
-    int mins = (int) secs / 60;
-    int hours = mins / 60;
-
-    secs -= 60 * mins;
-    mins -= 60 * hours;
-    if (hours)
-	fprintf(stderr, "%d:%02d:%05.2f", hours, mins, secs);
-    else if (mins)
-	fprintf(stderr,      "%d:%05.2f",        mins, secs);
-    else
-	fprintf(stderr,           "%.3f",              secs);
-}
-
-/**/
-void
-printtime(struct timeval *real, struct timeinfo *ti, char *desc)
-{
-    char *s;
-    double elapsed_time, user_time, system_time;
-    int percent;
-
-    if (!desc)
-	desc = "";
-
-    set_clktck();
-    /* go ahead and compute these, since almost every TIMEFMT will have them */
-    elapsed_time = real->tv_sec + real->tv_usec / 1000000.0;
-    user_time    = ti->ut / (double) clktck;
-    system_time  = ti->st / (double) clktck;
-    percent      =  100.0 * (ti->ut + ti->st)
-	/ (clktck * real->tv_sec + clktck * real->tv_usec / 1000000.0);
-
-    if (!(s = getsparam("TIMEFMT")))
-	s = DEFAULT_TIMEFMT;
-
-    for (; *s; s++)
-	if (*s == '%')
-	    switch (*++s) {
-	    case 'E':
-		fprintf(stderr, "%4.2fs", elapsed_time);
-		break;
-	    case 'U':
-		fprintf(stderr, "%4.2fs", user_time);
-		break;
-	    case 'S':
-		fprintf(stderr, "%4.2fs", system_time);
-		break;
-	    case '*':
-		switch (*++s) {
-		case 'E':
-		    printhhmmss(elapsed_time);
-		    break;
-		case 'U':
-		    printhhmmss(user_time);
-		    break;
-		case 'S':
-		    printhhmmss(system_time);
-		    break;
-		default:
-		    fprintf(stderr, "%%*");
-		    s--;
-		    break;
-		}
-		break;
-	    case 'P':
-		fprintf(stderr, "%d%%", percent);
-		break;
-	    case 'J':
-		fprintf(stderr, "%s", desc);
-		break;
-	    case '%':
-		putc('%', stderr);
-		break;
-	    case '\0':
-		s--;
-		break;
-	    default:
-		fprintf(stderr, "%%%c", *s);
-		break;
-	} else
-	    putc(*s, stderr);
-    putc('\n', stderr);
-    fflush(stderr);
-}
-
-/**/
-void
-dumptime(Job jn)
-{
-    Process pn;
-
-    if (!jn->procs)
-	return;
-    for (pn = jn->procs; pn; pn = pn->next)
-	printtime(dtime(&dtimeval, &pn->bgtime, &pn->endtime), &pn->ti, pn->text);
-}
-
 /**/
 void
 shelltime(void)
@@ -830,30 +832,489 @@ scanjobs(void)
             printjob(jobtab + i, 0, 1);
 }
 
-/* check to see if user has jobs running/stopped */
+/**** job control builtins ****/
+
+/* This simple function indicates whether or not s may represent      *
+ * a number.  It returns true iff s consists purely of digits and     *
+ * minuses.  Note that minus may appear more than once, and the empty *
+ * string will produce a `true' response.                             */
 
 /**/
-void
-checkjobs(void)
+static int
+isanum(char *s)
 {
-    int i;
+    while (*s == '-' || idigit(*s))
+	s++;
+    return *s == '\0';
+}
 
-    for (i = 1; i < MAXJOB; i++)
-	if (i != thisjob && (jobtab[i].stat & STAT_LOCKED) &&
-	    !(jobtab[i].stat & STAT_NOPRINT))
-	    break;
-    if (i < MAXJOB) {
-	if (jobtab[i].stat & STAT_STOPPED) {
+/* Make sure we have a suitable current and previous job set. */
 
-#ifdef USE_SUSPENDED
-	    zerr("you have suspended jobs.", NULL, 0);
-#else
-	    zerr("you have stopped jobs.", NULL, 0);
-#endif
-
-	} else
-	    zerr("you have running jobs.", NULL, 0);
-	stopmsg = 1;
+/**/
+static void
+setcurjob(void)
+{
+    if (curjob == thisjob ||
+	(curjob != -1 && !(jobtab[curjob].stat & STAT_INUSE))) {
+	curjob = prevjob;
+	setprevjob();
+	if (curjob == thisjob ||
+	    (curjob != -1 && !((jobtab[curjob].stat & STAT_INUSE) &&
+			       curjob != thisjob))) {
+	    curjob = prevjob;
+	    setprevjob();
+	}
     }
 }
 
+/* Convert a job specifier ("%%", "%1", "%foo", "%?bar?", etc.) *
+ * to a job number.                                             */
+
+/**/
+static int
+getjob(char *s, char *prog)
+{
+    int jobnum, returnval;
+
+    /* if there is no %, treat as a name */
+    if (*s != '%')
+	goto jump;
+    s++;
+    /* "%%", "%+" and "%" all represent the current job */
+    if (*s == '%' || *s == '+' || !*s) {
+	if (curjob == -1) {
+	    zwarnnam(prog, "no current job", NULL, 0);
+	    returnval = -1;
+	    goto done;
+	}
+	returnval = curjob;
+	goto done;
+    }
+    /* "%-" represents the previous job */
+    if (*s == '-') {
+	if (prevjob == -1) {
+	    zwarnnam(prog, "no previous job", NULL, 0);
+	    returnval = -1;
+	    goto done;
+	}
+	returnval = prevjob;
+	goto done;
+    }
+    /* a digit here means we have a job number */
+    if (idigit(*s)) {
+	jobnum = atoi(s);
+	if (jobnum && jobnum < MAXJOB && jobtab[jobnum].stat &&
+	    !(jobtab[jobnum].stat & STAT_SUBJOB) && jobnum != thisjob) {
+	    returnval = jobnum;
+	    goto done;
+	}
+	zwarnnam(prog, "%%%s: no such job", s, 0);
+	returnval = -1;
+	goto done;
+    }
+    /* "%?" introduces a search string */
+    if (*s == '?') {
+	struct process *pn;
+
+	for (jobnum = MAXJOB - 1; jobnum >= 0; jobnum--)
+	    if (jobtab[jobnum].stat && !(jobtab[jobnum].stat & STAT_SUBJOB) &&
+		jobnum != thisjob)
+		for (pn = jobtab[jobnum].procs; pn; pn = pn->next)
+		    if (strstr(pn->text, s + 1)) {
+			returnval = jobnum;
+			goto done;
+		    }
+	zwarnnam(prog, "job not found: %s", s, 0);
+	returnval = -1;
+	goto done;
+    }
+  jump:
+    /* anything else is a job name, specified as a string that begins the
+    job's command */
+    if ((jobnum = findjobnam(s)) != -1) {
+	returnval = jobnum;
+	goto done;
+    }
+    /* if we get here, it is because none of the above succeeded and went
+    to done */
+    zwarnnam(prog, "job not found: %s", s, 0);
+    returnval = -1;
+  done:
+    return returnval;
+}
+
+/* For jobs -Z (which modifies the shell's name as seen in ps listings).  *
+ * hackzero is the start of the safely writable space, and hackspace is   *
+ * its length, excluding a final NUL terminator that will always be left. */
+
+static char *hackzero;
+static int hackspace;
+
+/* Initialise the jobs -Z system.  The technique is borrowed from perl: *
+ * check through the argument and environment space, to see how many of *
+ * the strings are in contiguous space.  This determines the value of   *
+ * hackspace.                                                           */
+
+/**/
+void
+init_hackzero(char **argv, char **envp)
+{
+    char *p, *q;
+
+    hackzero = *argv;
+    p = strchr(hackzero, 0);
+    while(*++argv) {
+	q = *argv;
+	if(q != p+1)
+	    goto done;
+	p = strchr(q, 0);
+    }
+    for(; *envp; envp++) {
+	q = *envp;
+	if(q != p+1)
+	    goto done;
+	p = strchr(q, 0);
+    }
+    done:
+    hackspace = p - hackzero;
+}
+
+/* bg, disown, fg, jobs, wait: most of the job control commands are     *
+ * here.  They all take the same type of argument.  Exception: wait can *
+ * take a pid or a job specifier, whereas the others only work on jobs. */
+
+/**/
+int
+bin_fg(char *name, char **argv, char *ops, int func)
+{
+    int job, lng, firstjob = -1, retval = 0;
+
+    if (ops['Z']) {
+	int len;
+
+	if(isset(RESTRICTED)) {
+	    zwarnnam(name, "-Z is restricted", NULL, 0);
+	    return 1;
+	}
+	if(!argv[0] || argv[1]) {
+	    zwarnnam(name, "-Z requires one argument", NULL, 0);
+	    return 1;
+	}
+	unmetafy(*argv, &len);
+	if(len > hackspace)
+	    len = hackspace;
+	memcpy(hackzero, *argv, len);
+	memset(hackzero + len, 0, hackspace - len);
+	return 0;
+    }
+
+    lng = (ops['l']) ? 1 : (ops['p']) ? 2 : 0;
+    if (ops['d'])
+	lng |= 4;
+    
+    if ((func == BIN_FG || func == BIN_BG) && !jobbing) {
+	/* oops... maybe bg and fg should have been disabled? */
+	zwarnnam(name, "no job control in this shell.", NULL, 0);
+	return 1;
+    }
+
+    /* If necessary, update job table. */
+    if (unset(NOTIFY))
+	scanjobs();
+
+    setcurjob();
+
+    if (func == BIN_JOBS)
+        /* If you immediately type "exit" after "jobs", this      *
+         * will prevent zexit from complaining about stopped jobs */
+	stopmsg = 2;
+    if (!*argv)
+	/* This block handles all of the default cases (no arguments).  bg,
+	fg and disown act on the current job, and jobs and wait act on all the
+	jobs. */
+ 	if (func == BIN_FG || func == BIN_BG || func == BIN_DISOWN) {
+	    /* W.r.t. the above comment, we'd better have a current job at this
+	    point or else. */
+	    if (curjob == -1 || (jobtab[curjob].stat & STAT_NOPRINT)) {
+		zwarnnam(name, "no current job", NULL, 0);
+		return 1;
+	    }
+	    firstjob = curjob;
+	} else if (func == BIN_JOBS) {
+	    /* List jobs. */
+	    for (job = 0; job != MAXJOB; job++)
+		if (job != thisjob && jobtab[job].stat) {
+		    if ((!ops['r'] && !ops['s']) ||
+			(ops['r'] && ops['s']) ||
+			(ops['r'] && !(jobtab[job].stat & STAT_STOPPED)) ||
+			(ops['s'] && jobtab[job].stat & STAT_STOPPED))
+			printjob(job + jobtab, lng, 2);
+		}
+	    return 0;
+	} else {   /* Must be BIN_WAIT, so wait for all jobs */
+	    for (job = 0; job != MAXJOB; job++)
+		if (job != thisjob && jobtab[job].stat)
+		    waitjob(job, SIGINT);
+	    return 0;
+	}
+
+    /* Defaults have been handled.  We now have an argument or two, or three...
+    In the default case for bg, fg and disown, the argument will be provided by
+    the above routine.  We now loop over the arguments. */
+    for (; (firstjob != -1) || *argv; (void)(*argv && argv++)) {
+	int stopped, ocj = thisjob;
+
+	if (func == BIN_WAIT && isanum(*argv)) {
+	    /* wait can take a pid; the others can't. */
+	    waitforpid((long)atoi(*argv));
+	    retval = lastval2;
+	    thisjob = ocj;
+	    continue;
+	}
+	/* The only type of argument allowed now is a job spec.  Check it. */
+	job = (*argv) ? getjob(*argv, name) : firstjob;
+	firstjob = -1;
+	if (job == -1) {
+	    retval = 1;
+	    break;
+	}
+	if (!(jobtab[job].stat & STAT_INUSE) ||
+	    (jobtab[job].stat & STAT_NOPRINT)) {
+	    zwarnnam(name, "no such job: %d", 0, job);
+	    return 1;
+	}
+	/* We have a job number.  Now decide what to do with it. */
+	switch (func) {
+	case BIN_FG:
+	case BIN_BG:
+	case BIN_WAIT:
+	    if ((stopped = (jobtab[job].stat & STAT_STOPPED)))
+		makerunning(jobtab + job);
+	    else if (func == BIN_BG) {
+		/* Silly to bg a job already running. */
+		zwarnnam(name, "job already in background", NULL, 0);
+		thisjob = ocj;
+		return 1;
+	    }
+	    /* It's time to shuffle the jobs around!  Reset the current job,
+	    and pick a sensible secondary job. */
+	    if (curjob == job) {
+		curjob = prevjob;
+		prevjob = (func == BIN_BG) ? -1 : job;
+	    }
+	    if (prevjob == job || prevjob == -1)
+		setprevjob();
+	    if (curjob == -1) {
+		curjob = prevjob;
+		setprevjob();
+	    }
+	    if (func != BIN_WAIT)
+		/* for bg and fg -- show the job we are operating on */
+		printjob(jobtab + job, (stopped) ? -1 : 0, 1);
+	    if (func != BIN_BG) {		/* fg or wait */
+		if (strcmp(jobtab[job].pwd, pwd)) {
+		    fprintf(shout, "(pwd : ");
+		    fprintdir(jobtab[job].pwd, shout);
+		    fprintf(shout, ")\n");
+		}
+		fflush(shout);
+		if (func != BIN_WAIT) {		/* fg */
+		    thisjob = job;
+		    attachtty(jobtab[job].gleader);
+		}
+	    }
+	    if (stopped) {
+		if (func != BIN_BG && jobtab[job].ty)
+		    settyinfo(jobtab[job].ty);
+		killjb(jobtab + job, SIGCONT);
+	    }
+	    if (func == BIN_WAIT)
+	        waitjob(job, SIGINT);
+	    if (func != BIN_BG) {
+		waitjobs();
+		retval = lastval2;
+	    }
+	    break;
+	case BIN_JOBS:
+	    printjob(job + jobtab, lng, 2);
+	    break;
+	case BIN_DISOWN:
+	    {
+		static struct job zero;
+
+		jobtab[job] = zero;
+		break;
+	    }
+	}
+	thisjob = ocj;
+    }
+    return retval;
+}
+
+/* kill: send a signal to a process.  The process(es) may be specified *
+ * by job specifier (see above) or pid.  A signal, defaulting to       *
+ * SIGTERM, may be specified by name or number, preceded by a dash.    */
+
+/**/
+int
+bin_kill(char *nam, char **argv, char *ops, int func)
+{
+    int sig = SIGTERM;
+    int returnval = 0;
+
+    /* check for, and interpret, a signal specifier */
+    if (*argv && **argv == '-') {
+	if (idigit((*argv)[1]))
+	    /* signal specified by number */
+	    sig = atoi(*argv + 1);
+	else if ((*argv)[1] != '-' || (*argv)[2]) {
+	    char *signame;
+
+	    /* with argument "-l" display the list of signal names */
+	    if ((*argv)[1] == 'l' && (*argv)[2] == '\0') {
+		if (argv[1]) {
+		    while (*++argv) {
+			sig = zstrtol(*argv, &signame, 10);
+			if (signame == *argv) {
+			    for (sig = 1; sig <= SIGCOUNT; sig++)
+				if (!cstrpcmp(sigs + sig, &signame))
+				    break;
+			    if (sig > SIGCOUNT) {
+				zwarnnam(nam, "unknown signal: SIG%s",
+					 signame, 0);
+				returnval++;
+			    } else
+				printf("%d\n", sig);
+			} else {
+			    if (*signame) {
+				zwarnnam(nam, "unknown signal: SIG%s",
+					 signame, 0);
+				returnval++;
+			    } else {
+				if (WIFSIGNALED(sig))
+				    sig = WTERMSIG(sig);
+				else if (WIFSTOPPED(sig))
+				    sig = WSTOPSIG(sig);
+				if (1 <= sig && sig <= SIGCOUNT)
+				    printf("%s\n", sigs[sig]);
+				else
+				    printf("%d\n", sig);
+			    }
+			}
+		    }
+		    return returnval;
+		}
+		printf("%s", sigs[1]);
+		for (sig = 2; sig <= SIGCOUNT; sig++)
+		    printf(" %s", sigs[sig]);
+		putchar('\n');
+		return 0;
+	    }
+	    if ((*argv)[1] == 's' && (*argv)[2] == '\0')
+		signame = *++argv;
+	    else
+		signame = *argv + 1;
+
+	    /* check for signal matching specified name */
+	    for (sig = 1; sig <= SIGCOUNT; sig++)
+		if (!cstrpcmp(sigs + sig, &signame))
+		    break;
+	    if (*signame == '0' && !signame[1])
+		sig = 0;
+	    if (sig > SIGCOUNT) {
+		zwarnnam(nam, "unknown signal: SIG%s", signame, 0);
+		zwarnnam(nam, "type kill -l for a List of signals", NULL, 0);
+		return 1;
+	    }
+	}
+	argv++;
+    }
+
+    setcurjob();
+
+    /* Remaining arguments specify processes.  Loop over them, and send the
+    signal (number sig) to each process. */
+    for (; *argv; argv++) {
+	if (**argv == '%') {
+	    /* job specifier introduced by '%' */
+	    int p;
+
+	    if ((p = getjob(*argv, nam)) == -1) {
+		returnval++;
+		continue;
+	    }
+	    if (killjb(jobtab + p, sig) == -1) {
+		zwarnnam("kill", "kill %s failed: %e", *argv, errno);
+		returnval++;
+		continue;
+	    }
+	    /* automatically update the job table if sending a SIGCONT to a
+	    job, and send the job a SIGCONT if sending it a non-stopping
+	    signal. */
+	    if (jobtab[p].stat & STAT_STOPPED) {
+		if (sig == SIGCONT)
+		    jobtab[p].stat &= ~STAT_STOPPED;
+		if (sig != SIGKILL && sig != SIGCONT && sig != SIGTSTP
+		    && sig != SIGTTOU && sig != SIGTTIN && sig != SIGSTOP)
+		    killjb(jobtab + p, SIGCONT);
+	    }
+	} else if (!isanum(*argv)) {
+	    zwarnnam("kill", "illegal pid: %s", *argv, 0);
+	    returnval++;
+	} else if (kill(atoi(*argv), sig) == -1) {
+	    zwarnnam("kill", "kill %s failed: %e", *argv, errno);
+	    returnval++;
+	}
+    }
+    return returnval < 126 ? returnval : 1;
+}
+
+/* Suspend this shell */
+
+/**/
+int
+bin_suspend(char *name, char **argv, char *ops, int func)
+{
+    /* won't suspend a login shell, unless forced */
+    if (islogin && !ops['f']) {
+	zwarnnam(name, "can't suspend login shell", NULL, 0);
+	return 1;
+    }
+    if (jobbing) {
+	/* stop ignoring signals */
+	signal_default(SIGTTIN);
+	signal_default(SIGTSTP);
+	signal_default(SIGTTOU);
+    }
+    /* suspend ourselves with a SIGTSTP */
+    kill(0, SIGTSTP);
+    if (jobbing) {
+	/* stay suspended */
+	while (gettygrp() != mypgrp) {
+	    sleep(1);
+	    if (gettygrp() != mypgrp)
+		kill(0, SIGTTIN);
+	}
+	/* restore signal handling */
+	signal_ignore(SIGTTOU);
+	signal_ignore(SIGTSTP);
+	signal_ignore(SIGTTIN);
+    }
+    return 0;
+}
+
+/* find a job named s */
+
+/**/
+int
+findjobnam(char *s)
+{
+    int jobnum;
+
+    for (jobnum = MAXJOB - 1; jobnum >= 0; jobnum--)
+	if (!(jobtab[jobnum].stat & (STAT_SUBJOB | STAT_NOPRINT)) &&
+	    jobtab[jobnum].stat && jobtab[jobnum].procs && jobnum != thisjob &&
+	    jobtab[jobnum].procs->text && strpfx(s, jobtab[jobnum].procs->text))
+	    return jobnum;
+    return -1;
+}
