@@ -1,9 +1,11 @@
 /*
+ * $Id: lex.c,v 3.1.2.7 1997/06/01 06:13:15 hzoli Exp $
+ *
  * lex.c - lexical analysis
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -28,11 +30,13 @@
  */
 
 #include "zsh.h"
+#include "lex.pro"
 
 /* lexical state */
 
-int len = 0, bsiz = 256;
-char *bptr;
+static int dbparens;
+static int len = 0, bsiz = 256;
+static char *bptr;
 extern int hwgetword;
 
 struct lexstack {
@@ -41,6 +45,7 @@ struct lexstack {
     int incmdpos;
     int incond;
     int incasepat;
+    int dbparens;
     int isfirstln;
     int isfirstch;
     int histactive;
@@ -88,6 +93,7 @@ lexsave(void)
     ls->incmdpos = incmdpos;
     ls->incond = incond;
     ls->incasepat = incasepat;
+    ls->dbparens = dbparens;
     ls->isfirstln = isfirstln;
     ls->isfirstch = isfirstch;
     ls->histactive = histactive;
@@ -132,6 +138,7 @@ lexrestore(void)
     incmdpos = lstack->incmdpos;
     incond = lstack->incond;
     incasepat = lstack->incasepat;
+    dbparens = lstack->dbparens;
     isfirstln = lstack->isfirstln;
     isfirstch = lstack->isfirstch;
     histactive = lstack->histactive;
@@ -209,6 +216,7 @@ ctxtlex(void)
     case NEWLIN:
     case SEMI:
     case DSEMI:
+    case SEMIAMP:
     case AMPER:
     case AMPERBANG:
     case INPAR:
@@ -234,6 +242,8 @@ ctxtlex(void)
 	incmdpos = 0;
 	break;
     }
+    if (tok != DINPAR)
+	infor = tok == FOR ? 2 : 0;
     if (IS_REDIROP(tok) || tok == FOR || tok == FOREACH || tok == SELECT) {
 	inredir = 1;
 	oldpos = incmdpos;
@@ -315,7 +325,7 @@ void
 lexinit(void)
 {
     incond = incasepat = nocorrect =
-    lexstop = 0;
+    infor = dbparens = lexstop = 0;
     incmdpos = 1;
     tok = ENDINPUT;
 }
@@ -340,12 +350,11 @@ add(int c)
 
 #define SETPARBEGIN {if (zleparse && !(inbufflags & INP_ALIAS) && cs >= ll+1-inbufct) parbegin = inbufct;}
 #define SETPAREND {\
-	    if (zleparse && !(inbufflags & INP_ALIAS) && parbegin != -1 && parend == -1) {\
+	    if (zleparse && !(inbufflags & INP_ALIAS) && parbegin != -1 && parend == -1)\
 		if (cs >= ll + 1 - inbufct)\
 		    parbegin = -1;\
 		else\
-		    parend = inbufct;\
-	    } }
+		    parend = inbufct;}
 
 static int
 cmd_or_math(int cs_type)
@@ -413,7 +422,25 @@ gettok(void)
     wordbeg = inbufct - (qbang && c == bangchar);
     hwbegin(-1-(qbang && c == bangchar));
     /* word includes the last character read and possibly \ before ! */
-    if (idigit(c)) {	/* handle 1< foo */
+    if (dbparens) {
+	len = 0;
+	bptr = tokstr = (char *)ncalloc(bsiz = 256);
+	hungetc(c);
+	cmdpush(CS_MATH);
+	c = dquote_parse(infor ? ';' : ')', 0);
+	cmdpop();
+	*bptr = '\0';
+	if (!c && infor) {
+	    infor--;
+	    return DINPAR;
+	}
+	if (c || (c = hgetc()) != ')') {
+	    hungetc(c);
+	    return LEXERR;
+	}
+	dbparens = 0;
+	return DOUTPAR;
+    } else if (idigit(c)) {	/* handle 1< foo */
 	d = hgetc();
 	if (d == '>' || d == '<') {
 	    peekfd = c - '0';
@@ -461,12 +488,13 @@ gettok(void)
 	return NEWLIN;
     case LX1_SEMI:
 	d = hgetc();
-	if (d != ';') {
-	    hungetc(d);
-	    lexstop = 0;
-	    return SEMI;
-	}
-	return DSEMI;
+	if(d == ';')
+	    return DSEMI;
+	else if(d == '&')
+	    return SEMIAMP;
+	hungetc(d);
+	lexstop = 0;
+	return SEMI;
     case LX1_AMPER:
 	d = hgetc();
 	if (d == '&')
@@ -487,7 +515,6 @@ gettok(void)
 	    }
 	    hungetc(d);
 	    lexstop = 0;
-	    tokfd = -1;
 	    return AMPOUTANG;
 	}
 	hungetc(d);
@@ -504,10 +531,16 @@ gettok(void)
 	return BAR;
     case LX1_INPAR:
 	d = hgetc();
-	if (d == '(' && incmdpos) {
-	    len = 0;
-	    bptr = tokstr = (char *)ncalloc(bsiz = 256);
-	    return cmd_or_math(CS_MATH) ? DINPAR : INPAR;
+	if (d == '(') {
+	    if (infor) {
+		dbparens = 1;
+		return DINPAR;
+	    }
+	    if (incmdpos) {
+		len = 0;
+		bptr = tokstr = (char *)ncalloc(bsiz = 256);
+		return cmd_or_math(CS_MATH) ? DINPAR : INPAR;
+	    }
 	} else if (d == ')')
 	    return INOUTPAR;
 	hungetc(d);
@@ -639,12 +672,12 @@ gettok(void)
 }
 
 /**/
-int
+static int
 gettokstr(int c, int sub)
 {
-    int bct = 0, pct = 0, brct = 0, fdpar = 0;
+    int bct = 0, pct = 0, brct = 0;
     int intpos = 1, in_brace_param = 0;
-    int peek, inquote, unmatched = 0;
+    int peek, inquote;
 #ifdef DEBUG
     int ocmdsp = cmdsp;
 #endif
@@ -657,12 +690,8 @@ gettokstr(int c, int sub)
     for (;;) {
 	int act;
 	int e;
-	int inbl = inblank(c);
-	
-	if (fdpar && !inbl && c != ')')
-	    fdpar = 0;
 
-	if (inbl && !in_brace_param && !pct)
+	if (inblank(c) && !in_brace_param && !pct)
 	    act = LX2_BREAK;
 	else {
 	    act = lexact2[STOUC(c)];
@@ -685,30 +714,22 @@ gettokstr(int c, int sub)
 	    add(Meta);
 	    break;
 	case LX2_OUTPAR:
-	    if (fdpar) {
-		/* this is a single word `(   )', treat as INOUTPAR */
-		add(c);
-		*bptr = '\0';
-		return INOUTPAR;
-	    }
 	    if ((sub || in_brace_param) && isset(SHGLOB))
 		break;
-	    if (!in_brace_param && !pct--) {
+	    if (!in_brace_param && !pct--)
 		if (sub) {
 		    pct = 0;
 		    break;
 		} else
 		    goto brk;
-	    }
 	    c = Outpar;
 	    break;
 	case LX2_BAR:
-	    if (!pct && !in_brace_param) {
+	    if (!pct && !in_brace_param)
 		if (sub)
 		    break;
 		else
 		    goto brk;
-	    }
 	    if (unset(SHGLOB) || (!sub && !in_brace_param))
 		c = Bar;
 	    break;
@@ -779,40 +800,11 @@ gettokstr(int c, int sub)
 		    e = hgetc();
 		    hungetc(e);
 		    lexstop = 0;
-		    /* For command words, parentheses are only
-		     * special at the start.  But now we're tokenising
-		     * the remaining string.  So I don't see what
-		     * the old incmdpos test here is for.
-		     *   pws 1999/6/8
-		     *
-		     * Oh, no.
-		     *  func1(   )
-		     * is a valid function definition in [k]sh.  The best
-		     * thing we can do, without really nasty lookahead tricks,
-		     * is break if we find a blank after a parenthesis.  At
-		     * least this can't happen inside braces or brackets.  We
-		     * only allow this with SHGLOB (set for both sh and ksh).
-		     *
-		     * Things like `print @( |foo)' should still
-		     * work, because [k]sh don't allow multiple words
-		     * in a function definition, so we only do this
-		     * in command position.
-		     *   pws 1999/6/14
-		     */
-		    if (e == ')' || (isset(SHGLOB) && inblank(e) && !bct &&
-				     !brct && !intpos && incmdpos))
+		    if (e == ')' ||
+			(incmdpos && !brct && peek != ENVSTRING))
 			goto brk;
 		}
-		/*
-		 * This also handles the [k]sh `foo( )' function definition.
-		 * Maintain a variable fdpar, set as long as a single set of
-		 * parentheses contains only space.  Then if we get to the
-		 * closing parenthesis and it is still set, we can assume we
-		 * have a function definition.  Only do this at the start of
-		 * the word, since the (...) must be a separate token.
-		 */
-		if (!pct++ && isset(SHGLOB) && intpos && !bct && !brct)
-		    fdpar = 1;
+		pct++;
 	    }
 	    c = Inpar;
 	    break;
@@ -825,9 +817,8 @@ gettokstr(int c, int sub)
 		    *bptr = '\0';
 		    return STRING;
 		}
-		if (in_brace_param) {
+		if (in_brace_param)
 		    cmdpush(CS_BRACE);
-		}
 		bct++;
 	    }
 	    break;
@@ -836,9 +827,8 @@ gettokstr(int c, int sub)
 		break;
 	    if (!bct)
 		break;
-	    if (in_brace_param) {
+	    if (in_brace_param)
 		cmdpop();
-	    }
 	    if (bct-- == in_brace_param)
 		in_brace_param = 0;
 	    c = Outbrace;
@@ -848,12 +838,11 @@ gettokstr(int c, int sub)
 		c = Comma;
 	    break;
 	case LX2_OUTANG:
-	    if (!intpos) {
+	    if (!intpos)
 		if (in_brace_param || sub)
 		    break;
 		else
 		    goto brk;
-	    }
 	    e = hgetc();
 	    if (e != '(') {
 		hungetc(e);
@@ -963,7 +952,7 @@ gettokstr(int c, int sub)
 		}
 		ALLOWHIST
 		if (c != '\'') {
-		    unmatched = '\'';
+		    zerr("unmatched \'", NULL, 0);
 		    peek = LEXERR;
 		    cmdpop();
 		    goto brk;
@@ -984,7 +973,7 @@ gettokstr(int c, int sub)
 	    c = dquote_parse('"', sub);
 	    cmdpop();
 	    if (c) {
-		unmatched = '"';
+		zerr("unmatched \"", NULL, 0);
 		peek = LEXERR;
 		goto brk;
 	    }
@@ -995,7 +984,7 @@ gettokstr(int c, int sub)
 	    cmdpush(CS_BQUOTE);
 	    SETPARBEGIN
 	    inquote = 0;
-	    while ((c = hgetc()) != '`' && !lexstop) {
+	    while ((c = hgetc()) != '`' && !lexstop)
 		if (c == '\\') {
 		    c = hgetc();
 		    if (c != '\n') {
@@ -1009,19 +998,17 @@ gettokstr(int c, int sub)
 			break;
 		    }
 		    add(c);
-		    if (c == '\'') {
+		    if (c == '\'')
 			if ((inquote = !inquote))
 			    STOPHIST
 			else
 			    ALLOWHIST
-		    }
 		}
-	    }
 	    if (inquote)
 		ALLOWHIST
 	    cmdpop();
 	    if (c != '`') {
-		unmatched = '`';
+		zerr("unmatched `", NULL, 0);
 		peek = LEXERR;
 		goto brk;
 	    }
@@ -1038,8 +1025,6 @@ gettokstr(int c, int sub)
     }
   brk:
     hungetc(c);
-    if (unmatched)
-	zerr("unmatched %c", NULL, unmatched);
     if (in_brace_param) {
 	while(bct-- >= in_brace_param)
 	    cmdpop();
@@ -1060,7 +1045,7 @@ gettokstr(int c, int sub)
 extern int addedx;
 
 /**/
-int
+static int
 dquote_parse(char endchar, int sub)
 {
     int pct = 0, brct = 0, bct = 0, intick = 0, err = 0;
@@ -1174,9 +1159,8 @@ dquote_parse(char endchar, int sub)
     }
     if (intick == 2)
 	ALLOWHIST
-    if (intick) {
+    if (intick)
 	cmdpop();
-    }
     while (bct--)
 	cmdpop();
     if (lexstop)
@@ -1339,7 +1323,7 @@ exalias(void)
 /* skip (...) */
 
 /**/
-int
+static int
 skipcomm(void)
 {
     int pct = 1, c;
