@@ -1,11 +1,11 @@
 /*
- * $Id: linklist.c,v 2.3 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: linklist.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
  *
  * linklist.c - linked lists
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "linklist.pro"
 
 /* Get an empty linked list header */
 
@@ -48,20 +49,38 @@ newlinklist(void)
 /* Insert a node in a linked list after a given node */
 
 /**/
-void
+LinkNode
 insertlinknode(LinkList list, LinkNode node, void *dat)
 {
-    LinkNode tmp;
+    LinkNode tmp, new;
 
     tmp = node->next;
-    node->next = (LinkNode) alloc(sizeof *tmp);
-    node->next->last = node;
-    node->next->dat = dat;
-    node->next->next = tmp;
+    node->next = new = (LinkNode) alloc(sizeof *tmp);
+    new->last = node;
+    new->dat = dat;
+    new->next = tmp;
     if (tmp)
-	tmp->last = node->next;
+	tmp->last = new;
     else
-	list->last = node->next;
+	list->last = new;
+    return new;
+}
+
+/* Insert an already-existing node into a linked list after a given node */
+
+/**/
+LinkNode
+uinsertlinknode(LinkList list, LinkNode node, LinkNode new)
+{
+    LinkNode tmp = node->next;
+    node->next = new;
+    new->last = node;
+    new->next = tmp;
+    if (tmp)
+	tmp->last = new;
+    else
+	list->last = new;
+    return new;
 }
 
 /* Insert a list in another list */
