@@ -1,9 +1,11 @@
 /*
+ * $Id: loop.c,v 3.1.2.4 1997/06/01 06:13:15 hzoli Exp $
+ *
  * loop.c - loop execution
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -28,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "loop.pro"
 
 /**/
 int
@@ -36,11 +39,19 @@ execfor(Cmd cmd)
     List list;
     Forcmd node;
     char *str;
+    int val;
     LinkList args;
 
     node = cmd->u.forcmd;
     args = cmd->args;
-    if (!node->inflag) {
+    if (node->condition) {
+	str = node->name;
+	singsub(&str);
+	if (!errflag)
+	    matheval(str);
+	if (errflag)
+	    return lastval = errflag;
+    } else if (!node->inflag) {
 	char **x;
 
 	args = newlinklist();
@@ -50,8 +61,32 @@ execfor(Cmd cmd)
     lastval = 0;
     loops++;
     pushheap();
-    while ((str = (char *)ugetnode(args))) {
-	setsparam(node->name, ztrdup(str));
+    for (;;) {
+	if (node->condition) {
+	    str = dupstring(node->condition);
+	    singsub(&str);
+	    if (!errflag) {
+		while (iblank(*str))
+		    str++;
+		if (*str)
+		    val = matheval(str);
+		else
+		    val = 1;
+	    }
+	    if (errflag) {
+		if (breaks)
+		    breaks--;
+		lastval = 1;
+		break;
+	    }
+	    if (!val)
+		break;
+	} else {
+	    str = (char *) ugetnode(args);
+	    if (!str)
+		break;
+	    setsparam(node->name, ztrdup(str));
+	}
 	list = (List) dupstruct(node->list);
 	execlist(list, 1, (cmd->flags & CFLAG_EXEC) && empty(args));
 	if (breaks) {
@@ -60,7 +95,15 @@ execfor(Cmd cmd)
 		break;
 	    contflag = 0;
 	}
+	if (node->condition && !errflag) {
+	    str = dupstring(node->advance);
+	    singsub(&str);
+	    if (!errflag)
+		matheval(str);
+	}
 	if (errflag) {
+	    if (breaks)
+		breaks--;
 	    lastval = 1;
 	    break;
 	}
@@ -80,7 +123,7 @@ execselect(Cmd cmd)
     char *str, *s;
     LinkList args;
     LinkNode n;
-    int i, usezle;
+    int i;
     FILE *inp;
 
     node = cmd->u.forcmd;
@@ -97,21 +140,14 @@ execselect(Cmd cmd)
     loops++;
     lastval = 0;
     pushheap();
-    usezle = interact && SHTTY != -1 && isset(USEZLE);
-    inp = fdopen(dup(usezle ? SHTTY : 0), "r");
+    inp = fdopen(dup((SHTTY == -1) ? 0 : SHTTY), "r");
     for (;;) {
 	do {
 	    selectlist(args);
 	    if (empty(bufstack)) {
-	    	if (usezle) {
-		    int oef = errflag;
-
-		    isfirstln = 1;
-		    str = (char *)zleread(prompt3, NULL);
-		    if (errflag)
-			str = NULL;
-		    errflag = oef;
-	    	} else {
+	    	if (interact && SHTTY != -1 && isset(USEZLE))
+		    str = (char *)zleread(prompt3, NULL, 0);
+	    	else {
 		    int pptlen;
 		    str = putprompt(prompt3, &pptlen, NULL, 1);
 		    fwrite(str, pptlen, 1, stderr);
@@ -163,6 +199,65 @@ execselect(Cmd cmd)
     return lastval;
 }
 
+/* And this is used to print select lists. */
+
+/**/
+static void
+selectlist(LinkList l)
+{
+    size_t longest = 1, fct, fw = 0, colsz, t0, t1, ct;
+    LinkNode n;
+    char **arr, **ap;
+
+    trashzle();
+    ct = countlinknodes(l);
+    ap = arr = (char **)alloc((countlinknodes(l) + 1) * sizeof(char **));
+
+    for (n = (LinkNode) firstnode(l); n; incnode(n))
+	*ap++ = (char *)getdata(n);
+    *ap = NULL;
+    for (ap = arr; *ap; ap++)
+	if (strlen(*ap) > longest)
+	    longest = strlen(*ap);
+    t0 = ct;
+    longest++;
+    while (t0)
+	t0 /= 10, longest++;
+    /* to compensate for added ')' */
+    fct = (columns - 1) / (longest + 3);
+    if (fct == 0)
+	fct = 1;
+    else
+	fw = (columns - 1) / fct;
+    colsz = (ct + fct - 1) / fct;
+    for (t1 = 0; t1 != colsz; t1++) {
+	ap = arr + t1;
+	do {
+	    int t2 = strlen(*ap) + 2, t3;
+
+	    fprintf(stderr, "%d) %s", t3 = ap - arr + 1, *ap);
+	    while (t3)
+		t2++, t3 /= 10;
+	    for (; t2 < fw; t2++)
+		fputc(' ', stderr);
+	    for (t0 = colsz; t0 && *ap; t0--, ap++);
+	}
+	while (*ap);
+	fputc('\n', stderr);
+    }
+
+ /* Below is a simple attempt at doing it the Korn Way..
+       ap = arr;
+       t0 = 0;
+       do {
+           t0++;
+           fprintf(stderr,"%d) %s\n",t0,*ap);
+           ap++;
+       }
+       while (*ap);*/
+    fflush(stderr);
+}
+
 /**/
 int
 execwhile(Cmd cmd)
@@ -222,7 +317,7 @@ execrepeat(Cmd cmd)
     count = atoi(peekfirst(cmd->args));
     pushheap();
     loops++;
-    while (count-- > 0) {
+    while (count--) {
 	list = (List) dupstruct(cmd->u.list);
 	execlist(list, 1, 0);
 	freeheap();
@@ -294,14 +389,17 @@ execcase(Cmd cmd)
 
     if (node) {
 	while (*p) {
-	    singsub(p);
-	    if (matchpat(word, *p))
+	    char *pat = *p + 1;
+	    singsub(&pat);
+	    if (matchpat(word, pat)) {
+		do {
+		    execlist(*l++, 1, **p == ';' && (cmd->flags & CFLAG_EXEC));
+		} while(**p++ == '&' && *p);
 		break;
+	    }
 	    p++;
 	    l++;
 	}
-	if (*l)
-	    execlist(*l, 1, cmd->flags & CFLAG_EXEC);
     }
     return lastval;
 }
