@@ -1,11 +1,11 @@
 /*
- * $Id: math.c,v 2.11 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: math.c,v 3.1.2.1 1997/06/01 06:13:15 hzoli Exp $
  *
  * math.c - mathematical expression evaluation
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "math.pro"
 
 static char *ptr;
 
@@ -148,7 +149,7 @@ static char **lvals;
 
 
 /**/
-int
+static int
 zzlex(void)
 {
     int cct = 0;
@@ -329,7 +330,6 @@ zzlex(void)
 	    break;
 	case '0':
 	    if (*ptr == 'x' || *ptr == 'X') {
-		unary = 0;
 		/* Should we set lastbase here? */
 		yyval = zstrtol(++ptr, &ptr, lastbase = 16);
 		return NUM;
@@ -404,8 +404,7 @@ struct mathvalue {
 
 static struct mathvalue *stack;
 
-static void push _((long val, LV lval));
-
+/**/
 static void
 push(long val, LV lval)
 {
@@ -419,7 +418,7 @@ push(long val, LV lval)
 
 
 /**/
-long
+static long
 getcvar(LV s)
 {
     char *t;
@@ -431,7 +430,7 @@ getcvar(LV s)
 
 
 /**/
-long
+static long
 setvar(LV s, long v)
 {
     if (s == -1 || s >= lvc) {
@@ -446,7 +445,7 @@ setvar(LV s, long v)
 
 
 /**/
-int
+static int
 notzero(long a)
 {
     if (a == 0) {
@@ -697,7 +696,7 @@ op(int what)
 
 
 /**/
-void
+static void
 bop(int tk)
 {
     switch (tk) {
@@ -716,7 +715,7 @@ bop(int tk)
 
 
 /**/
-long
+static long
 mathevall(char *s, int prek, char **ep)
 {
     int t0;
@@ -815,7 +814,7 @@ mathevalarg(char *s, char **ss)
 /* operator-precedence parse the string and execute */
 
 /**/
-void
+static void
 mathparse(int pc)
 {
     int q, otok, onoeval;
