@@ -1,9 +1,11 @@
 /*
+ * $Id: mem.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
+ *
  * mem.c - memory management
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -28,12 +30,9 @@
  */
 
 #include "zsh.h"
-
-void *(*alloc) _((size_t));
-void *(*ncalloc) _((size_t));
+#include "mem.pro"
 
 /*
-
 	There are two ways to allocate memory in zsh.  The first way is
 	to call zalloc/zcalloc, which call malloc/calloc directly.  It
 	is legal to call realloc() or free() on memory allocated this way.
@@ -47,7 +46,7 @@ void *(*ncalloc) _((size_t));
 	pushheap() saves the states of all currently allocated heaps and
 	popheap() resets them to the last state saved and destroys the
 	information about that state.  If you called pushheap() and
-	allocated some memory on the heaps and then come to a place where
+	allocated some meory on the heaps and then come to a place where
 	you don't need the allocated memory anymore but you still want
 	to allocate memory on the heap, you should call freeheap().  This
 	works like popheap(), only that it doesn't free the information
@@ -70,10 +69,10 @@ void *(*ncalloc) _((size_t));
 	If we use zsh's own allocator we use a simple trick to avoid that
 	the (*real*) heap fills up with empty zsh-heaps: we allocate a
 	large block of memory before allocating a heap pool, this memory
-	is freed again immediately after the pool is allocated. If there
-	are only small blocks on the free list this guarantees that the
+	is freed again immediatly after the pool is allocated. If there
+	are only small blocks on the free list this guarentees that the
 	memory for the pool is at the end of the memory which means that
-	we can give it back to the system when the pool is freed.
+	we can give it back to the systems when the pool is freed.
 */
 
 #ifdef ZSH_MEM_WARNING
@@ -88,7 +87,7 @@ int h_m[1025], h_push, h_pop, h_free;
 
 #endif
 
-#define H_ISIZE  sizeof(zlong)
+#define H_ISIZE  sizeof(long)
 #define HEAPSIZE (8192 - H_ISIZE)
 #define HEAP_ARENA_SIZE (HEAPSIZE - sizeof(struct heap))
 #define HEAPFREE (16384 - H_ISIZE)
@@ -134,9 +133,6 @@ struct heap {
     struct heap *next;		/* next one                                  */
     size_t used;		/* bytes used from the heap                  */
     struct heapstack *sp;	/* used by pushheap() to save the value used */
-#ifdef ZSH_64_BIT_TYPE
-    size_t dummy;		/* Make sure sizeof(heap) is a multiple of 8 */
-#endif
 #define arena(X)	((char *) (X) + sizeof(struct heap))
 };
 
@@ -158,6 +154,7 @@ pushheap(void)
 #endif
 
     for (h = heaps; h; h = h->next) {
+	DPUTS(!h->used, "BUG: empty heap");
 	hs = (Heapstack) zalloc(sizeof(*hs));
 	hs->next = h->sp;
 	h->sp = hs;
@@ -171,20 +168,26 @@ pushheap(void)
 void
 freeheap(void)
 {
-    Heap h;
+    Heap h, hn, hl = NULL;
 
 #if defined(ZSH_MEM) && defined(ZSH_MEM_DEBUG)
     h_free++;
 #endif
-    for (h = heaps; h; h = h->next) {
+    for (h = heaps; h; h = hn) {
+	hn = h->next;
+	if (h->sp) {
 #ifdef ZSH_MEM_DEBUG
-	if (h->sp)
 	    memset(arena(h) + h->sp->used, 0xff, h->used - h->sp->used);
-	else
-	    memset(arena(h), 0xff, h->used);
 #endif
-	h->used = h->sp ? h->sp->used : 0;
+	    h->used = h->sp->used;
+	    hl = h;
+	} else
+	    zfree(h, HEAPSIZE);
     }
+    if (hl)
+	hl->next = NULL;
+    else
+	heaps = NULL;
 }
 
 /* reset heap to previous state and destroy state information */
@@ -254,10 +257,10 @@ halloc(size_t size)
 #endif
 
 	queue_signals();
-	n = HEAP_ARENA_SIZE > size ? HEAP_ARENA_SIZE : size;
+	n = HEAP_ARENA_SIZE > size ? HEAPSIZE : size + sizeof(*h);
 	for (hp = NULL, h = heaps; h; hp = h, h = h->next);
 
-	h = (Heap) zalloc(n + sizeof(*h));
+	h = (Heap) zalloc(n);
 
 #ifdef ZSH_MEM
 	if (called)
@@ -279,6 +282,76 @@ halloc(size_t size)
     }
 }
 
+/**/
+void *
+hrealloc(char *p, size_t old, size_t new)
+{
+    Heap h, ph;
+
+    old = (old + H_ISIZE - 1) & ~(H_ISIZE - 1);
+    new = (new + H_ISIZE - 1) & ~(H_ISIZE - 1);
+
+    if (old == new)
+	return p;
+    if (!old && !p)
+	return halloc(new);
+
+    /* find the heap with p */
+
+    for (h = heaps, ph = NULL; h; ph = h, h = h->next)
+	if (p >= arena(h) && p < arena(h) + HEAP_ARENA_SIZE)
+	    break;
+
+    DPUTS(!h, "BUG: hrealloc() called for non-heap memory.");
+    DPUTS(h->sp && arena(h) + h->sp->used > p,
+	  "BUG: hrealloc() wants to realloc pushed memory");
+
+    if (p + old < arena(h) + h->used) {
+	if (new > old) {
+	    char *ptr = (char *) halloc(new);
+	    memcpy(ptr, p, old);
+	    return ptr;
+	} else
+	    return new ? p : NULL;
+    }
+
+    DPUTS(p + old != arena(h) + h->used, "BUG: hrealloc more than allocated");
+
+    if (p == arena(h)) {
+	if (!new) {
+	    if (ph)
+		ph->next = h->next;
+	    else
+		heaps = h->next;
+	    zfree(h, HEAPSIZE);
+	    return NULL;
+	}
+	if (old > HEAP_ARENA_SIZE || new > HEAP_ARENA_SIZE) {
+	    size_t n = HEAP_ARENA_SIZE > new ? HEAPSIZE : new + sizeof(*h);
+
+	    if (ph)
+		ph->next = h = (Heap) realloc(h, n);
+	    else
+		heaps = h = (Heap) realloc(h, n);
+	}
+	h->used = new;
+	return arena(h);
+    }
+    DPUTS(h->used > HEAP_ARENA_SIZE, "BUG: hrealloc at invalid address");
+    if (h->used + (new - old) <= HEAP_ARENA_SIZE) {
+	h->used += new - old;
+	return p;
+    } else {
+	char *t = halloc(new);
+	memcpy(t, p, old > new ? new : old);
+	h->used -= old;
+#ifdef ZSH_MEM_DEBUG
+	memset(p, 0xff, old);
+#endif
+	return t;
+    }
+}
+
 /* allocate memory from the current memory pool and clear it */
 
 /**/
@@ -292,17 +365,6 @@ hcalloc(size_t size)
     return ptr;
 }
 
-/**/
-void *
-hrealloc(char *p, int old, int new)
-{
-    char *ptr;
-
-    ptr = (char *)halloc(new);
-    memcpy(ptr, p, old);
-    return (void *) ptr;
-}
-
 /* allocate permanent memory */
 
 /**/
@@ -349,14 +411,9 @@ void *
 zrealloc(void *ptr, size_t size)
 {
     if (ptr) {
-	if (size) {
+	if (size)
 	    /* Do normal realloc */
-	    if (!(ptr = (void *) realloc(ptr, size))) {
-		zerr("fatal error: out of memory", NULL, 0);
-		exit(1);
-	    }
-	    return ptr;
-	}
+	    return realloc(ptr, size);
 	else
 	    /* If ptr is not NULL, but size is zero, *
 	     * then object pointed to is freed.      */
@@ -400,9 +457,8 @@ ztrdup(const char *s)
 /*
    Below is a simple segment oriented memory allocator for systems on
    which it is better than the system's one. Memory is given in blocks
-   aligned to an integer multiple of sizeof(zlong) (4 bytes on most machines,
-   but 8 bytes on e.g. a dec alpha, or if we are using 64-bit integer
-   arithmetic on a 32-bit machine). Each block is preceded by a header
+   aligned to an integer multiple of sizeof(long) (4 bytes on most machines,
+   but 8 bytes on e.g. a dec alpha). Each block is preceded by a header
    which contains the length of the data part (in bytes). In allocated
    blocks only this field of the structure m_hdr is senseful. In free
    blocks the second field (next) is a pointer to the next free segment
@@ -462,26 +518,22 @@ ztrdup(const char *s)
 
 struct m_shdr {
     struct m_shdr *next;	/* next one on free list */
-#ifdef ZSH_64_BIT_TYPE
-    /* dummy to make this 64-bit aligned */
-    struct m_shdr *dummy;
-#endif
 };
 
 struct m_hdr {
-    zlong len;			/* length of memory block */
+    long len;			/* length of memory block */
     struct m_hdr *next;		/* if free: next on free list
 				   if block of small blocks: next one with
 				                 small blocks of same size*/
     struct m_shdr *free;	/* if block of small blocks: free list */
-    zlong used;			/* if block of small blocks: number of used
+    long used;			/* if block of small blocks: number of used
 				                                     blocks */
 };
 
 
 /* alignment for memory blocks */
 
-#define M_ALIGN (sizeof(zlong))
+#define M_ALIGN (sizeof(long))
 
 /* length of memory header, length of first field of memory header and
    minimal size of a block left free (if we allocate memory and take a
@@ -490,7 +542,7 @@ struct m_hdr {
    the free list) */
 
 #define M_HSIZE (sizeof(struct m_hdr))
-#define M_ISIZE (sizeof(zlong))
+#define M_ISIZE (sizeof(long))
 #define M_MIN   (2 * M_ISIZE)
 
 /* a pointer to the last free block, a pointer to the free list (the blocks
@@ -526,9 +578,9 @@ char *m_high, *m_low;
 #define M_SNUM     50
 #define M_SLEN(M)  ((M)->len / M_SNUM)
 #define M_SBLEN(S) ((S) * M_SNUM + sizeof(struct m_shdr *) +  \
-		    sizeof(zlong) + sizeof(struct m_hdr *))
+		    sizeof(long) + sizeof(struct m_hdr *))
 #define M_BSLEN(S) (((S) - sizeof(struct m_shdr *) -  \
-		     sizeof(zlong) - sizeof(struct m_hdr *)) / M_SNUM)
+		     sizeof(long) - sizeof(struct m_hdr *)) / M_SNUM)
 #define M_NSMALL 8
 
 struct m_hdr *m_small[M_NSMALL];
@@ -653,24 +705,15 @@ malloc(MALLOC_ARG_T size)
 	for (mp = NULL, m = m_free; m && m->len < size; mp = m, m = m->next);
     }
     if (!m) {
-	long nal;
 	/* no matching free block was found, we have to request new
 	   memory from the system */
 	n = (size + M_HSIZE + m_pgsz - 1) & ~(m_pgsz - 1);
 
 	if (((char *)(m = (struct m_hdr *)sbrk(n))) == ((char *)-1)) {
-	    DPUTS(1, "allocation error at sbrk.");
+	    DPUTS(1, "MEM: allocation error at sbrk.");
 	    unqueue_signals();
 	    return NULL;
 	}
-	if ((nal = ((long)(char *)m) & (M_ALIGN-1))) {
-	    if ((char *)sbrk(M_ALIGN - nal) == (char *)-1) {
-		DPUTS(1, "MEM: allocation error at sbrk.");
-		unqueue_signals();
-		return NULL;
-	    }
-	    m = (struct m_hdr *) ((char *)m + (M_ALIGN - nal));
-	}
 	/* set m_low, for the check in free() */
 	if (!m_low)
 	    m_low = (char *)m;
@@ -786,7 +829,7 @@ zfree(void *p, int sz)
     /* first a simple check if the given address is valid */
     if (((char *)p) < m_low || ((char *)p) > m_high ||
 	((long)p) & (M_ALIGN - 1)) {
-	DPUTS(1, "attempt to free storage at invalid address");
+	DPUTS(1, "BUG: attempt to free storage at invalid address");
 	return;
     }
 
@@ -818,7 +861,7 @@ zfree(void *p, int sz)
 		if ((((char *)p) - (((char *)mt) + sizeof(struct m_hdr))) %
 		    M_BSLEN(mt->len)) {
 
-		    DPUTS(1, "attempt to free storage at invalid address");
+		    DPUTS(1, "BUG: attempt to free storage at invalid address");
 		    unqueue_signals();
 		    return;
 		}
@@ -826,13 +869,13 @@ zfree(void *p, int sz)
 		for (sh2 = mt->free; sh2; sh2 = sh2->next)
 		    if (((char *)p) == ((char *)sh2)) {
 
-			DPUTS(1, "attempt to free already free storage");
+			DPUTS(1, "BUG: attempt to free already free storage");
 			unqueue_signals();
 			return;
 		    }
 #endif
 		DPUTS(M_BSLEN(mt->len) < osz,
-		      "attempt to free more than allocated.");
+		      "BUG: attempt to free more than allocated.");
 
 #ifdef ZSH_MEM_DEBUG
 		m_f[M_BSLEN(mt->len) / M_ISIZE]++;
@@ -892,7 +935,7 @@ zfree(void *p, int sz)
 
     /* no block was found at the given address */
     if (((char *)mt) >= m_high) {
-	DPUTS(1, "attempt to free storage at invalid address");
+	DPUTS(1, "BUG: attempt to free storage at invalid address");
 	unqueue_signals();
 	return;
     }
@@ -903,11 +946,11 @@ zfree(void *p, int sz)
 
     if (m == mt) {
 	/* it is, ouch! */
-	DPUTS(1, "attempt to free already free storage");
+	DPUTS(1, "BUG: attempt to free already free storage");
 	unqueue_signals();
 	return;
     }
-    DPUTS(m->len < osz, "attempt to free more than allocated.");
+    DPUTS(m->len < osz, "BUG: attempt to free more than allocated");
 #ifdef ZSH_MEM_DEBUG
     memset(p, 0xff, m->len);
 #endif
@@ -948,7 +991,7 @@ zfree(void *p, int sz)
 
 	m_lfree->len -= n;
 	if (brk(m_high -= n) == -1)
-	    DPUTS(1, "allocation error at brk.");
+	    DPUTS(1, "MEM: allocation error at brk.");
 
 #ifdef ZSH_MEM_DEBUG
 	m_b += n;
@@ -1111,9 +1154,8 @@ bin_mem(char *name, char **argv, char *ops, int func)
 	if (m == mf)
 	    buf[0] = '\0';
 	else if (m == ms)
-	    sprintf(buf, "%ld %ld %ld", (long)(M_SNUM - ms->used),
-		    (long)ms->used,
-		    (long)((m->len - sizeof(struct m_hdr)) / M_SNUM + 1));
+	    sprintf(buf, "%ld %ld %ld", M_SNUM - ms->used, ms->used,
+		    (m->len - sizeof(struct m_hdr)) / M_SNUM + 1);
 
 	else {
 	    for (i = 0, b = buf, c = (char *)&m->next; i < 20 && i < m->len;
@@ -1124,7 +1166,7 @@ bin_mem(char *name, char **argv, char *ops, int func)
 
 	printf("%d\t%d\t%ld\t%ld\t%s\t%ld\t%s\n", ii,
 	       (m == mf) ? fi++ : ui++,
-	       (long)m, (long)m->len,
+	       (long)m, m->len,
 	       (m == mf) ? "free" : ((m == ms) ? "small" : "used"),
 	       (m == mf) ? (f += m->len) : (u += m->len),
 	       buf);
@@ -1145,7 +1187,7 @@ bin_mem(char *name, char **argv, char *ops, int func)
 	    printf("%ld\t", (long)i * M_ISIZE);
 
 	    for (ii = 0, m = m_small[i]; m; m = m->next) {
-		printf("(%ld/%ld) ", (long)(M_SNUM - m->used), (long)m->used);
+		printf("(%ld/%ld) ", M_SNUM - m->used, m->used);
 		if (!((++ii) & 7))
 		    printf("\n\t");
 	    }
