@@ -1,11 +1,11 @@
 /*
- * $Id: params.c,v 2.41 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: params.c,v 3.1.2.13 1997/06/03 06:00:23 hzoli Exp $
  *
  * params.c - parameters
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -31,6 +31,9 @@
 
 #include "zsh.h"
 
+#include "version.h"
+#include "params.pro"
+
 static Param argvparam;
 
 /* Set up parameter hash table.  This will add predefined  *
@@ -42,11 +45,11 @@ void
 createparamtable(void)
 {
     Param ip, pm;
-    char **new_environ, **envp, **envp2, **sigptr, **t;
+    char **old_environ, **envp, **envp2, **sigptr, **t;
     char buf[50], *str, *iname;
     int num_env;
 
-    paramtab = newhashtable(151);
+    paramtab = newhashtable(151, "paramtab", NULL);
 
     paramtab->hash        = hasher;
     paramtab->emptytable  = NULL;
@@ -59,10 +62,6 @@ createparamtable(void)
     paramtab->enablenode  = NULL;
     paramtab->freenode    = freeparamnode;
     paramtab->printnode   = printparamnode;
-#ifdef ZSH_HASH_DEBUG
-    paramtab->printinfo   = printhashtabinfo;
-    paramtab->tablename   = ztrdup("paramtab");
-#endif
 
     /* Add the special parameters to the hash table */
     for (ip = special_params; ip->nam; ip++)
@@ -76,10 +75,6 @@ createparamtable(void)
     noerrs = 1;
 
     HEAPALLOC {
-	int allexp = opts[ALLEXPORT];
-
-	opts[ALLEXPORT] = 0;
-
 	/* Add the standard non-special parameters which have to    *
 	 * be initialized before we copy the environment variables. *
 	 * We don't want to override whatever values the users has  *
@@ -100,35 +95,35 @@ createparamtable(void)
 
 	/* Copy the environment variables we are inheriting to dynamic *
 	 * memory, so we can do mallocs and frees on it.               */
-	num_env = arrlen(environ);
-	new_environ = (char **) zalloc(sizeof(char *) * (num_env + 1));
-	*new_environ = NULL;
+	old_environ = environ;
+	num_env = arrlen(old_environ);
+	environ = (char **) zalloc(sizeof(char *) * (num_env + 1));
+	for (envp = environ, envp2 = old_environ; *envp2; envp2++)
+	    *envp++ = ztrdup(*envp2);
+	*envp = NULL;
 
 	/* Now incorporate environment variables we are inheriting *
 	 * into the parameter hash table.                          */
-	for (envp = new_environ, envp2 = environ; *envp2; envp2++) {
+	for (envp = environ, envp2 = old_environ; *envp2; envp++, envp2++) {
 	    for (str = *envp2; *str && *str != '='; str++);
 	    if (*str == '=') {
-		iname = ztrdup(*envp2);
-		str = iname + (str - *envp2);
+		iname = NULL;
 		*str = '\0';
-		if (!idigit(*iname) && isident(iname) && !strchr(iname, '[') &&
-		    !((pm = (Param) paramtab->getnode(paramtab, iname)) &&
-		      (pm->flags & (PM_DONTIMPORT | PM_EXPORTED))) &&
-		    (pm = setsparam(iname, metafy(str + 1, -1, META_DUP)))) {
-		    *str = '=';
-		    pm->flags |= PM_EXPORTED;
-		    if (pm->flags & PM_SPECIAL) {
-			zsfree(iname);
-			iname = mkenvstr(pm->nam, getsparam(pm->nam));
-		    }
-		    pm->env = *envp++ = iname;
-		    *envp = NULL;
-		} else
-		    zsfree(iname);
+		if (!idigit(**envp2) && isident(*envp2) && !strchr(*envp2, '[')) {
+		    iname = *envp2;
+		    if ((!(pm = (Param) paramtab->getnode(paramtab, iname)) ||
+			 !(pm->flags & PM_DONTIMPORT)) &&
+			(pm = setsparam(iname, metafy(str + 1, -1, META_DUP)))) {
+			pm->flags |= PM_EXPORTED;
+			pm->env = *envp;
+			if (pm->flags & PM_SPECIAL)
+			    pm->env = replenv(pm->env, getsparam(iname));
+		    } else if (pm)
+			for (t = envp--; (t[0] = t[1]); t++);
+		}
+		*str = '=';
 	    }
 	}
-	environ = new_environ;
 
 	pm = (Param) paramtab->getnode(paramtab, "HOME");
 	if (!(pm->flags & PM_EXPORTED)) {
@@ -160,41 +155,41 @@ createparamtable(void)
 	setsparam("ZSH_VERSION", ztrdup(ZSH_VERSION));
 	setaparam("signals", sigptr = zalloc((SIGCOUNT+4) * sizeof(char *)));
 	for (t = sigs; (*sigptr++ = ztrdup(*t++)); );
-	opts[ALLEXPORT] = allexp;
     } LASTALLOC;
 
     noerrs = 0;
 }
 
-/* Create a new parameter node */
+/* Create a parameter, so that it can be assigned to.  Returns NULL if the *
+ * parameter already exists or can't be created, otherwise returns the     *
+ * parameter node.  If a parameter of the same name exists in an outer     *
+ * scope, it is hidden by a newly created parameter.  An already existing  *
+ * parameter node at the current level may be `created' and returned       *
+ * provided it is unset and not special.  If the parameter can't be        *
+ * created because it already exists, the PM_UNSET flag is cleared.        */
 
 /**/
 Param
 createparam(char *name, int flags)
 {
-    Param pm, oldpm, altpm;
-    int spec;
+    Param pm, oldpm;
 
     if (name != nulstring) {
 	oldpm = (Param) paramtab->getnode(paramtab, name);
-	spec = oldpm && (oldpm->flags & PM_SPECIAL);
 
-	if ((oldpm && oldpm->level == locallevel) || spec) {
-	    if (oldpm && !(oldpm->flags & PM_UNSET))
+	if (oldpm && oldpm->level == locallevel) {
+	    if (!(oldpm->flags & PM_UNSET) || (oldpm->flags & PM_SPECIAL)) {
+		oldpm->flags &= ~PM_UNSET;
 		return NULL;
+	    }
+	    if ((oldpm->flags & PM_RESTRICTED) && isset(RESTRICTED)) {
+		zerr("%s: restricted", name, 0);
+		return NULL;
+	    }
+
 	    pm = oldpm;
 	    pm->ct = 0;
 	    oldpm = pm->old;
-	    pm->flags = (flags & (PM_EXPORTED | PM_LEFT | PM_RIGHT_B |
-				  PM_RIGHT_Z | PM_LOWER | PM_UPPER |
-				  PM_READONLY | PM_TAGGED | PM_UNIQUE)) |
-		(pm->flags & (PM_SCALAR | PM_INTEGER | PM_ARRAY | PM_SPECIAL));
-	    if (pm->ename &&
-		(altpm = (Param) paramtab->getnode(paramtab, pm->ename))) {
-		altpm->flags &= ~(PM_UNSET | PM_UNIQUE | PM_UPPER | PM_LEFT |
-				  PM_RIGHT_B | PM_RIGHT_Z | PM_LOWER |
-				  PM_READONLY | PM_TAGGED | PM_EXPORTED);
-	    }
 	} else {
 	    pm = (Param) zcalloc(sizeof *pm);
 	    if ((pm->old = oldpm)) {
@@ -205,14 +200,12 @@ createparam(char *name, int flags)
 	}
 
 	if (isset(ALLEXPORT) && !oldpm)
-	    pm->flags |= PM_EXPORTED;
-    } else {
+	    flags |= PM_EXPORTED;
+    } else
 	pm = (Param) alloc(sizeof *pm);
-	spec = 0;
-    }
+    pm->flags = flags;
 
-    if (!spec) {
-	pm->flags = flags;
+    if(!(pm->flags & PM_SPECIAL)) {
 	switch (PM_TYPE(flags)) {
 	case PM_SCALAR:
 	    pm->sets.cfn = strsetfn;
@@ -226,12 +219,11 @@ createparam(char *name, int flags)
 	    pm->sets.afn = arrsetfn;
 	    pm->gets.afn = arrgetfn;
 	    break;
-#ifdef DEBUG
 	default:
-	    DPUTS(1, "oops, tried to create param node without valid flag");
+	    DPUTS(1, "BUG: tried to create param node without valid flag");
 	    break;
-#endif
 	}
+	pm->unsetfn = stdunsetfn;
     }
     return pm;
 }
@@ -276,7 +268,7 @@ isident(char *s)
 char **garr;
 
 /**/
-long
+static long
 getarg(char **str, int *inv, Value v, int a2, long *w)
 {
     int num = 1, word = 0, rev = 0, ind = 0, down = 0, l, i;
@@ -709,7 +701,7 @@ getstrvalue(Value v)
 	    break;
 	default:
 	    s = NULL;
-	    DPUTS(1, "oops, param node without valid type");
+	    DPUTS(1, "BUG: param node without valid type");
 	    break;
 	}
 
@@ -778,7 +770,7 @@ getintvalue(Value v)
 }
 
 /**/
-void
+static void
 setstrvalue(Value v, char *val)
 {
     char buf[(SIZEOF_LONG * 8) + 4];
@@ -788,6 +780,11 @@ setstrvalue(Value v, char *val)
 	zsfree(val);
 	return;
     }
+    if ((v->pm->flags & PM_RESTRICTED) && isset(RESTRICTED)) {
+	zerr("%s: restricted", v->pm->nam, 0);
+	zsfree(val);
+	return;
+    }
     switch (PM_TYPE(v->pm->flags)) {
     case PM_SCALAR:
 	MUSTUSEHEAP("setstrvalue");
@@ -858,7 +855,7 @@ setstrvalue(Value v, char *val)
 }
 
 /**/
-void
+static void
 setintvalue(Value v, long val)
 {
     char buf[DIGBUFSIZE];
@@ -867,6 +864,10 @@ setintvalue(Value v, long val)
 	zerr("read-only variable: %s", v->pm->nam, 0);
 	return;
     }
+    if ((v->pm->flags & PM_RESTRICTED) && isset(RESTRICTED)) {
+	zerr("%s: restricted", v->pm->nam, 0);
+	return;
+    }
     switch (PM_TYPE(v->pm->flags)) {
     case PM_SCALAR:
     case PM_ARRAY:
@@ -881,7 +882,7 @@ setintvalue(Value v, long val)
 }
 
 /**/
-void
+static void
 setarrvalue(Value v, char **val)
 {
     if (v->pm->flags & PM_READONLY) {
@@ -889,6 +890,11 @@ setarrvalue(Value v, char **val)
 	freearray(val);
 	return;
     }
+    if ((v->pm->flags & PM_RESTRICTED) && isset(RESTRICTED)) {
+	zerr("%s: restricted", v->pm->nam, 0);
+	freearray(val);
+	return;
+    }
     if (PM_TYPE(v->pm->flags) != PM_ARRAY) {
 	freearray(val);
 	zerr("attempt to assign array value to non-array", NULL, 0);
@@ -1036,9 +1042,8 @@ setaparam(char *s, char **val)
 	    createparam(t, PM_ARRAY);
 	else if (PM_TYPE(v->pm->flags) != PM_ARRAY &&
 		 !(v->pm->flags & PM_SPECIAL)) {
-	    int uniq = v->pm->flags & PM_UNIQUE;
 	    unsetparam(t);
-	    createparam(t, PM_ARRAY | uniq);
+	    createparam(t, PM_ARRAY);
 	    v = NULL;
 	}
     }
@@ -1084,31 +1089,26 @@ unsetparam(char *s)
     Param pm;
 
     if ((pm = (Param) paramtab->getnode(paramtab, s)))
-	unsetparam_pm(pm, 0);
+	unsetparam_pm(pm, 0, 1);
 }
 
 /* Unset a parameter */
 
 /**/
 void
-unsetparam_pm(Param pm, int altflag)
+unsetparam_pm(Param pm, int altflag, int exp)
 {
     Param oldpm, altpm;
-    int spec;
 
     if ((pm->flags & PM_READONLY) && pm->level <= locallevel) {
 	zerr("read-only variable: %s", pm->nam, 0);
 	return;
     }
-    spec = (pm->flags & PM_SPECIAL) && !pm->level;
-    switch (PM_TYPE(pm->flags)) {
-    case PM_SCALAR:
-	(pm->sets.cfn) (pm, NULL);
-	break;
-    case PM_ARRAY:
-	(pm->sets.afn) (pm, NULL);
-	break;
+    if ((pm->flags & PM_RESTRICTED) && isset(RESTRICTED)) {
+	zerr("%s: restricted", pm->nam, 0);
+	return;
     }
+    pm->unsetfn(pm, exp);
     if ((pm->flags & PM_EXPORTED) && pm->env) {
 	delenv(pm->env);
 	zsfree(pm->env);
@@ -1118,13 +1118,11 @@ unsetparam_pm(Param pm, int altflag)
     /* remove it under its alternate name if necessary */
     if (pm->ename && !altflag) {
 	altpm = (Param) paramtab->getnode(paramtab, pm->ename);
-	unsetparam_pm(altpm, 1);
+	unsetparam_pm(altpm, 1, exp);
     }
 
-    if ((locallevel && locallevel == pm->level) || spec) {
-	pm->flags |= PM_UNSET;
+    if (locallevel >= pm->level && (pm->flags & PM_SPECIAL))
 	return;
-    }
 
     paramtab->removenode(paramtab, pm->nam); /* remove parameter node from table */
 
@@ -1138,10 +1136,24 @@ unsetparam_pm(Param pm, int altflag)
     paramtab->freenode((HashNode) pm); /* free parameter node */
 }
 
+/* Standard function to unset a parameter.  This is mostly delegated to *
+ * the specific set function.                                           */
+
+/**/
+void
+stdunsetfn(Param pm, int exp)
+{
+    switch (PM_TYPE(pm->flags)) {
+	case PM_SCALAR: pm->sets.cfn(pm, NULL); break;
+	case PM_ARRAY:  pm->sets.afn(pm, NULL); break;
+    }
+    pm->flags |= PM_UNSET;
+}
+
 /* Function to get value of an integer parameter */
 
 /**/
-long
+static long
 intgetfn(Param pm)
 {
     return pm->u.val;
@@ -1150,7 +1162,7 @@ intgetfn(Param pm)
 /* Function to set value of an integer parameter */
 
 /**/
-void
+static void
 intsetfn(Param pm, long x)
 {
     pm->u.val = x;
@@ -1168,7 +1180,7 @@ strgetfn(Param pm)
 /* Function to set value of a scalar (string) parameter */
 
 /**/
-void
+static void
 strsetfn(Param pm, char *x)
 {
     zsfree(pm->u.str);
@@ -1179,7 +1191,7 @@ strsetfn(Param pm, char *x)
 /* Function to get value of an array parameter */
 
 /**/
-char **
+static char **
 arrgetfn(Param pm)
 {
     static char *nullarray = NULL;
@@ -1190,7 +1202,7 @@ arrgetfn(Param pm)
 /* Function to set value of an array parameter */
 
 /**/
-void
+static void
 arrsetfn(Param pm, char **x)
 {
     if (pm->u.arr && pm->u.arr != x)
@@ -1218,7 +1230,7 @@ nullsetfn(Param pm, char *x)
 long
 intvargetfn(Param pm)
 {
-    return *((long *)pm->data);
+    return *((long *)pm->u.data);
 }
 
 /* Function to set value of generic special integer *
@@ -1229,7 +1241,7 @@ intvargetfn(Param pm)
 void
 intvarsetfn(Param pm, long x)
 {
-    *((long *)pm->data) = x;
+    *((long *)pm->u.data) = x;
 }
 
 /* Function to set value of any ZLE-related integer *
@@ -1240,11 +1252,23 @@ intvarsetfn(Param pm, long x)
 void
 zlevarsetfn(Param pm, long x)
 {
-    long *p = (long *)pm->data;
+    if ((long *)pm->u.data == & columns) {
+	if(x <= 0)
+	    x = tccolumns > 0 ? tccolumns : 80;
+	if (x > 2)
+	    termflags &= ~TERM_NARROW;
+	else
+	    termflags |= TERM_NARROW;
+    } else if ((long *)pm->u.data == & lines) {
+	if(x <= 0)
+	    x = tclines > 0 ? tclines : 24;
+	if (x > 2)
+	    termflags &= ~TERM_SHORT;
+	else
+	    termflags |= TERM_SHORT;
+    }
 
-    *p = x;
-    if (p == &lines || p == &columns)
-	adjustwinsize(2 + (p == &columns));
+    *((long *)pm->u.data) = x;
 }
 
 /* Function to set value of generic special scalar    *
@@ -1255,7 +1279,7 @@ zlevarsetfn(Param pm, long x)
 void
 strvarsetfn(Param pm, char *x)
 {
-    char **q = ((char **)pm->data);
+    char **q = ((char **)pm->u.data);
 
     zsfree(*q);
     *q = x;
@@ -1269,7 +1293,7 @@ strvarsetfn(Param pm, char *x)
 char *
 strvargetfn(Param pm)
 {
-    char *s = *((char **)pm->data);
+    char *s = *((char **)pm->u.data);
 
     if (!s)
 	return hcalloc(1);
@@ -1285,7 +1309,7 @@ strvargetfn(Param pm)
 char **
 arrvargetfn(Param pm)
 {
-    return *((char ***)pm->data);
+    return *((char ***)pm->u.data);
 }
 
 /* Function to set value of generic special array parameter.    *
@@ -1298,7 +1322,7 @@ arrvargetfn(Param pm)
 void
 arrvarsetfn(Param pm, char **x)
 {
-    char ***dptr = (char ***)pm->data;
+    char ***dptr = (char ***)pm->u.data;
 
     if (*dptr != x)
 	freearray(*dptr);
@@ -1313,18 +1337,17 @@ arrvarsetfn(Param pm, char **x)
 char *
 colonarrgetfn(Param pm)
 {
-    return zjoin(*(char ***)pm->data, ':');
+    return zjoin(*(char ***)pm->u.data, ':');
 }
 
 /**/
 void
 colonarrsetfn(Param pm, char *x)
 {
-    char ***dptr = (char ***)pm->data;
+    char ***dptr = (char ***)pm->u.data;
 
     freearray(*dptr);
     *dptr = x ? colonsplit(x, pm->flags & PM_UNIQUE) : mkarray(NULL);
-    zsfree(x);
     if (pm->ename)
 	arrfixenv(pm->nam, *dptr);
 }
@@ -1419,12 +1442,10 @@ usernamesetfn(Param pm, char *x)
 #endif
 	if(!setgid(pswd->pw_gid) && !setuid(pswd->pw_uid)) {
 	    zsfree(cached_username);
-	    cached_username = x;
+	    cached_username = ztrdup(pswd->pw_name);
 	    cached_uid = pswd->pw_uid;
-	    return;
 	}
     }
-    zsfree(x);
 #endif
 }
 
@@ -1562,7 +1583,7 @@ static struct localename {
 };
 
 /**/
-void
+static void
 setlang(char *x)
 {
     struct localename *ln;
@@ -1596,12 +1617,17 @@ langsetfn(Param pm, char *x)
 void
 lcsetfn(Param pm, char *x)
 {
+    struct localename *ln;
+
     strsetfn(pm, x);
     if (getsparam("LC_ALL"))
 	return;
     if (!x)
 	x = getsparam("LANG");
-    setlocale((int) pm->data, x ? x : "");
+
+    for (ln = lc_names; ln->name; ln++)
+	if (!strcmp(ln->name, pm->nam))
+	    setlocale(ln->category, x ? x : "");
 }
 #endif
 
@@ -1634,24 +1660,6 @@ errnogetfn(Param pm)
     return errno;
 }
 
-/* Function to get value for special parameter `-' */
-
-/**/
-char *
-dashgetfn(Param pm)
-{
-    static char buf[63];
-    char *val = buf;
-    int optno;
-
-    for(optno = 1; optno < OPT_SIZE; optno++)
-	if(optid(optns[optno]) &&
-		opts[optno] == !(optid(optns[optno]) & OPT_REV))
-	    *val++ = optid(optns[optno]) & ~OPT_REV;
-    *val = '\0';
-    return buf;
-}
-
 /* Function to get value for special parameter `histchar' */
 
 /**/
@@ -1767,7 +1775,7 @@ termsetfn(Param pm, char *x)
  * do the replacing, since we've already scanned for the string. */
 
 /**/
-void
+static void
 arrfixenv(char *s, char **t)
 {
     char **ep, *u;
@@ -1780,16 +1788,15 @@ arrfixenv(char *s, char **t)
     u = zjoin(t, ':');
     len_s = strlen(s);
     pm = (Param) paramtab->getnode(paramtab, s);
+    for (ep = environ; *ep; ep++)
+	if (!strncmp(*ep, s, len_s) && (*ep)[len_s] == '=') {
+	    pm->env = replenv(*ep, u);
+	    return;
+	}
     if (isset(ALLEXPORT))
 	pm->flags |= PM_EXPORTED;
-    if (pm->flags & PM_EXPORTED) {
-	for (ep = environ; *ep; ep++)
-	    if (!strncmp(*ep, s, len_s) && (*ep)[len_s] == '=') {
-		pm->env = replenv(*ep, u);
-		return;
-	    }
+    if (pm->flags & PM_EXPORTED)
 	pm->env = addenv(s, u);
-    }
 }
 
 /* Given *name = "foo", it searchs the environment for string *
@@ -1838,7 +1845,7 @@ replenv(char *e, char *value)
  * return a new string *str = "foo=bar".        */
 
 /**/
-char *
+static char *
 mkenvstr(char *name, char *value)
 {
     char *str, *s;
@@ -1908,7 +1915,7 @@ delenv(char *x)
 }
 
 /**/
-void
+static void
 convbase(char *s, long v, int base)
 {
     int digs = 0;
@@ -1936,3 +1943,31 @@ convbase(char *s, long v, int base)
 	x /= base;
     }
 }
+
+/* Start a parameter scope */
+
+/**/
+void
+startparamscope(void)
+{
+    locallevel++;
+}
+
+/* End a parameter scope: delete the parameters local to the scope. */
+
+/**/
+void
+endparamscope(void)
+{
+    locallevel--;
+    scanhashtable(paramtab, 0, 0, 0, scanendscope, 0);
+}
+
+/**/
+static void
+scanendscope(HashNode hn, int flags)
+{
+    Param pm = (Param)hn;
+    if(pm->level > locallevel)
+	unsetparam_pm(pm, 0, 0);
+}
