@@ -1,11 +1,11 @@
 /*
- * $Id: parse.c,v 2.28 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: parse.c,v 3.1.2.4 1997/06/01 06:13:15 hzoli Exp $
  *
  * parse.c - parser
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,10 +30,10 @@
  */
 
 #include "zsh.h"
+#include "parse.pro"
 
 #define YYERROR  { tok = LEXERR; return NULL; }
 #define YYERRORV { tok = LEXERR; return; }
-#define COND_ERROR(X,Y) do{zerr(X,Y,0);discard_input();YYERROR}while(0)
 
 #define make_list()     allocnode(N_LIST)
 #define make_sublist()  allocnode(N_SUBLIST)
@@ -46,23 +46,6 @@
 #define make_varnode()  allocnode(N_VARASG)
 #define make_cond()     allocnode(N_COND)
 
-static void
-discard_input(void)
-{
-    errflag = 0;
-    if (isnewlin <= 0) {
-	/* Discard remaining stuff after a parse error. */
-	int c;
-
-	hwbegin(0);
-	while ((c = hgetc()) != '\n' && !lexstop);
-	if (c == '\n')
-	    hungetc('\n');
-	hwend();
-    }
-    errflag = 1;
-}
-
 /*
  * event	: ENDINPUT
  *			| SEPER
@@ -79,7 +62,7 @@ parse_event(void)
 }
 
 /**/
-List
+static List
 par_event(void)
 {
     Sublist sl;
@@ -121,7 +104,18 @@ par_event(void)
 	    return NULL;
 	}
 	yyerror();
-	discard_input();
+	errflag = 0;
+	if (isnewlin <= 0) {
+	    /* Discard remaining stuff after a parse error. */
+	    int c;
+
+	    hwbegin(0);
+	    while ((c = hgetc()) != '\n' && !lexstop);
+	    if (c == '\n')
+		hungetc('\n');
+	    hwend();
+	}
+	errflag = 1;
 	return NULL;
     } else {
 	l->right = par_event();
@@ -151,7 +145,7 @@ parse_list(void)
  */
 
 /**/
-List
+static List
 par_list(void)
 {
     Sublist sl;
@@ -179,7 +173,7 @@ par_list(void)
 }
 
 /**/
-List
+static List
 par_list1(void)
 {
     Sublist sl;
@@ -198,7 +192,7 @@ par_list1(void)
  */
 
 /**/
-Sublist
+static Sublist
 par_sublist(void)
 {
     Sublist sl;
@@ -223,7 +217,7 @@ par_sublist(void)
  */
 
 /**/
-Sublist
+static Sublist
 par_sublist2(void)
 {
     Sublist sl;
@@ -248,7 +242,7 @@ par_sublist2(void)
  */
 
 /**/
-Pline
+static Pline
 par_pline(void)
 {
     Cmd c;
@@ -299,7 +293,7 @@ par_pline(void)
  */
 
 /**/
-Cmd
+static Cmd
 par_cmd(void)
 {
     Cmd c;
@@ -392,13 +386,13 @@ par_cmd(void)
 }
 
 /*
- * for	: ( FOR[EACH] | SELECT ) name ( "in" wordlist | INPAR wordlist OUTPAR )
-				{ SEPER } ( DO list DONE | INBRACE list OUTBRACE |
-				list ZEND | list1 )
+ * for  : ( FOR DINPAR expr SEMI expr SEMI expr DOUTPAR |
+ *    ( FOR[EACH] | SELECT ) name ( "in" wordlist | INPAR wordlist OUTPAR ) )
+ *	{ SEPER } ( DO list DONE | INBRACE list OUTBRACE | list ZEND | list1 )
  */
 
 /**/
-void
+static void
 par_for(Cmd c)
 {
     Forcmd f;
@@ -407,28 +401,48 @@ par_for(Cmd c)
     f = (Forcmd) make_forcmd();
     c->type = (tok == SELECT) ? CSELECT : CFOR;
     incmdpos = 0;
+    infor = tok == FOR ? 2 : 0;
     yylex();
-    if (tok != STRING || !isident(tokstr))
-	YYERRORV;
-    f->name = tokstr;
-    incmdpos = 1;
-    yylex();
-    if (tok == STRING && !strcmp(tokstr, "in")) {
-	f->inflag = 1;
-	incmdpos = 0;
+    if (tok == DINPAR) {
 	yylex();
-	c->args = par_wordlist();
-	if (tok != SEPER)
+	if (tok != DINPAR)
 	    YYERRORV;
-    } else if (tok == INPAR) {
-	f->inflag = 1;
-	incmdpos = 0;
+	f->name = tokstr;
 	yylex();
-	c->args = par_nl_wordlist();
-	if (tok != OUTPAR)
+	if (tok != DINPAR)
 	    YYERRORV;
+	f->condition = tokstr;
+	yylex();
+	if (tok != DOUTPAR)
+	    YYERRORV;
+	f->advance = tokstr;
+	infor = 0;
 	incmdpos = 1;
 	yylex();
+    } else {
+	infor = 0;
+	if (tok != STRING || !isident(tokstr))
+	    YYERRORV;
+	f->name = tokstr;
+	incmdpos = 1;
+	yylex();
+	if (tok == STRING && !strcmp(tokstr, "in")) {
+	    f->inflag = 1;
+	    incmdpos = 0;
+	    yylex();
+	    c->args = par_wordlist();
+	    if (tok != SEPER)
+		YYERRORV;
+	} else if (tok == INPAR) {
+	    f->inflag = 1;
+	    incmdpos = 0;
+	    yylex();
+	    c->args = par_nl_wordlist();
+	    if (tok != OUTPAR)
+		YYERRORV;
+	    incmdpos = 1;
+	    yylex();
+	}
     }
     incmdpos = 1;
     while (tok == SEPER)
@@ -459,12 +473,13 @@ par_for(Cmd c)
 
 /*
  * case	: CASE STRING { SEPER } ( "in" | INBRACE )
-				{ { SEPER } STRING { BAR STRING } OUTPAR list [ DSEMI ] }
+				{ { SEPER } STRING { BAR STRING } OUTPAR
+					list [ DSEMI | SEMIAMP ] }
 				{ SEPER } ( "esac" | OUTBRACE )
  */
 
 /**/
-void
+static void
 par_case(Cmd c)
 {
     int brflag;
@@ -509,7 +524,9 @@ par_case(Cmd c)
 	    yylex();
 	    break;
 	}
-	str = tokstr;
+	str = ncalloc(strlen(tokstr) + 2);
+	*str = ';';
+	strcpy(str + 1, tokstr);
 	incasepat = 0;
 	incmdpos = 1;
 	for (;;) {
@@ -536,9 +553,9 @@ par_case(Cmd c)
 		if (str[sl - 1] != Bar) {
 		    /* POSIX allows (foo*) patterns */
 		    int pct;
-		    char *s = str;
+		    char *s;
 
-		    for (s = str, pct = 0; *s; s++) {
+		    for (s = str + 1, pct = 0; *s; s++) {
 			if (*s == Inpar)
 			    pct++;
 			if (!pct)
@@ -555,7 +572,7 @@ par_case(Cmd c)
 			if (*s == Outpar)
 			    pct--;
 		    }
-		    if (*s || pct || s == str)
+		    if (*s || pct || s == str + 1)
 			YYERRORV;
 		    break;
 		} else {
@@ -577,7 +594,9 @@ par_case(Cmd c)
 	    yylex();
 	    break;
 	}
-	if (tok != DSEMI)
+	if(tok == SEMIAMP)
+	    *str = '&';
+	else if (tok != DSEMI)
 	    YYERRORV;
 	incasepat = 1;
 	incmdpos = 0;
@@ -604,7 +623,7 @@ par_case(Cmd c)
  */
 
 /**/
-void
+static void
 par_if(Cmd c)
 {
     struct ifcmd *i;
@@ -726,7 +745,7 @@ par_if(Cmd c)
  */
 
 /**/
-void
+static void
 par_while(Cmd c)
 {
     struct whilecmd *w;
@@ -765,7 +784,7 @@ par_while(Cmd c)
  */
 
 /**/
-void
+static void
 par_repeat(Cmd c)
 {
     c->type = CREPEAT;
@@ -806,7 +825,7 @@ par_repeat(Cmd c)
  */
 
 /**/
-void
+static void
 par_subsh(Cmd c)
 {
     c->type = (tok == INPAR) ? SUBSH : CURSH;
@@ -824,7 +843,7 @@ par_subsh(Cmd c)
  */
 
 /**/
-void
+static void
 par_funcdef(Cmd c)
 {
     nocorrect = 1;
@@ -863,7 +882,7 @@ par_funcdef(Cmd c)
  */
 
 /**/
-void
+static void
 par_time(Cmd c)
 {
     yylex();
@@ -876,7 +895,7 @@ par_time(Cmd c)
  */
 
 /**/
-void
+static void
 par_dinbrack(Cmd c)
 {
     c->type = COND;
@@ -898,7 +917,7 @@ par_dinbrack(Cmd c)
  */
 
 /**/
-Cmd
+static Cmd
 par_simple(Cmd c)
 {
     int isnull = 1;
@@ -1008,7 +1027,7 @@ par_cond(void)
  */
 
 /**/
-Cond
+static Cond
 par_cond_1(void)
 {
     Cond c, c2;
@@ -1038,7 +1057,7 @@ par_cond_1(void)
  */
 
 /**/
-Cond
+static Cond
 par_cond_2(void)
 {
     Cond c, c2;
@@ -1145,7 +1164,7 @@ par_cond_2(void)
  */
 
 /**/
-void
+static void
 par_redir(LinkList l)
 {
     struct redir *fn = (struct redir *)allocnode(N_REDIR);
@@ -1211,7 +1230,7 @@ par_redir(LinkList l)
  */
 
 /**/
-LinkList
+static LinkList
 par_wordlist(void)
 {
     LinkList l;
@@ -1229,7 +1248,7 @@ par_wordlist(void)
  */
 
 /**/
-LinkList
+static LinkList
 par_nl_wordlist(void)
 {
     LinkList l;
@@ -1244,13 +1263,15 @@ par_nl_wordlist(void)
 }
 
 /**/
-Cond
+static Cond
 par_cond_double(char *a, char *b)
 {
     Cond n = (Cond) make_cond();
 
-    if (a[0] != '-' || !a[1] || a[2])
-	COND_ERROR("parse error: condition expected: %s", a);
+    if (a[0] != '-' || !a[1] || a[2]) {
+	zerr("parse error: condition expected: %s", a, 0);
+	return NULL;
+    }
     n->left = (void *) b;
     n->type = a[1];
     n->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
@@ -1258,7 +1279,7 @@ par_cond_double(char *a, char *b)
 }
 
 /**/
-int
+static int
 get_cond_num(char *tst)
 {
     static char *condstrs[] =
@@ -1274,7 +1295,7 @@ get_cond_num(char *tst)
 }
 
 /**/
-Cond
+static Cond
 par_cond_triple(char *a, char *b, char *c)
 {
     Cond n = (Cond) make_cond();
@@ -1289,9 +1310,9 @@ par_cond_triple(char *a, char *b, char *c)
 	if ((t0 = get_cond_num(b + 1)) > -1)
 	    n->type = t0 + COND_NT;
 	else
-	    COND_ERROR("unrecognized condition: %s", b);
+	    zerr("unrecognized condition: %s", b, 0);
     } else
-	COND_ERROR("condition expected: %s", b);
+	zerr("condition expected: %s", b, 0);
     n->left = (void *) a;
     n->right = (void *) c;
     n->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
@@ -1299,7 +1320,7 @@ par_cond_triple(char *a, char *b, char *c)
 }
 
 /**/
-void
+static void
 yyerror(void)
 {
     int t0;
