@@ -1,11 +1,11 @@
 /*
- * $Id: prototypes.h,v 2.5 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: prototypes.h,v 3.1.2.5 1997/06/01 06:13:15 hzoli Exp $
  *
  * prototypes.h - prototypes header file
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -29,6 +29,8 @@
  *
  */
 
+#define GLOBAL_PROTOTYPES
+
 #include "builtin.pro"
 #include "compat.pro"
 #include "cond.pro"
@@ -44,29 +46,19 @@
 #include "loop.pro"
 #include "math.pro"
 #include "mem.pro"
+#include "options.pro"
 #include "params.pro"
 #include "parse.pro"
+#include "prompt.pro"
 #include "signals.pro"
 #include "subst.pro"
 #include "text.pro"
 #include "utils.pro"
 #include "watch.pro"
-#include "zle_hist.pro"
-#include "zle_main.pro"
-#include "zle_misc.pro"
-#include "zle_move.pro"
-#include "zle_refresh.pro"
-#include "zle_tricky.pro"
-#include "zle_utils.pro"
-#include "zle_vi.pro"
-#include "zle_word.pro"
+#include "module.pro"
+
+#undef GLOBAL_PROTOTYPES
 
-/* memory allocation routines - changed with permalloc()/heapalloc() */
- 
-/* real things in mem.c */
-extern void *(*alloc) _((size_t));
-extern void *(*ncalloc) _((size_t));
- 
 #ifndef HAVE_STDLIB_H
 char *malloc _((size_t));
 char *realloc _((void *, size_t));
@@ -155,3 +147,6 @@ extern char *strerror _((int errnum));
 /*** end of prototypes for functions in compat.c ***/
 /***************************************************/
 
+#ifndef HAVE_MEMMOVE
+extern void bcopy _((const void *, void *, int));
+#endif
