@@ -1,11 +1,11 @@
 /*
- * $Id: signals.c,v 2.24 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: signals.c,v 3.1.2.7 1997/06/01 06:13:15 hzoli Exp $
  *
  * signals.c - signals handling code
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "signals.pro"
  
 /* This is only used on machines that don't understand signal sets.  *
  * On SYSV machines this will represent the signals that are blocked *
@@ -110,6 +111,7 @@ intr(void)
         install_handler(SIGINT);
 }
 
+#if 0
 /* disable ^C interrupts */
  
 /**/
@@ -119,6 +121,7 @@ nointr(void)
     if (interact)
         signal_ignore(SIGINT);
 }
+#endif
  
 /* temporarily block ^C interrupts */
  
@@ -473,7 +476,8 @@ handler(int sig)
         else {
             extern int list_pipe, simple_pline;
  
-	    if (isset(PRIVILEGED) && isset(INTERACTIVE) && noerrexit < 0)
+	    if ((isset(PRIVILEGED) || isset(RESTRICTED)) &&
+		isset(INTERACTIVE) && noerrexit < 0)
 		zexit(SIGINT, 1);
             if (list_pipe || chline || simple_pline) {
                 breaks = loops;
@@ -485,7 +489,7 @@ handler(int sig)
 
 #ifdef SIGWINCH
     case SIGWINCH:
-        adjustwinsize(1);  /* check window size and adjust */
+        adjustwinsize();  /* check window size and adjust */
 	if (sigtrapped[SIGWINCH])
 	    dotrap(SIGWINCH);
         break;
