@@ -1,11 +1,11 @@
 /*
- * $Id: subst.c,v 2.52 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: subst.c,v 3.1.2.1 1997/06/01 06:13:15 hzoli Exp $
  *
  * subst.c - various substitutions
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "subst.pro"
 
 /* Do substitutions before fork. These are:
  *  - Process substitution: <(...), >(...), =(...)
@@ -89,7 +90,7 @@ prefork(LinkList list, int flags)
 }
 
 /**/
-LinkNode
+static LinkNode
 stringsubst(LinkList list, LinkNode node, int ssub)
 {
     int qt;
@@ -131,14 +132,14 @@ stringsubst(LinkList list, LinkNode node, int ssub)
 		endchar = Outpar;
 		str[-1] = '\0';
 		if (skipparens(Inpar, Outpar, &str))
-		    DPUTS(1, "Oops. parse error in command substitution");
+		    DPUTS(1, "BUG: parse error in command substitution");
 		str--;
 	    } else {
 		endchar = *str;
 		*str = '\0';
 
 		while (*++str != endchar)
-		    DPUTS(!*str, "Oops. parse error in command substitution");
+		    DPUTS(!*str, "BUG: parse error in command substitution");
 	    }
 	    *str++ = '\0';
 	    if (endchar == Outpar && str2[1] == '(' && str[-2] == ')') {
@@ -241,7 +242,7 @@ singsub(char **s)
  * resulted in an empty list                                            */
 
 /**/
-int
+static int
 multsub(char **s, char ***a, int *isarr, char *sep)
 {
     LinkList foo;
@@ -402,7 +403,7 @@ filesubstr(char **namptr, int assign)
 }
 
 /**/
-char *
+static char *
 strcatsub(char **d, char *pb, char *pe, char *src, int l, char *s, int glbsub)
 {
     int pl = pe - pb;
@@ -420,8 +421,6 @@ strcatsub(char **d, char *pb, char *pe, char *src, int l, char *s, int glbsub)
     return dest;
 }
 
-typedef int (*CompareFn) _((const void *, const void *));
-
 /**/
 int
 strpcmp(const void *a, const void *b)
@@ -449,13 +448,8 @@ int
 cstrpcmp(const void *a, const void *b)
 {
 #ifdef HAVE_STRCOLL
-# ifdef __GNUC__
-    char c[strlen(*(char **) a) + 1];
-    char d[strlen(*(char **) b) + 1];
-# else
-    char *c = halloc(strlen(*(char **) a) + 1);
-    char *d = halloc(strlen(*(char **) b) + 1);
-# endif
+    VARARR(char, c, strlen(*(char **) a) + 1);
+    VARARR(char, d, strlen(*(char **) b) + 1);
     char *s, *t;
     int   cmp;
 
@@ -479,13 +473,8 @@ int
 invcstrpcmp(const void *a, const void *b)
 {
 #ifdef HAVE_STRCOLL
-# ifdef __GNUC__
-    char c[strlen(*(char **) a) + 1];
-    char d[strlen(*(char **) b) + 1];
-# else
-    char *c = halloc(strlen(*(char **) a) + 1);
-    char *d = halloc(strlen(*(char **) b) + 1);
-# endif
+    VARARR(char, c, strlen(*(char **) a) + 1);
+    VARARR(char, d, strlen(*(char **) b) + 1);
     char *s, *t;
     int   cmp;
 
@@ -505,7 +494,7 @@ invcstrpcmp(const void *a, const void *b)
 }
 
 /**/
-char *
+static char *
 dopadding(char *str, int prenum, int postnum, char *preone, char *postone, char *premul, char *postmul)
 {
     char def[3], *ret, *t, *r;
@@ -657,7 +646,7 @@ get_strarg(char *s)
 }
 
 /**/
-int
+static int
 get_intarg(char **s)
 {
     char *t = get_strarg(*s + 1);
@@ -706,7 +695,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
     int vunset = 0;
     int spbreak = isset(SHWORDSPLIT) && !ssub && !qt;
     char *val = NULL, **aval = NULL;
-    unsigned int fwidth = 0;
+    int fwidth = 0;
     Value v;
     int flags = 0;
     int flnum = 0;
@@ -747,7 +736,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		untokenize(X = dupstring(s + 1));\
 		if (escapes) {\
 		    X = getkeystring(X, &klen, 3, NULL);\
-		    X = metafy(X, klen, META_USEHEAP);\
+		    X = metafy(X, klen, META_HREALLOC);\
 		}\
 	    }
 
@@ -1007,7 +996,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		fwidth = v->pm->ct ? v->pm->ct : strlen(val);
 		switch (v->pm->flags & (PM_LEFT | PM_RIGHT_B | PM_RIGHT_Z)) {
 		    char *t;
-		    unsigned int t0;
+		    int t0;
 
 		case PM_LEFT:
 		case PM_LEFT | PM_RIGHT_Z:
@@ -1374,7 +1363,12 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	int i;
 	LinkNode on = n;
 
-	if (!aval[0] && !plan9) {
+	if (!aval[0]) {
+	    if (plan9) {
+		uremnode(l, n);
+		*str = aptr;
+		return n;
+	    }
 	    if (aptr > (char *) getdata(n) &&
 		aptr[-1] == Dnull && *s == Dnull)
 		*--aptr = '\0', s++;
@@ -1385,62 +1379,53 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	    setdata(n, y);
 	    return n;
 	}
-	if (sortit) {
-	    static CompareFn sortfn[] = {
-		strpcmp, invstrpcmp, cstrpcmp, invcstrpcmp
-	    };
+	if (sortit && !copied)
+	    aval = arrdup(aval);
+	if (sortit == 1)
+	    qsort(aval, arrlen(aval), sizeof(char *), strpcmp);
 
-	    if (!copied)
-		aval = arrdup(aval);
+	else if (sortit == 2)
+	    qsort(aval, arrlen(aval), sizeof(char *), invstrpcmp);
+
+	else if (sortit == 3)
+	    qsort(aval, arrlen(aval), sizeof(char *), cstrpcmp);
+
+	else if (sortit)
+	    qsort(aval, arrlen(aval), sizeof(char *), invcstrpcmp);
 
-	    i = arrlen(aval);
-	    if (i && (*aval[i-1] || --i))
-		qsort(aval, i, sizeof(char *), sortfn[sortit-1]);
-	}
 	if (plan9) {
 	    LinkList tl = newlinklist();
-	    LinkNode tn;
 
-	    *--s = Marker;
 	    addlinknode(tl, s);
 	    if (!eval && !stringsubst(tl, firstnode(tl), ssub))
 		return NULL;
 	    *str = aptr;
-	    tn = firstnode(tl);
-	    while ((x = *aval++)) {
+	    if (empty(tl)) {
+		uremnode(l, n);
+		return n;
+	    }
+	    i = 0;
+	    while (aval[i]) {
+		LinkNode tn;
+
+		x = aval[i++];
 		if (prenum || postnum)
 		    x = dopadding(x, prenum, postnum, preone, postone,
 				  premul, postmul);
 		if (eval && parsestr(x))
 		    return NULL;
 		xlen = strlen(x);
-		for (tn = firstnode(tl);
-		     tn && *(y = (char *) getdata(tn)) == Marker;
-		     incnode(tn)) {
-		    strcatsub(&y, ostr, aptr, x, xlen, y + 1, globsubst);
+		for (tn = firstnode(tl); tn; incnode(tn)) {
+		    strcatsub(&y, ostr, aptr, x, xlen,
+			      (char *) getdata(tn), globsubst);
 		    if (qt && !*y && isarr != 2)
 			y = dupstring(nulstring);
-		    if (plan9)
-			setdata(n, (void *) y), plan9 = 0;
+		    if (i == 1)
+			setdata(n, (void *) y);
 		    else
 			insertlinknode(l, n, (void *) y), incnode(n);
 		}
 	    }
-	    for (; tn; incnode(tn)) {
-		y = (char *) getdata(tn);
-		if (*y == Marker)
-		    continue;
-		if (qt && !*y && isarr != 2)
-		    y = dupstring(nulstring);
-		if (plan9)
-		    setdata(n, (void *) y), plan9 = 0;
-		else
-		    insertlinknode(l, n, (void *) y), incnode(n);
-	    }
-	    if (plan9) {
-		uremnode(l, n);
-		return NULL;
-	    }
 	} else {
 	    x = aval[0];
 	    if (prenum || postnum)
@@ -1465,11 +1450,10 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		    return NULL;
 		if (qt && !*x && isarr != 2)
 		    y = dupstring(nulstring);
-		else {
-		    y = dupstring(x);
-		    if (globsubst)
-			tokenize(y);
-		}
+		else if (globsubst)
+		    tokenize(y = dupstring(x));
+		else
+		    y = x;
 		insertlinknode(l, n, (void *) y), incnode(n);
 	    }
 
@@ -1519,7 +1503,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
  */
 
 /**/
-char *
+static char *
 arithsubst(char *a, char **bptr, char *rest)
 {
     char *s = *bptr, *t, buf[DIGBUFSIZE];
@@ -1751,7 +1735,7 @@ modify(char **str, char **ptr)
 /* get a directory stack entry */
 
 /**/
-char *
+static char *
 dstackent(char ch, int val)
 {
     int backwards;
