@@ -1,11 +1,11 @@
 /*
- * $Id: system.h,v 2.7 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: system.h,v 3.1.2.4 1997/06/01 08:02:19 hzoli Exp $
  *
  * system.h - system configuration header file
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -48,6 +48,32 @@
 # define _(Args) ()
 #endif
 
+#ifndef HAVE_ALLOCA
+# define alloca halloc
+#else
+# ifdef __GNUC__
+#  define alloca __builtin_alloca
+# else
+#  if HAVE_ALLOCA_H
+#   include <alloca.h>
+#  else
+#   ifdef _AIX
+ #   pragma alloca
+#   else
+#    ifndef alloca
+char *alloca _((size_t));
+#    endif
+#   endif
+#  endif
+# endif
+#endif
+
+#ifdef HAVE_VARIABLE_LENGTH_ARRAYS
+# define VARARR(X,Y,Z)	X (Y)[Z]
+#else
+# define VARARR(X,Y,Z)	X *(Y) = (X *) alloca(sizeof(X) * (Z))
+#endif
+
 #ifdef HAVE_LIBC_H     /* NeXT */
 # include <libc.h>
 #endif
@@ -381,6 +407,10 @@ struct timezone {
 # define RLIMIT_VMEM RLIMIT_AS
 #endif
 
+#ifdef HAVE_SYS_CAPABILITY_H
+# include <sys/capability.h>
+#endif
+
 /* DIGBUFSIZ is the length of a buffer which can hold the -LONG_MAX-1 *
  * converted to printable decimal form including the sign and the     *
  * terminating null character. Below 0.30103 > lg 2.                  */
@@ -466,6 +496,14 @@ struct timezone {
 # define R_OK 4
 #endif
 
+#ifndef HAVE_MEMCPY
+# define memcpy memmove
+#endif
+
+#ifndef HAVE_MEMMOVE
+# define memmove(dest, src, len) bcopy(src, dest, len)
+#endif
+
 extern char **environ;     /* environment variable list */
 
 /* These variables are sometimes defined in, *
