@@ -1,9 +1,11 @@
 /*
+ * $Id: text.c,v 3.1.2.3 1997/06/01 06:13:15 hzoli Exp $
+ *
  * text.c - textual representations of syntax trees
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -28,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "text.pro"
 
 static char *tptr, *tbuf, *tlim;
 static int tsiz, tindent, tnewlins;
@@ -35,7 +38,7 @@ static int tsiz, tindent, tnewlins;
 /* add a character to the text buffer */
 
 /**/
-void
+static void
 taddchr(int c)
 {
     *tptr++ = c;
@@ -53,7 +56,7 @@ taddchr(int c)
 /* add a string to the text buffer */
 
 /**/
-void
+static void
 taddstr(char *s)
 {
     int sl = strlen(s);
@@ -71,6 +74,7 @@ taddstr(char *s)
     tptr += sl;
 }
 
+#if 0
 /* add an integer to the text buffer */
 
 /**/
@@ -82,11 +86,12 @@ taddint(int x)
     sprintf(buf, "%d", x);
     taddstr(buf);
 }
+#endif
 
 /* add a newline, or something equivalent, to the text buffer */
 
 /**/
-void
+static void
 taddnl(void)
 {
     int t0;
@@ -152,7 +157,7 @@ getjobtext(struct node *n)
 #define _whilecmd(X) ((struct whilecmd *) (X))
 
 /**/
-void
+static void
 gettext2(struct node *n)
 {
     Cmd nn;
@@ -233,13 +238,23 @@ gettext2(struct node *n)
 	case CFOR:
 	case CSELECT:
 	    taddstr((nn->type == CFOR) ? "for " : "select ");
-	    taddstr(nn->u.forcmd->name);
-	    if (nn->u.forcmd->inflag) {
-		taddstr(" in ");
-		taddlist(nn->args);
+	    if (nn->u.forcmd->condition) {
+		taddstr("((");
+		taddstr(nn->u.forcmd->name);
+		taddstr("; ");
+		taddstr(nn->u.forcmd->condition);
+		taddstr("; ");
+		taddstr(nn->u.forcmd->advance);
+		taddstr(")) do");
+	    } else {
+		taddstr(nn->u.forcmd->name);
+		if (nn->u.forcmd->inflag) {
+		    taddstr(" in ");
+		    taddlist(nn->args);
+		}
+		taddnl();
+		taddstr("do");
 	    }
-	    taddnl();
-	    taddstr("do");
 	    tindent++;
 	    taddnl();
 	    gt2(nn->u.forcmd->list);
@@ -302,12 +317,13 @@ gettext2(struct node *n)
 		    taddnl();
 		else
 		    taddchr(' ');
-		taddstr(*p);
+		taddstr(*p + 1);
 		taddstr(") ");
 		tindent++;
 		gt2(*l);
 		tindent--;
-		taddstr(" ;;");
+		taddstr(" ;");
+		taddchr(**p);
 	    }
 	    tindent--;
 	    if (tnewlins)
@@ -368,7 +384,7 @@ gettext2(struct node *n)
  * With addpar non-zero, parenthesise the subexpression. */
 
 /**/
-void
+static void
 getcond(Cond nm, int addpar)
 {
     static char *c1[] =
@@ -422,7 +438,7 @@ getcond(Cond nm, int addpar)
 }
 
 /**/
-void
+static void
 getsimptext(Cmd cmd)
 {
     LinkNode n;
@@ -451,7 +467,7 @@ getredirs(Cmd cmd)
     LinkNode n;
     static char *fstr[] =
     {
-	">", ">|", ">>", ">>|", "&>", "&>|", "&>>", "&>>|", "<", "<>",
+	">", ">|", ">>", ">>|", "&>", "&>|", "&>>", "&>>|", "<>", "<",
 	"<<", "<<-", "<<<", "<&", ">&", NULL /* >&- */, "<", ">"
     };
 
@@ -484,12 +500,12 @@ getredirs(Cmd cmd)
 	    break;
 #ifdef DEBUG
 	case CLOSE:
-	    DPUTS(1, "What?  CLOSE in getredirs?");
+	    DPUTS(1, "BUG: CLOSE in getredirs()");
 	    taddchr(f->fd1 + '0');
 	    taddstr(">&- ");
 	    break;
 	default:
-	    DPUTS(1, "unknown redirection in getredirs()");
+	    DPUTS(1, "BUG: unknown redirection in getredirs()");
 #endif
 	}
     }
@@ -497,7 +513,7 @@ getredirs(Cmd cmd)
 }
 
 /**/
-void
+static void
 taddlist(LinkList l)
 {
     LinkNode n;
