@@ -1,11 +1,11 @@
 /*
- * $Id: watch.c,v 2.3 1996/10/15 20:16:35 hzoli Exp $
+ * $Id: watch.c,v 3.1.2.2 1997/06/01 06:13:15 hzoli Exp $
  *
  * watch.c - login/logout watching
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -30,6 +30,7 @@
  */
 
 #include "zsh.h"
+#include "watch.pro"
 
 static int wtabsz;
 STRUCT_UTMP *wtab;
@@ -38,7 +39,7 @@ static time_t lastutmpcheck;
 /* get the time of login/logout for WATCH */
 
 /**/
-time_t
+static time_t
 getlogtime(STRUCT_UTMP *u, int inout)
 {
     FILE *in;
@@ -84,7 +85,7 @@ getlogtime(STRUCT_UTMP *u, int inout)
 #define END3	')'
 
 /**/
-char *
+static char *
 watch3ary(int inout, STRUCT_UTMP *u, char *fmt, int prnt)
 {
     int truth = 1, sep;
@@ -120,7 +121,7 @@ watch3ary(int inout, STRUCT_UTMP *u, char *fmt, int prnt)
 /* print a login/logout event */
 
 /**/
-char *
+static char *
 watchlog2(int inout, STRUCT_UTMP *u, char *fmt, int prnt, int fini)
 {
     char buf[40], buf2[80];
@@ -191,10 +192,10 @@ watchlog2(int inout, STRUCT_UTMP *u, char *fmt, int prnt, int fini)
 			fm2 = "%l:%M%p";
 			break;
 		    case 'T':
-			fm2 = "%k:%M";
+			fm2 = "%K:%M";
 			break;
 		    case 'w':
-			fm2 = "%a %e";
+			fm2 = "%a %f";
 			break;
 		    case 'W':
 			fm2 = "%m/%d/%y";
@@ -268,7 +269,7 @@ watchlog2(int inout, STRUCT_UTMP *u, char *fmt, int prnt, int fini)
 /* check the List for login/logouts */
 
 /**/
-void
+static void
 watchlog(int inout, STRUCT_UTMP *u, char **w, char *fmt)
 {
     char *v, *vv, sav;
@@ -331,7 +332,7 @@ watchlog(int inout, STRUCT_UTMP *u, char **w, char *fmt)
 /* compare 2 utmp entries */
 
 /**/
-int
+static int
 ucmp(STRUCT_UTMP *u, STRUCT_UTMP *v)
 {
     if (u->ut_time == v->ut_time)
@@ -342,7 +343,7 @@ ucmp(STRUCT_UTMP *u, STRUCT_UTMP *v)
 /* initialize the user List */
 
 /**/
-void
+static void
 readwtab(void)
 {
     STRUCT_UTMP *uptr;
