@@ -1,9 +1,11 @@
 /*
+ * $Id: zsh.h,v 3.1.2.9 1997/05/06 05:03:03 hzoli Exp $
+ *
  * zsh.h - standard header file
  *
  * This file is part of zsh, the Z shell.
  *
- * Copyright (c) 1992-1996 Paul Falstad
+ * Copyright (c) 1992-1997 Paul Falstad
  * All rights reserved.
  *
  * Permission is hereby granted, without written agreement and without
@@ -27,45 +29,33 @@
  *
  */
 
+#ifndef _ZSH_H
+#define _ZSH_H
+
 /* We use <config.h> instead of "config.h" so that a compilation        *
  * using -I. -I$srcdir will use ./config.h rather than $srcdir/config.h *
  * (which it would do because it found this file in $srcdir).           */
 #include <config.h>
-
-/*
- * Our longest integer type:  will be a 64 bit either if long already is,
- * or if we found some alternative such as long long.
- * Currently we only define this to be longer than a long if --enable-lfs
- * was given.  That enables internal use of 64-bit types even if
- * no actual large file support is present.
- *
- * This comes before system.h, where we use SIZEOF_ZLONG for
- * defining the digit buffer size.
- */
-#ifdef ZSH_64_BIT_TYPE
-# define SIZEOF_ZLONG 8
-typedef ZSH_64_BIT_TYPE zlong;
-#ifdef ZSH_64_BIT_UTYPE
-typedef ZSH_64_BIT_UTYPE zulong;
-#else
-typedef unsigned zlong zulong;
-#endif
-#else
-# define SIZEOF_ZLONG SIZEOF_LONG
-typedef long zlong;
-typedef unsigned long zulong;
-#endif
-
 #include <system.h>
 
+#ifndef IN_ZLE
+# define trashzle()      trashzleptr()
+# define zleread(X,Y,H)  zlereadptr(X,Y,H)
+# define spaceinline(X)  spaceinlineptr(X)
+# define gotword()       gotwordptr()
+# define refresh()       refreshptr()
+#endif /* !IN_ZLE */
+
+#ifndef IN_COMP
+# define compctlread(N,A,O,R) compctlreadptr(N,A,O,R)
+#endif /* !IN_COMP */
+
 /* A few typical macros */
 #define minimum(a,b)  ((a) < (b) ? (a) : (b))
 
 /* math.c */
 typedef int LV;
 
-#include "zle.h"
-
 /* Character tokens are sometimes casted to (unsigned char)'s.         * 
  * Unfortunately, some compilers don't correctly cast signed to        * 
  * unsigned promotions; i.e. (int)(unsigned char)((char) -1) evaluates * 
@@ -79,12 +69,6 @@ typedef int LV;
 # define STOUC(X)	((unsigned char)(X))
 #endif
 
-/* Meta together with the character following Meta denotes the character *
- * which is the exclusive or of 32 and the character following Meta.     *
- * This is used to represent characters which otherwise has special      *
- * meaning for zsh.  These are the characters for which the imeta() test *
- * is true: the null character, and the characters from Meta to Marker.  */
-
 #define Meta		((char) 0x83)
 
 /* Note that the fourth character in DEFAULT_IFS is Meta *
@@ -120,9 +104,6 @@ typedef int LV;
 
 #define INULL(x)	(((x) & 0xfc) == 0x98)
 
-/* Marker used in paramsubst for rc_expand_param */
-#define Marker		((char) 0x9c)
-
 /* chars that need to be quoted if meant literally */
 
 #define SPECCHARS "#$^*()=|{}[]`<>?~;&\n\t \\\'\""
@@ -159,38 +140,39 @@ enum {
     DINPAR,
     DOUTPAR,
     AMPERBANG,		/* 30 */
+    SEMIAMP,
     DOUTBRACK,
     STRING,
     ENVSTRING,
-    ENVARRAY,
-    ENDINPUT,		/* 35 */
+    ENVARRAY,		/* 35 */
+    ENDINPUT,
     LEXERR,
 
     /* Tokens for reserved words */
     BANG,	/* !         */
     DINBRACK,	/* [[        */
-    INBRACE,    /* {         */
-    OUTBRACE,   /* }         */	/* 40 */
+    INBRACE,    /* {         */	/* 40 */
+    OUTBRACE,   /* }         */
     CASE,	/* case      */
     COPROC,	/* coproc    */
     DO,		/* do        */
-    DONE,	/* done      */
-    ELIF,	/* elif      */ /* 45 */
+    DONE,	/* done      */ /* 45 */
+    ELIF,	/* elif      */
     ELSE,	/* else      */
     ZEND,	/* end       */
     ESAC,	/* esac      */
-    FI,		/* fi        */
-    FOR,	/* for       */ /* 50 */
+    FI,		/* fi        */ /* 50 */
+    FOR,	/* for       */
     FOREACH,	/* foreach   */
     FUNC,	/* function  */
     IF,		/* if        */
-    NOCORRECT,	/* nocorrect */
-    REPEAT,	/* repeat    */ /* 55 */
+    NOCORRECT,	/* nocorrect */ /* 55 */
+    REPEAT,	/* repeat    */
     SELECT,	/* select    */
     THEN,	/* then      */
     TIME,	/* time      */
-    UNTIL,	/* until     */
-    WHILE	/* while     */ /* 60 */
+    UNTIL,	/* until     */ /* 60 */
+    WHILE	/* while     */
 };
 
 /* Redirection types.  If you modify this, you may also have to modify *
@@ -206,8 +188,8 @@ enum {
     ERRWRITENOW,	/* >&| */
     ERRAPP,		/* >>& */
     ERRAPPNOW,		/* >>&| */
-    READ,		/* < */
     READWRITE,		/* <> */
+    READ,		/* < */
     HEREDOC,		/* << */
     HEREDOCDASH,	/* <<- */
     HERESTR,		/* <<< */
@@ -218,10 +200,11 @@ enum {
     OUTPIPE		/* > >(...) */
 };
 
-#define IS_APPEND_REDIR(X)    ((X)>=WRITE && (X)<=ERRAPPNOW && ((X) & 2))
-#define IS_CLOBBER_REDIR(X)   ((X)>=WRITE && (X)<=ERRAPPNOW && ((X) & 1))
+#define IS_WRITE_FILE(X)      ((X)>=WRITE && (X)<=READWRITE)
+#define IS_APPEND_REDIR(X)    (IS_WRITE_FILE(X) && ((X) & 2))
+#define IS_CLOBBER_REDIR(X)   (IS_WRITE_FILE(X) && ((X) & 1))
 #define IS_ERROR_REDIR(X)     ((X)>=ERRWRITE && (X)<=ERRAPPNOW)
-#define IS_READFD(X)          (((X)>=READ && (X)<=MERGEIN) || (X)==INPIPE)
+#define IS_READFD(X)          (((X)>=READWRITE && (X)<=MERGEIN) || (X)==INPIPE)
 #define IS_REDIROP(X)         ((X)>=OUTANG && (X)<=TRINANG)
 
 /* Flags for input stack */
@@ -230,7 +213,6 @@ enum {
 #define INP_HIST      (1<<2)	/* expanding history                       */
 #define INP_CONT      (1<<3)	/* continue onto previously stacked input  */
 #define INP_ALCONT    (1<<4)	/* stack is continued from alias expn.     */
-#define INP_LINENO    (1<<5)    /* update line number                      */
 
 /* Flags for metafy */
 #define META_REALLOC	0
@@ -240,6 +222,7 @@ enum {
 #define META_ALLOC	4
 #define META_NOALLOC	5
 #define META_HEAPDUP	6
+#define META_HREALLOC	7
 
 
 /**************************/
@@ -258,8 +241,8 @@ typedef struct cmdnam    *Cmdnam;
 typedef struct shfunc    *Shfunc;
 typedef struct builtin   *Builtin;
 typedef struct nameddir  *Nameddir;
+typedef struct module    *Module;
 
-typedef struct schedcmd  *Schedcmd;
 typedef struct process   *Process;
 typedef struct job       *Job;
 typedef struct value     *Value;
@@ -275,10 +258,8 @@ typedef struct complist  *Complist;
 typedef struct heap      *Heap;
 typedef struct heapstack *Heapstack;
 typedef struct histent   *Histent;
-typedef struct compctlp  *Compctlp;
-typedef struct compctl   *Compctl;
-typedef struct compcond  *Compcond;
 typedef struct forcmd    *Forcmd;
+typedef struct autofn    *AutoFn;
 
 typedef struct asgment  *Asgment;
 
@@ -303,6 +284,7 @@ struct linklist {
 /* Macros for manipulating link lists */
 
 #define addlinknode(X,Y) insertlinknode(X,(X)->last,Y)
+#define uaddlinknode(X,Y) uinsertlinknode(X,(X)->last,Y)
 #define empty(X)     ((X)->first == NULL)
 #define nonempty(X)  ((X)->first != NULL)
 #define firstnode(X) ((X)->first)
@@ -342,7 +324,8 @@ struct node {
 #define N_IF      8
 #define N_WHILE   9
 #define N_VARASG 10
-#define N_COUNT  11
+#define N_AUTOFN 11
+#define N_COUNT  12
 
 /* values for types[4] */
 
@@ -419,6 +402,7 @@ struct cmd {
 	struct whilecmd *whilecmd;
 	Sublist pline;
 	Cond cond;
+	AutoFn autofn;
 	void *generic;
     } u;
     LinkList args;		/* command & argmument List (char *'s)   */
@@ -440,6 +424,7 @@ struct cmd {
 #define CSELECT 10
 #define COND    11
 #define CARITH  12
+#define AUTOFN  13
 
 /* flags for command modifiers */
 #define CFLAG_EXEC	(1<<0)	/* exec ...    */
@@ -483,7 +468,9 @@ struct forcmd {			/* for/select */
 /* Cmd->args contains list of words to loop thru */
     int ntype;			/* node type                          */
     int inflag;			/* if there is an in ... clause       */
-    char *name;			/* parameter to assign values to      */
+    char *name;			/* initializer or parameter name      */
+    char *condition;		/* arithmetic terminating condition   */
+    char *advance;		/* evaluated after each loop          */
     List list;			/* list to look through for each name */
 };
 
@@ -515,6 +502,13 @@ struct whilecmd {
     List loop;			/* list to execute until condition met */
 };
 
+/* node for autoloading functions */
+
+struct autofn {
+    int ntype;			/* node type                             */
+    Shfunc shf;			/* the shell function to define	         */
+};
+
 /* The number of fds space is allocated for  *
  * each time a multio must increase in size. */
 #define MULTIOUNIT 8
@@ -597,12 +591,7 @@ struct job {
 #define STAT_INUSE	(1<<6)	/* this job entry is in use             */
 #define STAT_SUPERJOB	(1<<7)	/* job has a subjob                     */
 #define STAT_SUBJOB	(1<<8)	/* job is a subjob                      */
-#define STAT_WASSUPER   (1<<9)  /* was a super-job, sub-job needs to be */
-				/* deleted */
-#define STAT_CURSH	(1<<10)	/* last command is in current shell     */
-#define STAT_NOSTTY	(1<<11)	/* the tty settings are not inherited   */
-				/* from this job when it exits.         */
-#define STAT_ATTACH	(1<<12)	/* delay reattaching shell to tty       */
+#define STAT_CURSH	(1<<9)	/* last command is in current shell     */
 
 #define SP_RUNNING -1		/* fake status for jobs currently running */
 
@@ -651,6 +640,13 @@ struct heredocs {
     Redir rd;
 };
 
+struct dirsav {
+    int dirfd, level;
+    char *dirname;
+    dev_t dev;
+    ino_t ino;
+};
+
 /*******************************/
 /* Definitions for Hash Tables */
 /*******************************/
@@ -669,9 +665,7 @@ typedef void     (*FreeNodeFunc)   _((HashNode));
  * scanhashtable or scanmatchtable    */
 typedef void     (*ScanFunc)       _((HashNode, int));
 
-#ifdef ZSH_HASH_DEBUG
 typedef void (*PrintTableStats) _((HashTable));
-#endif
 
 /* hash table for standard open hashing */
 
@@ -681,10 +675,6 @@ struct hashtable {
     int ct;			/* number of elements                         */
     HashNode *nodes;		/* array of size hsize                        */
 
-#ifdef ZSH_HASH_DEBUG
-    char *tablename;		/* string containing name of the hash table */
-#endif
-
     /* HASHTABLE METHODS */
     HashFunc hash;		/* pointer to hash function for this table    */
     TableFunc emptytable;	/* pointer to function to empty table         */
@@ -699,8 +689,8 @@ struct hashtable {
     FreeNodeFunc freenode;	/* pointer to function to free a node         */
     ScanFunc printnode;		/* pointer to function to print a node        */
 
-#ifdef ZSH_HASH_DEBUG
-    PrintTableStats printinfo;	/* pointer to function to print table stats */
+#ifdef HASHTABLE_INTERNAL_MEMBERS
+    HASHTABLE_INTERNAL_MEMBERS	/* internal use in hashtable.c                */
 #endif
 };
 
@@ -800,6 +790,15 @@ struct builtin {
 
 #define BINF_TYPEOPTS   (BINF_TYPEOPT|BINF_PLUSOPTS)
 
+struct module {
+    char *nam;
+    int flags;
+    void *handle;
+    LinkList deps;
+};
+
+#define MOD_BUSY    (1<<0)
+
 /* node used in parameter hash table (paramtab) */
 
 struct param {
@@ -809,27 +808,30 @@ struct param {
 
     /* the value of this parameter */
     union {
+	void *data;		/* used by special parameter functions    */
 	char **arr;		/* value if declared array   (PM_ARRAY)   */
 	char *str;		/* value if declared string  (PM_SCALAR)  */
-	zlong val;		/* value if declared integer (PM_INTEGER) */
+	long val;		/* value if declared integer (PM_INTEGER) */
     } u;
 
     /* pointer to function to set value of this parameter */
     union {
 	void (*cfn) _((Param, char *));
-	void (*ifn) _((Param, zlong));
+	void (*ifn) _((Param, long));
 	void (*afn) _((Param, char **));
     } sets;
 
     /* pointer to function to get value of this parameter */
     union {
 	char *(*cfn) _((Param));
-	zlong (*ifn) _((Param));
+	long (*ifn) _((Param));
 	char **(*afn) _((Param));
     } gets;
 
+    /* pointer to function to unset this parameter */
+    void (*unsetfn) _((Param, int));
+
     int ct;			/* output base or field width            */
-    void *data;			/* used by getfns                        */
     char *env;			/* location in environment, if exported  */
     char *ename;		/* name of corresponding environment var */
     Param old;			/* old struct for use with local         */
@@ -861,16 +863,8 @@ struct param {
 #define PM_UNIQUE	(1<<10)	/* remove duplicates                          */
 #define PM_SPECIAL	(1<<11) /* special builtin parameter                  */
 #define PM_DONTIMPORT	(1<<12)	/* do not import this variable                */
-#define PM_UNSET	(1<<13)
-
-/* node for compctl hash table (compctltab) */
-
-struct compctlp {
-    HashNode next;		/* next in hash chain               */
-    char *nam;			/* command name                     */
-    int flags;			/* CURRENTLY UNUSED                 */
-    Compctl cc;			/* pointer to the compctl desc.     */
-};
+#define PM_RESTRICTED	(1<<13) /* cannot be changed in restricted mode       */
+#define PM_UNSET	(1<<14)	/* has null value                             */
 
 /* node for named directory hash table (nameddirtab) */
 
@@ -898,20 +892,6 @@ struct nameddir {
 #define PRINT_WHENCE_SIMPLE	(1<<5)
 #define PRINT_WHENCE_FUNCDEF	(1<<6)
 
-
-/******************************/
-/* Definitions for sched list */
-/******************************/
- 
-/* node in sched list */
-
-struct schedcmd {
-    struct schedcmd *next;
-    char *cmd;			/* command to run */
-    time_t time;		/* when to run it */
-};
-
-
 /***********************************/
 /* Definitions for history control */
 /***********************************/
@@ -929,8 +909,8 @@ struct histent {
     int flags;			/* Misc flags                       */
 };
 
-#define HIST_OLD	0x00000002	/* Command is already written to disk*/
-#define HIST_READ	0x00000004	/* Command was read back from disk*/
+#define HIST_OLD	0x00000001	/* Command is already written to disk*/
+#define HIST_READ	0x00000002	/* Command was read back from disk*/
 
 /* Parts of the code where history expansion is disabled *
  * should be within a pair of STOPHIST ... ALLOWHIST     */
@@ -941,146 +921,34 @@ struct histent {
 #define HISTFLAG_DONE   1
 #define HISTFLAG_NOEXEC 2
 #define HISTFLAG_RECALL 4
-#define HISTFLAG_SETTY  8
+
 
 /******************************************/
 /* Definitions for programable completion */
 /******************************************/
 
-struct compcond {
-    struct compcond *and, *or;	/* the next or'ed/and'ed conditions    */
-    int type;			/* the type (CCT_*)                    */
-    int n;			/* the array length                    */
-    union {			/* these structs hold the data used to */
-	struct {		/* test this condition                 */
-	    int *a, *b;		/* CCT_POS, CCT_NUMWORDS               */
-	}
-	r;
-	struct {		/* CCT_CURSTR, CCT_CURPAT,... */
-	    int *p;
-	    char **s;
-	}
-	s;
-	struct {		/* CCT_RANGESTR,... */
-	    char **a, **b;
-	}
-	l;
-    }
-    u;
-};
-
-#define CCT_UNUSED     0
-#define CCT_POS        1
-#define CCT_CURSTR     2
-#define CCT_CURPAT     3
-#define CCT_WORDSTR    4
-#define CCT_WORDPAT    5
-#define CCT_CURSUF     6
-#define CCT_CURPRE     7
-#define CCT_CURSUB     8
-#define CCT_CURSUBC    9
-#define CCT_NUMWORDS  10
-#define CCT_RANGESTR  11
-#define CCT_RANGEPAT  12
-
-/* Contains the real description for compctls */
-
-struct compctl {
-    int refc;			/* reference count                         */
-    struct compctl *next;	/* next compctl for -x                     */
-    unsigned long mask;		/* mask of things to complete (CC_*)       */
-    char *keyvar;		/* for -k (variable)                       */
-    char *glob;			/* for -g (globbing)                       */
-    char *str;			/* for -s (expansion)                      */
-    char *func;			/* for -K (function)                       */
-    char *explain;		/* for -X (explanation)                    */
-    char *prefix, *suffix;	/* for -P and -S (prefix, suffix)          */
-    char *subcmd;		/* for -l (command name to use)            */
-    char *hpat;			/* for -H (history pattern)                */
-    int hnum;			/* for -H (number of events to search)     */
-    struct compctl *ext;	/* for -x (first of the compctls after -x) */
-    struct compcond *cond;	/* for -x (condition for this compctl)     */
-    struct compctl *xor;	/* for + (next of the xor'ed compctls)     */
-};
-
-/* objects to complete */
-#define CC_FILES	(1<<0)
-#define CC_COMMPATH	(1<<1)
-#define CC_REMOVE	(1<<2)
-#define CC_OPTIONS	(1<<3)
-#define CC_VARS		(1<<4)
-#define CC_BINDINGS	(1<<5)
-#define CC_ARRAYS	(1<<6)
-#define CC_INTVARS	(1<<7)
-#define CC_SHFUNCS	(1<<8)
-#define CC_PARAMS	(1<<9)
-#define CC_ENVVARS	(1<<10)
-#define CC_JOBS		(1<<11)
-#define CC_RUNNING	(1<<12)
-#define CC_STOPPED	(1<<13)
-#define CC_BUILTINS	(1<<14)
-#define CC_ALREG	(1<<15)
-#define CC_ALGLOB	(1<<16)
-#define CC_USERS	(1<<17)
-#define CC_DISCMDS	(1<<18)
-#define CC_EXCMDS	(1<<19)
-#define CC_SCALARS	(1<<20)
-#define CC_READONLYS	(1<<21)
-#define CC_SPECIALS	(1<<22)
-#define CC_DELETE	(1<<23)
-#define CC_NAMED	(1<<24)
-#define CC_QUOTEFLAG	(1<<25)
-#define CC_EXTCMDS	(1<<26)
-#define CC_RESWDS	(1<<27)
-
-#define CC_RESERVED	(1<<31)
+/* Nothing special. */
+#define IN_NOTHING 0
+/* In command position. */
+#define IN_CMD     1
+/* In a mathematical environment. */
+#define IN_MATH    2
+/* In a condition. */
+#define IN_COND    3
+/* In a parameter assignment (e.g. `foo=bar'). */
+#define IN_ENV     4
 
 
 /******************************/
 /* Definition for zsh options */
 /******************************/
 
-/* Possible values of emulation: the order must match the OPT_*SH values *
- * below.                                                                */
+/* Possible values of emulation */
 
-enum {
-  EMULATE_CSH, /* C shell */
-  EMULATE_KSH, /* Korn shell */
-  EMULATE_SH,  /* Bourne shell */
-  EMULATE_ZSH  /* `native' mode */
-};
-
-/* the option name table */
-
-struct option {
-    char *name;			/* full name */
-    char id_zsh;		/* single letter name in zsh/csh mode */
-    char id_ksh;		/* single letter name in ksh/sh mode */
-    unsigned char flags;	/* see below */
-};
-
-#define optid(X) ( isset(SHOPTIONLETTERS) ? (X).id_ksh : (X).id_zsh )
-
-/* option flags: the order of the shells here must match the order of the *
- * EMULATE_ values above                                                  */
-
-#define OPT_CSH		0x01	/* option is set by default for csh */
-#define OPT_KSH		0x02	/* option is set by default for ksh */
-#define OPT_SH		0x04	/* option is set by default for sh */
-#define OPT_ZSH		0x08	/* option is set by default for zsh */
-
-#define OPT_ALL		(OPT_CSH|OPT_KSH|OPT_SH|OPT_ZSH)
-#define OPT_BOURNE	(OPT_KSH|OPT_SH)
-#define OPT_BSHELL	(OPT_KSH|OPT_SH|OPT_ZSH)
-#define OPT_NONBOURNE	(OPT_ALL & ~OPT_BOURNE)
-#define OPT_NONZSH	(OPT_ALL & ~OPT_ZSH)
-
-#define OPT_EMULATE	0x40	/* option is relevant to emulation */
-#define OPT_SPECIAL	0x80	/* option should never be set by emulate() */
-
-/* this can be ored with an option letter to invert its sense */
-
-#define OPT_REV		((char) 0x80)
+#define EMULATE_CSH  (1<<1) /* C shell */
+#define EMULATE_KSH  (1<<2) /* Korn shell */
+#define EMULATE_SH   (1<<3) /* Bourne shell */
+#define EMULATE_ZSH  (1<<4) /* `native' mode */
 
 /* option indices */
 
@@ -1144,6 +1012,7 @@ enum {
     INTERACTIVE,
     INTERACTIVECOMMENTS,
     KSHARRAYS,
+    KSHAUTOLOAD,
     KSHOPTIONPRINT,
     LISTAMBIGUOUS,
     LISTBEEP,
@@ -1164,7 +1033,6 @@ enum {
     OVERSTRIKE,
     PATHDIRS,
     POSIXBUILTINS,
-    PRINTEIGHTBIT,
     PRINTEXITVALUE,
     PRIVILEGED,
     PROMPTCR,
@@ -1177,6 +1045,7 @@ enum {
     RCQUOTES,
     RCS,
     RECEXACT,
+    RESTRICTED,
     RMSTARSILENT,
     SHFILEEXPANSION,
     SHGLOB,
@@ -1197,7 +1066,6 @@ enum {
 #undef isset
 #define isset(X) (opts[X])
 #define unset(X) (!opts[X])
-#define defset(X) (!!(optns[X].flags & (1 << emulation)))
 
 #define interact (isset(INTERACTIVE))
 #define jobbing  (isset(MONITOR))
@@ -1277,8 +1145,7 @@ struct ttyinfo {
 #define TCALLATTRSOFF  21
 #define TCSTANDOUTEND  22
 #define TCUNDERLINEEND 23
-#define TCHORIZPOS     24
-#define TC_COUNT       25
+#define TC_COUNT       24
 
 #define tccan(X) (tclen[X])
 
@@ -1304,19 +1171,14 @@ struct ttyinfo {
 
 #include "ztype.h"
 
-#define cmdpush(X) do { \
-		       if (cmdsp >= 0 && cmdsp < 256) \
-			   cmdstack[cmdsp++] = (X); \
-		   } while (0)
+#define cmdpush(X) if (!(cmdsp >= 0 && cmdsp < 256)) {;} else cmdstack[cmdsp++]=(X)
 #ifdef DEBUG
-# define cmdpop()  do { \
-		       if (cmdsp <= 0) { \
-			   fputs("BUG: cmdstack empty\n", stderr); \
-			   fflush(stderr); \
-		       } else cmdsp--; \
-		   } while (0)
+# define cmdpop()  if (cmdsp <= 0) { \
+			fputs("BUG: cmdstack empty\n", stderr); \
+			fflush(stderr); \
+		   } else cmdsp--
 #else
-# define cmdpop()  do { if (cmdsp > 0) cmdsp--; } while (0)
+# define cmdpop()   if (cmdsp <= 0) {;} else cmdsp--
 #endif
 
 #define CS_FOR          0
@@ -1366,8 +1228,8 @@ struct ttyinfo {
 		} while(0)
 
 # define LASTALLOC_RETURN \
-			for (nonlocal_useheap ? global_heapalloc() : \
-			     global_permalloc(); 1;) return
+			if ((nonlocal_useheap ? global_heapalloc() : \
+			     global_permalloc()), 0) {;} else return
 #else
 # define HEAPALLOC	do { int nonlocal_useheap = global_heapalloc(); \
 			alloc_stackp++; do
@@ -1381,8 +1243,8 @@ struct ttyinfo {
 		} while(0)
 
 # define LASTALLOC_RETURN \
-			for (nonlocal_useheap ? global_heapalloc() : \
-			     global_permalloc(); alloc_stackp--;) return
+			if ((nonlocal_useheap ? global_heapalloc() : \
+			    global_permalloc()),alloc_stackp--,0){;}else return
 #endif
 
 /****************/
@@ -1390,11 +1252,10 @@ struct ttyinfo {
 /****************/
 
 #ifdef DEBUG
-# define DPUTS(X,Y) do { if ((X)) dputs(Y); } while (0)
-# define MUSTUSEHEAP(X) do { if (useheap == 0) { \
-		    fprintf(stderr, "BUG: permanent allocation in %s\n", X); \
-		    fflush(stderr); \
-		} } while (0)
+# define DPUTS(X,Y) if (!(X)) {;} else dputs(Y)
+# define MUSTUSEHEAP(X) if (useheap) {;} else \
+		fprintf(stderr, "BUG: permanent allocation in %s\n", X), \
+		fflush(stderr)
 #else
 # define DPUTS(X,Y)
 # define MUSTUSEHEAP(X)
@@ -1410,12 +1271,30 @@ struct ttyinfo {
 #define ZSIG_IGNORED	(1<<1)
 #define ZSIG_FUNC	(1<<2)
 
+/*********************************/
+/* Dynamic builtin table munging */
+/*********************************/
+
+typedef struct binlist *BinList;
+
+#define BINL_ADDED (1<<0)
+
+struct binlist {
+    char *nam;
+    int binf;
+    HandlerFunc func;
+    int minargs, maxargs, funcid;
+    char *optstr, *defopts;
+    int flags;
+};
+
 /***********************/
 /* Shared header files */
 /***********************/
 
-#include "version.h"
 #include "signals.h"
-#include "globals.h"
 #include "prototypes.h"
+#include "globals.h"
 #include "hashtable.h"
+
+#endif /*!_ZSH_H*/
