@@ -58,6 +58,9 @@
 /* Define to 1 if compiler incorrectly cast signed to unsigned */
 #undef BROKEN_SIGNED_TO_UNSIGNED_CASTING
 
+/* Define to 1 if compiler supports variable-length arrays */
+#undef HAVE_VARIABLE_LENGTH_ARRAYS
+
 /* Define if your system defines TIOCGWINSZ in sys/ioctl.h.  */
 #undef GWINSZ_IN_SYS_IOCTL
 
@@ -150,6 +153,9 @@
 /* Define to 1 if struct timezone is defined by a system header */
 #undef HAVE_STRUCT_TIMEZONE
 
+/* Define if your system's typeahead disappears from the shell editor. */
+#undef CLOBBERS_TYPEAHEAD
+
 /* Define to 1 if there is a prototype defined for brk() on your system */
 #undef HAVE_BRK_PROTO
 
@@ -174,26 +180,14 @@
 /* Define to 1 if /bin/sh does not interpret \ escape sequences */
 #undef SH_USE_BSD_ECHO
 
-/* Define to `unsigned long' if <sys/types.h> doesn't define. */
-#undef ino_t
+/* Define to 1 if an underscore has to be prepended to dlsym() argument */
+#undef DLSYM_NEEDS_UNDERSCORE
 
-/*
- * Definitions used when a long is less than eight byte, to try to
- * provide some support for eight byte operations.
- *
- * Note that ZSH_64_BIT_TYPE, OFF_T_IS_64_BIT, INO_T_IS_64_BIT do *not* get
- * defined if long is already 64 bits, since in that case no special handling
- * is required.
- */
+/* Define to 1 if multiple modules defining the same symbol are OK */
+#undef DYNAMIC_NAME_CLASH_OK
 
-/* Define to a 64 bit integer type if there is one, but long is shorter */
-#undef ZSH_64_BIT_TYPE
+/* The exension used for dynamically loaded modules */
+#undef DL_EXT
 
-/* Define to an unsigned variant of ZSH_64_BIT_TYPE if that is defined */
-#undef ZSH_64_BIT_UTYPE
-
-/* Define to 1 if off_t is 64 bit (for large file support) */
-#undef OFF_T_IS_64_BIT
-
-/* Define to 1 if ino_t is 64 bit (for large file support) */
-#undef INO_T_IS_64_BIT
+/* Define to 1 if you want to use dynamically loaded modules */
+#undef DYNAMIC
